package dev.pfaff.altargui;

import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.util.GrowableMemorySegment;
import dev.pfaff.altargui.pathfinder.PFCanvasRef;
import dev.pfaff.altargui.pathfinder.Pathfinder;
import dev.pfaff.altargui.pathfinder.PathfinderGlobals;
import dev.pfaff.altargui.render.BufferOverflowException;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.PFCanvasDrawer;
import dev.pfaff.altargui.render.RenderObject;
import dev.pfaff.altargui.texture.SinglePixelLoadableTexture;
import dev.pfaff.altargui.texture.TextureAllocator;
import dev.pfaff.altargui.texture.TextureHandle;
import dev.pfaff.altargui.util.OpenArrayList;
import dev.pfaff.altargui.util.Preconditions;
import dev.pfaff.altargui.util.ThreadFuture;
import dev.pfaff.altargui.widgets.*;
import dev.pfaff.glass.opengl.enums.GLEnableCap;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.log4truth.NamedLogger;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import org.jetbrains.annotations.NotNull;

import java.lang.foreign.MemorySegment;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.texture.TextureAllocator.MAX_TEXTURE_SIZE;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.lang.foreign.ValueLayout.*;

public final class CanvasImpl implements Canvas, AutoCloseable {
	private static final NamedLogger LOGGER = NamedLogger.forCaller();
	private static final List<String> TAGS = List.of(DEBUG);
	public static final List<String> DRAW_TAGS = List.of(DEBUG, "CANVAS_DRAW");

	// TODO: dynamic debug layer registration
	public static final List<DebugLayer> DEBUG_LAYERS = List.of(
			FiniteScrollListWidget.DEBUG_LAYER_PRE_UNSCISSORED,
			FiniteScrollListWidget.DEBUG_LAYER_PRE_SCISSORED,
			FiniteScrollListWidget.DEBUG_LAYER_POST_UNSCISSORED,
			MultiChildWidget.DEBUG_LAYER,
			PaddedWidget.DEBUG_LAYER_WHOLE,
			PaddedWidget.DEBUG_LAYER_INSIDE,
			ScrollWidget.DEBUG_LAYER_PRE_UNSCISSORED,
			ScrollWidget.DEBUG_LAYER_PRE_SCISSORED,
			ScrollWidget.DEBUG_LAYER_POST_UNSCISSORED,
			ClipWidget.DEBUG_LAYER_PRE_UNSCISSORED,
			ClipWidget.DEBUG_LAYER_PRE_SCISSORED,
			ClipWidget.DEBUG_LAYER_POST_UNSCISSORED
	);

	static {
		for (int i = 0; i < DEBUG_LAYERS.size(); i++) {
			DEBUG_LAYERS.get(i).setId(i);
		}
	}

	private final OpenArrayList<RenderObject> uploadablesNormal = OpenArrayList.wrap(new RenderObject[1024]);
	private final OpenArrayList<RenderObject> uploadablesTooltip = OpenArrayList.wrap(new RenderObject[1024]);
	private final OpenArrayList<PFCanvasRef> canvasPool = OpenArrayList.wrap(new PFCanvasRef[1024]);
	private int currentCanvasNormal = -1;
	private int currentCanvasTooltip = -1;
	private int nextPFCanvasId = 0;
	private final GrowableMemorySegment uploadBufNormal = new GrowableMemorySegment(1024);
	private final GrowableMemorySegment uploadBufTooltip = new GrowableMemorySegment(1024);
	private final GrowableMemorySegment drawBuf = new GrowableMemorySegment(1024);
	private int uploadBufNormalPos;
	private int uploadBufTooltipPos;
	private final LayeredRenderBuffer outputBuffer = Config.LAYERED_RENDER_BUFFER_IMPL.get();
	private int combinedLength;
	private Layer layer = Layer.Normal;
	private int currentProgram = -1;
	private double scaleFactor;
	private int windowWidth;
	private int windowHeight;

	//	private final TextureArray texture = new TextureArray();
	private final TextureAllocator texAlloc = new TextureAllocator(MAX_TEXTURE_SIZE, 1);

	private final IntArrayList scissorStack = new IntArrayList();

	public final BitSet enabledDebugLayers = new BitSet(DEBUG_LAYERS.size());

	@SuppressWarnings("FieldCanBeLocal")
	private final TextureHandle[] keepAliveTextures;

	public CanvasImpl() {
		{
			var keepAliveTextures = this.keepAliveTextures = new TextureHandle[3];
			var texAlloc = this.texAlloc;
			texAlloc.reset();
			// preload some textures
			var tasks = List.of(
					texAlloc.alloc(SinglePixelLoadableTexture.WHITE)
			);
			ThreadFuture.spawnVirtualTask(() -> {
				for (int i = 0; i < tasks.size(); i++) {
					var handle = tasks.get(i).join();
					keepAliveTextures[i] = handle;
				}
			});
		}
	}

	public void nextDebugLayer(boolean forward) {
		int activeDebugLayer = this.enabledDebugLayers.nextSetBit(0);
		this.enabledDebugLayers.clear();
		int next = activeDebugLayer + (forward ? 1 : -1);
		if (next != -1 && next != DEBUG_LAYERS.size()) {
			this.enabledDebugLayers.set(next == -2 ? DEBUG_LAYERS.size() - 1 : next);
		}
	}

	@Override
	public boolean isDebugLayerEnabled(DebugLayer layer) {
		return this.enabledDebugLayers.get(layer.id());
	}

	public void reset() {
		this.uploadBufNormalPos = 0;
		this.uploadBufTooltipPos = 0;
		this.uploadablesNormal.clear();
		this.uploadablesTooltip.clear();
		this.currentCanvasNormal = -1;
		this.currentCanvasTooltip = -1;
		this.nextPFCanvasId = 0;
		this.layer = Layer.Normal;
		this.combinedLength = 0;
		this.currentProgram = -1;

		if (!this.scissorStack.isEmpty()) {
			throw new IllegalStateException(STR."Caught attempt to reset the canvas without popping all scissors: \{this.scissorStack.size()} remain");
		}
	}

	public void setWindowProps(double scaleFactor, int windowWidth, int windowHeight) {
		this.scaleFactor = scaleFactor;
		this.windowWidth = windowWidth;
		this.windowHeight = windowHeight;

		PathfinderGlobals.renderer.setViewport(0, 0, windowWidth, windowHeight);
		PathfinderGlobals.renderer.setWindowSize(windowWidth, windowHeight);
		PathfinderGlobals.renderer.destFramebufferSizeChanged();
	}

	private static int write(GrowableMemorySegment buffer, int pos, int additional, Writer writer) {
		buffer.ensureTotalCapacity(pos + additional);
		var retAdditional = writer.write(buffer.ms().asSlice(pos));
		assert retAdditional == additional;
		pos += additional;
		return pos;
	}

	public int renderBufferByteCount() {
//		return Arrays.stream(this.outputBuffers).mapToInt(LayeredRenderBuffer::byteCount).sum();
		return this.outputBuffer.byteCount();
	}

	@FunctionalInterface
	private interface Writer {
		/**
		 * @return the number of bytes written.
		 */
		int write(MemorySegment ms);
	}

	private void writeUpload(int additional, Writer writer) {
		this.writeUpload(this.layer, additional, writer);
	}

	private void writeUpload(Layer layer, int additional, Writer writer) {
		switch (layer) {
			case Normal -> this.uploadBufNormalPos = write(this.uploadBufNormal, this.uploadBufNormalPos, additional, writer);
			case Tooltip -> this.uploadBufTooltipPos = write(this.uploadBufTooltip, this.uploadBufTooltipPos, additional, writer);
		}
	}

	private int writeDraw(int drawBufPos, int additional, Writer writer) {
		return write(this.drawBuf, drawBufPos, additional, writer);
	}

	/**
	 * If the program differs from the current program, emits a draw call for the current layer.
	 */
	private int writeDrawLayer(int drawBufPos, int newProgram) {
		assert this.outputBuffer.currentProgram() == this.currentProgram : "State inconsistency";

		var currentProgram = this.currentProgram;
		if (newProgram != currentProgram) {
			return writeEndLayer(drawBufPos, currentProgram, newProgram);
		}
		return drawBufPos;
	}

	private int writeEndLayer(int drawBufPos, int currentProgram, int newProgram) {
		LOGGER.log(() -> STR."switching active program from \{currentProgram} to \{newProgram} at \{drawBufPos}", TAGS);

		this.currentProgram = newProgram;
		if (currentProgram != -1) {
			if (!this.outputBuffer.endCurrentLayer()) throw new AssertionError();

			// write both commands at once
			return writeDraw(drawBufPos, 1, ms -> {
				ms.set(JAVA_BYTE, 0, DrawCommand.DRAW.asByte());
				return 1;
			});
		}
		return drawBufPos;
	}

	private static short checkScissorValue(float i, String name) {
		Preconditions.checkArgument(i >= Short.MIN_VALUE && i <= Short.MAX_VALUE, () -> {
			return STR."Expected scissor coordinates and dimensions in the range [\{Short.MIN_VALUE}, \{Short.MAX_VALUE}], found \{name}=\{i}";
		});
		return (short) i;
	}

	@Override
	public TextureAllocator texAlloc() {
		return texAlloc;
	}

	public void internalPushScissor(float x, float y, float width, float height) {
		x = (float) (x * this.scaleFactor);
		height = (float) (height * this.scaleFactor);
		y = this.windowHeight - (float) (y * this.scaleFactor) - height;
		width = (float) (width * this.scaleFactor);

		var x_ = checkScissorValue(x, "x");
		var y_ = checkScissorValue(y, "y");
		var width_ = checkScissorValue(width, "width");
		var height_ = checkScissorValue(height, "height");
		writeUpload(Byte.BYTES + Short.BYTES * 4, ms -> {
			int pos = 0;
			ms.set(JAVA_BYTE, pos++, UploadCommand.PUSH_SCISSOR.asByte());
			ms.set(JAVA_SHORT_UNALIGNED, pos, x_);
			pos += (int) JAVA_SHORT_UNALIGNED.byteSize();
			ms.set(JAVA_SHORT_UNALIGNED, pos, y_);
			pos += (int) JAVA_SHORT_UNALIGNED.byteSize();
			ms.set(JAVA_SHORT_UNALIGNED, pos, width_);
			pos += (int) JAVA_SHORT_UNALIGNED.byteSize();
			ms.set(JAVA_SHORT_UNALIGNED, pos, height_);
			pos += (int) JAVA_SHORT_UNALIGNED.byteSize();
			return pos;
		});
	}

	public void internalPopScissor() {
		writeUpload(1, ms -> {
			ms.set(JAVA_BYTE, 0, UploadCommand.POP_SCISSOR.asByte());
			return 1;
		});
	}

	private OpenArrayList<RenderObject> getUploadablesBuffer(Layer layer) {
		return switch (layer) {
			case Normal -> this.uploadablesNormal;
			case Tooltip -> this.uploadablesTooltip;
		};
	}

	private OpenArrayList<RenderObject> getUploadablesBuffer() {
		return this.getUploadablesBuffer(this.layer);
	}

	private int getCurrentPFCanvas(Layer layer) {
		return switch (layer) {
			case Normal -> this.currentCanvasNormal;
			case Tooltip -> this.currentCanvasTooltip;
		};
	}

	private int getCurrentPFCanvas() {
		return this.getCurrentPFCanvas(this.layer);
	}

	private void setCurrentPFCanvas(Layer layer, int canvas) {
		switch (layer) {
			case Normal -> this.currentCanvasNormal = canvas;
			case Tooltip -> this.currentCanvasTooltip = canvas;
		}
	}

	private void setCurrentPFCanvas(int canvas) {
		this.setCurrentPFCanvas(this.layer, canvas);
	}

	private GrowableMemorySegment getBuffer(Layer layer) {
		return switch (layer) {
			case Normal -> this.uploadBufNormal;
			case Tooltip -> this.uploadBufTooltip;
		};
	}

	private int getBufferPos(Layer layer) {
		return switch (layer) {
			case Normal -> this.uploadBufNormalPos;
			case Tooltip -> this.uploadBufTooltipPos;
		};
	}

	private GrowableMemorySegment getBuffer() {
		return this.getBuffer(this.layer);
	}

	@Override
	public void setLayer(@NotNull Layer layer) {
		this.layer = Objects.requireNonNull(layer);
	}

	public void iSpecUpload(RenderObject uploadable, int size) {
		this.combinedLength += size;
		this.getUploadablesBuffer().add(uploadable);
		var pfCanvas = this.getCurrentPFCanvas();
		if (pfCanvas != -1) {
			this.setCurrentPFCanvas(-1);
		}
		writeUpload(pfCanvas != -1
				? UploadCommand.FLUSH_PF.sizeWithBytecode + UploadCommand.UPLOAD.sizeWithBytecode
				: UploadCommand.UPLOAD.sizeWithBytecode, ms -> {
			boolean flushPfCanvas = pfCanvas != -1;
			if (flushPfCanvas) {
				ms.set(JAVA_BYTE, 0, UploadCommand.FLUSH_PF.asByte());
				ms.set(JAVA_INT_UNALIGNED, 1, pfCanvas);
			}
			ms.set(JAVA_BYTE, flushPfCanvas ? UploadCommand.FLUSH_PF.sizeWithBytecode : 0, UploadCommand.UPLOAD.asByte());
			return flushPfCanvas
					? UploadCommand.FLUSH_PF.sizeWithBytecode + UploadCommand.UPLOAD.sizeWithBytecode
					: UploadCommand.UPLOAD.sizeWithBytecode;
		});
	}

	@Override
	public void upload(RenderObject uploadable) {
		uploadable.uSpecUploadCanvas(this);
	}

	@Override
	public void upload(PFCanvasDrawer drawer) {
		var canvasPool = this.canvasPool;
		PFCanvasRef canvas;
		if (this.getCurrentPFCanvas() == -1) {
			canvas = getNewCanvasIdAndMarkNotFinished(canvasPool);
		} else {
			canvas = canvasPool.getUnchecked(this.getCurrentPFCanvas());
		}
		drawer.draw(canvas);
	}

	private PFCanvasRef getNewCanvasIdAndMarkNotFinished(OpenArrayList<PFCanvasRef> canvasPool) {
		int id;
		setCurrentPFCanvas(id = this.nextPFCanvasId++);
		PFCanvasRef canvas;
		if (id >= canvasPool.size()) {
			canvas = allocNewCanvas(canvasPool);
		} else {
			canvas = canvasPool.getUnchecked(id);
		}
		canvas.setSize(this.windowWidth, this.windowHeight);
		return canvas;
	}

	private PFCanvasRef allocNewCanvas(OpenArrayList<PFCanvasRef> canvasPool) {
		var canvas = Pathfinder.createCanvas(PathfinderGlobals.fontContext);
		canvasPool.add(canvas);
		return canvas;
	}

	public void draw() {
		// shouldn't be necessary
//		for (var program : Layer.VALUES) {
//			int program = layerPrograms[program.ordinal()];
//			if (program != -1) {
//				write(program, 1, ms -> {
//					ms.set(JAVA_BYTE, 0, Command.DRAW.asByte());
//					return 1;
//				});
//				layerPrograms[program.ordinal()] = -1;
//			}
//		}

		int bufUpr = uploadAndPrepareDrawCommands();
		executeDrawCommands(bufUpr);
	}

	private int uploadAndPrepareDrawCommands() {
		if (!this.scissorStack.isEmpty()) {
			throw new AssertionError("Not all scissors were popped during preparation!");
		}
		var outputBuffer = this.outputBuffer;
		// TODO: document this capacity
		outputBuffer.beginUploading(this.combinedLength * 2);
		this.drawBuf.ensureTotalCapacity(this.uploadBufNormalPos + this.uploadBufTooltipPos);
		int drawBufPos = 0;
		{
			for (var layer : Layer.VALUES) {
				var buf = this.getBuffer(layer).ms();
				var uploadables = this.getUploadablesBuffer(layer);
				var uploadablesA = uploadables.elements();
				var uploadablesUpr = uploadables.size();
				if (uploadablesUpr > uploadablesA.length) throw new AssertionError();
				int uploadablesI = 0;

				var pos = 0;
				var bufUpr = this.getBufferPos(layer);
				while (pos < bufUpr) {
					byte cmd = buf.get(JAVA_BYTE, pos++);
					switch (UploadCommand.VALUES.get(cmd)) {
						case PUSH_SCISSOR -> {
							drawBufPos = writeDrawLayer(drawBufPos, -1);

							{
								var drawBufPos_ = drawBufPos;
								LOGGER.log(() -> STR."pushing scissor at \{drawBufPos_}", TAGS);
							}

							var bufSlice = buf.asSlice(pos-1, Byte.BYTES + Short.BYTES * 4);
							drawBufPos = writeDraw(drawBufPos, (int) bufSlice.byteSize(), ms -> {
								ms.copyFrom(bufSlice);
								return (int) bufSlice.byteSize();
							});
							pos += Short.BYTES * 4;
						}
						case POP_SCISSOR -> {
							drawBufPos = writeDrawLayer(drawBufPos, -1);
							{
								var drawBufPos_ = drawBufPos;
								LOGGER.log(() -> STR."popping scissor at \{drawBufPos_}", TAGS);
							}
							drawBufPos = writeDraw(drawBufPos, 1, ms -> {
								ms.set(JAVA_BYTE, 0, DrawCommand.POP_SCISSOR.asByte());
								return 1;
							});
						}
						case UPLOAD -> {
							if (uploadablesI >= uploadablesUpr) {
								throw new BufferOverflowException("Overflowed uploadable object buffer");
							}
							var uploadable = uploadablesA[uploadablesI++];
							int program = uploadable.program();
							drawBufPos = writeDrawLayer(drawBufPos, program);
							outputBuffer.iSpecUpload(uploadable, program);
						}
						case FLUSH_PF -> {
							var canvas = buf.get(JAVA_INT_UNALIGNED, pos);
							pos += (int) JAVA_INT_UNALIGNED.byteSize();
							drawBufPos = writeDrawLayer(drawBufPos, -1);
							drawBufPos = writeDraw(drawBufPos, 1 + (int) JAVA_INT.byteSize(), ms -> {
								ms.set(JAVA_BYTE, 0, DrawCommand.DRAW_PF.asByte());
								ms.set(JAVA_INT_UNALIGNED, 1, canvas);
								return 1 + (int) JAVA_INT.byteSize();
							});
						}
					}
				}

				if (uploadablesI != uploadablesUpr) {
					throw new IllegalStateException(STR."Not all objects uploaded: \{uploadablesI}/\{uploadablesUpr} uploaded, \{uploadablesUpr - uploadablesI} remaining");
				}

				int canvas = this.getCurrentPFCanvas(layer);
				if (canvas != -1) {
					drawBufPos = writeDrawLayer(drawBufPos, -1);
					drawBufPos = writeDraw(drawBufPos, 1 + (int) JAVA_INT.byteSize(), ms -> {
						ms.set(JAVA_BYTE, 0, DrawCommand.DRAW_PF.asByte());
						ms.set(JAVA_INT_UNALIGNED, 1, canvas);
						return 1 + (int) JAVA_INT.byteSize();
					});
				}
			}

			drawBufPos = writeDrawLayer(drawBufPos, -1);
		}
		outputBuffer.finishUploading();

		return drawBufPos;
	}

	private void executeDrawCommands(final int bufUpr) {
		var outputBuffer = this.outputBuffer;
		var texAllocUnbinder = this.texAlloc.boundForDraw();
		try {
//			LOGGER.log(Level.INFO, STR."final flush [\{rangeStart}, \{rangeStart + rangeCount}) n=\{rangeCount}");

			try (var drawer = outputBuffer.drawer()) {
				if (!this.scissorStack.isEmpty()) {
					throw new AssertionError("Not all scissors were popped during buffer building!");
				}

				var canvasPool = this.canvasPool;

				var buf = this.drawBuf.ms();
				var pos = 0;
				while (pos < bufUpr) {
					byte cmd = buf.get(JAVA_BYTE, pos++);
					int cmdI = Byte.toUnsignedInt(cmd);
					if (cmdI >= DrawCommand.VALUES.size()) {
						throw new IllegalStateException(STR."Invalid command byte in buffer at \{pos-1}: \{cmdI}");
					}
					switch (DrawCommand.VALUES.get(cmdI)) {
						case PUSH_SCISSOR -> {
							if (this.scissorStack.isEmpty()) GL.glEnable(GLEnableCap.SCISSOR_TEST);
							this.scissorStack.push(pos);
							//LOGGER.log(Level.INFO, "pushed scissor at " + i);
							applyScissor(buf, pos);
							pos += Short.BYTES * 4;
						}
						case POP_SCISSOR -> {
//					LOGGER.log(Level.INFO, "popped scissor at " + buf.readerIndex() + " from " + scissorStack.topInt
//					());
							this.scissorStack.popInt();
							if (this.scissorStack.isEmpty()) {
								Logger.log(() -> STR."disableScissor", DRAW_TAGS);
								GL.glDisable(GLEnableCap.SCISSOR_TEST);
							} else {
								applyScissor(buf, this.scissorStack.topInt());
							}
						}
						case DRAW -> {
							var pos_ = pos;
							LOGGER.log(() -> STR."drawing next layer at \{pos_-1}", TAGS);
							drawer.drawNext();
						}
						case DRAW_PF -> {
							texAllocUnbinder.close();
							texAllocUnbinder = null;
							var id = buf.get(JAVA_INT_UNALIGNED, pos);
							pos += (int) JAVA_INT_UNALIGNED.byteSize();
							var canvas = canvasPool.getUnchecked(id);
//							var _ = canvas.takeScene();
							PathfinderGlobals.sceneProxy.replaceScene(canvas.takeScene());
							PathfinderGlobals.sceneProxy.buildAndRenderGL(PathfinderGlobals.renderer, PathfinderGlobals.options);
							texAllocUnbinder = this.texAlloc.boundForDraw();
							LayeredRenderBuffer.applyGLStateForDrawing();
						}
					}
				}

				if (!this.scissorStack.isEmpty()) {
					throw new IllegalStateException(STR."Drawing completed without popping all scissors: \{this.scissorStack.size()} remain");
				}

				assert !drawer.drawNext() : "Extra layers in drawer";
			}
		} finally {
			if (texAllocUnbinder != null) texAllocUnbinder.close();
		}
	}

	private static void applyScissor(MemorySegment buf, int i) {
		short x = buf.get(JAVA_SHORT_UNALIGNED, i + Short.BYTES * 0);
		short y = buf.get(JAVA_SHORT_UNALIGNED, i + Short.BYTES * 1);
		short width = buf.get(JAVA_SHORT_UNALIGNED, i + Short.BYTES * 2);
		short height = buf.get(JAVA_SHORT_UNALIGNED, i + Short.BYTES * 3);
		Logger.log(() -> STR."applyScissor(\{x}, \{y}, \{width}, \{height})", DRAW_TAGS);
		GL.glScissor(x, y, width, height);
	}

	@Override
	public void close() {
		this.outputBuffer.close();
		this.texAlloc.close();
	}

	private enum UploadCommand {
		PUSH_SCISSOR((int) JAVA_SHORT_UNALIGNED.byteSize() * 4),
		POP_SCISSOR(0),
		UPLOAD(0),
		FLUSH_PF((int) JAVA_INT.byteSize()),
		;

		public static final List<UploadCommand> VALUES = List.of(values());

		public final int sizeWithBytecode;

		UploadCommand(int sizeWithoutBytecode) {
			this.sizeWithBytecode = 1 + sizeWithoutBytecode;
		}

		public final byte asByte() {
			return (byte) this.ordinal();
		}
	}

	private enum DrawCommand {
		PUSH_SCISSOR,
		POP_SCISSOR,
		DRAW,
		DRAW_PF,
		;

		public static final List<DrawCommand> VALUES = List.of(values());

		public final byte asByte() {
			return (byte) this.ordinal();
		}
	}
}
