package dev.pfaff.altargui;

import dev.pfaff.altargui.util.Hasher;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

/**
 * Identifies a widget in the tree. Most often you will want a {@link CallSite unique key for each call-site}, but
 * sometimes you will want a widget to maintain its state as it is moved around the tree. For this, you will need to
 * use a {@link Global Global} key.
 */
public final class Key {
	/**
	 * Together with the parent's key, identifies a widget by the
	 * {@link StateContext#getState(WidgetInstance, CallSite, Widget) getState}
	 * call-site in the tree.
	 * <p>
	 * To create a call-site key, use the syntax {@code () -> {}} as the argument. You <em>may</em> store this in a
	 * variable or a field, but you <strong>must not</strong> use it at more than one
	 * {@link StateContext#getState(WidgetInstance, CallSite, Widget) getState}
	 * call-site.
	 */
	@FunctionalInterface
	public interface CallSite {
		default void hash(Hasher hasher) {
			hasher.trackIdentity(this);
		}

		void __();
	}

	public static final class Local implements CallSite {
		private Local() {}

		@Override
		public void __() {
		}
	}

	public record IntValue(int value) implements CallSite {
		@Override
		public void hash(Hasher hasher) {
			hasher.track(value);
		}

		@Override
		public void __() {
		}
	}

	public record IdentityValue<T>(T value) implements CallSite {
		public static final IdentityValue NULL = new IdentityValue<>(null);

		@Override
		public void hash(Hasher hasher) {
			hasher.trackIdentity(value);
		}

		@Override
		public void __() {
		}
	}

	public record EqualityValue<T>(@Nullable T value) implements CallSite {
		public static final EqualityValue NULL = new EqualityValue<>(null);

		@Override
		public void hash(Hasher hasher) {
			hasher.track(value);
		}

		@Override
		public void __() {
		}
	}

	public record Bi(@NonNull CallSite first, @NonNull CallSite second) implements CallSite {
		@Override
		public void hash(Hasher hasher) {
			this.first.hash(hasher);
			this.second.hash(hasher);
		}

		@Override
		public void __() {
		}
	}

	/**
	 * Global keys, represented by {@link Global}, like all {@code CallSite}s, are compared by identity. Unlike
	 * regular {@code CallSite}s, however, they allow keyed widgets to be moved freely around the tree while
	 * maintaining their state.
	 * <p>
	 * To create a global key, you must call {@link #globalKey()}. The interface is sealed to prevent misuse. Be sure
	 * to store the returned key somewhere, like in the field of a {@link WidgetInstance}, rather than creating it on
	 * every {@link StateContext#getState(WidgetInstance, CallSite, Widget) getState}, lest the widget be completely
	 * recreated every time.
	 * <p>
	 * When the widget is truly unique in the tree, like a singleton, storing a regular {@code CallSite} in a {@code
	 * static final} field would have the same effect. However, this approach makes the developer's intentions less
	 * clear, and as such is not recommended. Additionally, code using this approach would suddenly break if the widget
	 * were ever changed to no longer be a singleton.
	 */
	public static final class Global implements CallSite {
		WidgetInstance element;

		private Global() {}

		@Override
		public void __() {
		}
	}

	/**
	 * Creates a new {@link Global Global} key.
	 */
	public static Global globalKey() {
		return new Global();
	}

	/**
	 * Creates a new {@link Local Local} key.
	 */
	public static Local localKey() {
		return new Local();
	}

	/**
	 * Creates a new {@link EqualityValue EqualityValue} key.
	 */
	public static <T> EqualityValue<T> valueKey(T value) {
		// Intellij's analyzer doesn't understand that T could be nullable.
		//noinspection ConstantValue
		if (value == null) //noinspection unchecked
			return EqualityValue.NULL;
		return new EqualityValue<>(value);
	}

	/**
	 * Creates a new {@link IdentityValue IdentityValue} key.
	 */
	public static <T> IdentityValue<T> identityKey(T value) {
		//noinspection ConstantValue
		if (value == null) //noinspection unchecked
			return IdentityValue.NULL;
		return new IdentityValue<>(value);
	}

	/**
	 * Creates a new {@link EqualityValue EqualityValue} key.
	 */
	public static IntValue intKey(int value) {
		return new IntValue(value);
	}

	/**
	 * Creates a new {@link Bi Bi} key.
	 */
	public static Bi biKey(CallSite first, CallSite second) {
		return new Bi(first, second);
	}
}
