package dev.pfaff.altargui;

import dev.pfaff.altargui.types.Alignment;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.widgets.*;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.Function;

import static dev.pfaff.altargui.widgets.MultiChildWidget.diagonal;
import static dev.pfaff.altargui.widgets.MultiChildWidget.stack;

/**
 * Inspired by Flutter's approach: Widgets are essentially factories for configured elements (subclasses of
 * {@link WidgetInstance}).
 */
public interface Widget {
	Widget[] EMPTY_ARRAY = new Widget[0];

	/**
	 * Instantiates an instance of the widget.
	 */
	@SuppressWarnings("rawtypes")
	@NotNull WidgetInstance instantiate(WidgetInstantiationArgs args);

	default SizedWidget.Absolute withAbsoluteSize(float x, float y) {
		return new SizedWidget.Absolute(x, y, this);
	}

	default SizedWidget.AbsoluteWidth withAbsoluteWidth(float width) {
		return new SizedWidget.AbsoluteWidth(width, this);
	}

	default SizedWidget.AbsoluteHeight withAbsoluteHeight(float height) {
		return new SizedWidget.AbsoluteHeight(height, this);
	}

	default SizedWidget.Expand expand() {
		if (this instanceof SizedWidget.Expand(var child, _, _)) {
			return new SizedWidget.Expand(child, true, true);
		}
		return new SizedWidget.Expand(this, true, true);
	}

	default SizedWidget.Expand expandWidth() {
		if (this instanceof SizedWidget.Expand(var child, _, var y)) {
			return new SizedWidget.Expand(child, true, y);
		}
		return new SizedWidget.Expand(this, true, false);
	}

	default SizedWidget.Expand expandHeight() {
		if (this instanceof SizedWidget.Expand(var child, var x, _)) {
			return new SizedWidget.Expand(child, x, true);
		}
		return new SizedWidget.Expand(this, false, true);
	}

	default MultiChildAttributesWidget withMultiChildAttributes() {
		return new MultiChildAttributesWidget(this);
	}

	default AspectRatioWidget withAspectRatio(double aspectRatio, Alignment horizontalAlignment, Alignment verticalAlignment) {
		return new AspectRatioWidget(this, aspectRatio, horizontalAlignment, verticalAlignment);
	}

	default AspectRatioWidget withAspectRatio(double aspectRatio, Alignment alignment) {
		return this.withAspectRatio(aspectRatio, alignment, alignment);
	}

	default AspectRatioWidget withAspectRatio(double aspectRatio) {
		return this.withAspectRatio(aspectRatio, Alignment.Start);
	}

	default PositionedWidget.Absolute atAbsolutePosition(float x, float y) {
		return new PositionedWidget.Absolute(x, y, this);
	}

	default PositionedWidget.Relative atRelativePosition(float x, float y) {
		return new PositionedWidget.Relative(x, y, this);
	}

	default PaddedWidget withMargin(float startX, float startY, float endX, float endY) {
		return new PaddedWidget(startX, startY, endX, endY, this);
	}

	default PaddedWidget withMargin(float all) {
		return withMargin(all, all, all, all);
	}

	default PaddedWidget withSymmetricMargin(float horizontal, float vertical) {
		return withMargin(horizontal, vertical, horizontal, vertical);
	}

	default PaddedWidget withHorizontalMargin(float start, float end) {
		return withMargin(start, 0, end, 0);
	}

	default PaddedWidget withVerticalMargin(float start, float end) {
		return withMargin(0, start, 0, end);
	}

	default PaddedWidget withHorizontalMargin(float both) {
		return withHorizontalMargin(both, both);
	}

	default PaddedWidget withVerticalMargin(float both) {
		return withVerticalMargin(both, both);
	}

	default PaddedWidget withMarginTop(float value) {
		return withMargin(0, value, 0, 0);
	}

	default PaddedWidget withMarginBottom(float value) {
		return withMargin(0, 0, 0, value);
	}

	default PaddedWidget withMarginStart(float value) {
		return withMargin(value, 0, 0, 0);
	}

	default PaddedWidget withMarginEnd(float value) {
		return withMargin(0, 0, value, 0);
	}

	default PositionedWidget.Aligned atCenter() {
		return atAlignment(Alignment.Center, Alignment.Center);
	}

	default PositionedWidget.Aligned atAlignment(Alignment horizontal, Alignment vertical) {
		return new PositionedWidget.Aligned(this, horizontal, vertical);
	}

	default BorderWidget withBorder(Color color, float thickness) {
		return new BorderWidget(color, thickness, this);
	}

	default Widget withBackground(Widget background) {
		return stack(List.of(background.withMultiChildAttributes().tighten(), this));
	}

	default Widget withBackgroundColor(Color color) {
		return withBackground(RectWidget.colored(color));
	}

	default Widget withBackgroundTexture(LoadableTex2d tex) {
		return withBackground(RectWidget.textured(tex));
	}

	default Widget withTooltip(Widget content) {
		var tooltip = new LayerWidget(content
				.withMargin(4)
				.withBackground(Widgets.TOOLTIP_BACKGROUND), Layer.Tooltip);
		return new MouseOverWithCoordsBuilder((hovered, mouseX, mouseY) -> diagonal(mounter -> {
			mounter.layout(() -> {}, this);
			if (hovered) {
				// I'm leaving this one as a mounter because the mouseX and mouseY are going to be changing
				// frequently, so using a list will provide little if any benefit.
				mounter.layout(() -> {}, tooltip.atAbsolutePosition(mouseX+6, mouseY+6));
			}
		}));
	}

	default Widget withId(String id) {
		return new IdWidget(id, this);
	}

	default Widget withKey(Key.CallSite key) {
		return new KeyedWidget(key, this);
	}

	default <W extends Widget> W wrap(Function<Widget, W> builder) {
		return builder.apply(this);
	}
}
