package dev.pfaff.altargui;

import dev.pfaff.altargui.widgets.state.HasDebugLayer;
import dev.pfaff.altargui.widgets.state.Painter;
import dev.pfaff.altargui.widgets.state.PostPainter;
import dev.pfaff.altargui.handlers.MouseDragHandler;
import dev.pfaff.altargui.handlers.MouseMoveHandler;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.inspector.InspectorMeta;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.elements.IElement;
import dev.pfaff.altargui.internal.elements.RootElement;
import dev.pfaff.altargui.internal.util.Specialize;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import dev.pfaff.altargui.render.objects.DebugRectRenderObject;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.util.CoordUtil;
import dev.pfaff.altargui.util.Preconditions;
import dev.pfaff.altargui.util.listenable.BooleanSignal;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.widgets.IdWidget;
import org.jetbrains.annotations.MustBeInvokedByOverriders;
import org.jetbrains.annotations.Nullable;

import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.GuiGlobals.profiler;
import static dev.pfaff.altargui.util.ClipUtil.clipExtent;
import static dev.pfaff.log4truth.StandardTags.WARN;
import static java.util.FormatProcessor.FMT;

public non-sealed abstract class WidgetInstance<W extends Widget> implements IElement {
	public static final WidgetInstance[] EMPTY_ARRAY = new WidgetInstance[0];

	/**
	 * The cached key of the element's widget. This <strong>must</strong> be the first field in the class for offset
	 * compatibility with {@link RootElement}.
	 */
	final long key;
	/**
	 * Only non-null if {@link Config#SANITIZE_KEYS} is {@code true}.
	 */
	final Key.@Nullable CallSite keyObject;
	final GuiContext guiContext;
	@InspectorMeta(tooltip = "The class of the element's Widget")
	final Class<? extends Widget> widgetClass;
	@InspectorMeta(tooltip = "The unique identifier of this element")
	boolean generation;
	// @Stable
	boolean initialized;

	@Nullable
	private IElement parent;

	final @Nullable DebugLayer debugLayer = this instanceof HasDebugLayer hasDebugLayer
			? hasDebugLayer.debugLayer()
			: null;
	final @Nullable DebugRectRenderObject debugRect = this instanceof HasDebugLayer hasDebugLayer
			? new DebugRectRenderObject()
			: null;

	final @Nullable MouseDragHandler mouseDragHandler = this instanceof MouseDragHandler handler ? handler : null;
	final @Nullable MouseMoveHandler mouseMoveHandler = this instanceof MouseMoveHandler handler ? handler : null;
	final @Nullable MouseScrollHandler mouseScrollHandler = this instanceof MouseScrollHandler handler ? handler : null;
	final @Nullable Painter painter = this instanceof Painter painter ? painter : null;
	final @Nullable PostPainter postPainter = this instanceof PostPainter painter ? painter : null;

	protected WidgetInstance(WidgetInstantiationArgs args) {
		this.guiContext = args.context;
		this.widgetClass = args.widgetClass;
		this.key = args.key;
		this.keyObject = args.keyObject;
	}

	@InspectorMeta(tooltip = "The allocated width of the widget")
	private float allocatedWidth;
	@InspectorMeta(tooltip = "The allocated height of the widget")
	private float allocatedHeight;
	@InspectorMeta(tooltip = "Whether the widget is permitted by the parent to overflow the X-axis")
	private boolean canOverflowX;
	@InspectorMeta(tooltip = "Whether the widget is permitted by the parent to overflow the Y-axis")
	private boolean canOverflowY;
	@InspectorMeta(tooltip = "The X-axis position of the widget")
	private float allocatedX;
	@InspectorMeta(tooltip = "The Y-axis position of the widget")
	private float allocatedY;
	@InspectorMeta(tooltip = "The computed width of the widget")
	private float computedWidth;
	@InspectorMeta(tooltip = "The computed height of the widget")
	private float computedHeight;
	@InspectorMeta(tooltip = "The clipped X-axis position of the widget")
	private float clippedX;
	@InspectorMeta(tooltip = "The clipped Y-axis position of the widget")
	private float clippedY;
	@InspectorMeta(tooltip = "The clipped width of the widget")
	private float clippedWidth;
	@InspectorMeta(tooltip = "The clipped height of the widget")
	private float clippedHeight;

	// @Stable
	private boolean hasStencilLaidOut = false;
	// @Stable
	private boolean hasTraceLaidOut = false;
	// @Stable
	private boolean hasScissorLaidOut = false;
	private boolean isStencilLayingOut = false;
	private boolean isTraceLayingOut = false;
	private boolean isScissorLayingOut = false;
	public final BooleanSignal needsStencilLayout = new BooleanSignal(true);
	public final BooleanSignal needsTraceLayout = new BooleanSignal(true);
	public final BooleanSignal needsScissorLayout = new BooleanSignal(true);
	private @Nullable ListenerDisposer needsStencilLayoutParentNotifier;
	private @Nullable ListenerDisposer needsTraceLayoutParentNotifier;
	private @Nullable ListenerDisposer needsScissorLayoutParentNotifier;

	/**
	 * Configures the instance's state according to the inputs in the specified {@code widget}.
	 * <p>
	 * Other initialization might be done if {@code init} is true.
	 */
	// TODO: explore whether moving this to the Widget interface would improve performance.
	public abstract void configure(StateContext stateContext, W widget, boolean init);

	/**
	 * Called when the element is mounted to a new parent.
	 */
	protected void onMount() {
		if (Config.CHECKS && this.parent == null) throw new IllegalStateException("Not mounted");
		// add listener to propagate needs layout upwards
		if (parent instanceof WidgetInstance<?> parentWidgetInstance) {
			needsStencilLayoutParentNotifier = needsStencilLayout.listenAndInvoke(parentWidgetInstance::markNeedsStencilLayout);
			needsTraceLayoutParentNotifier = needsTraceLayout.listenAndInvoke(parentWidgetInstance::markNeedsTraceLayout);
			needsScissorLayoutParentNotifier = needsScissorLayout.listenAndInvoke(parentWidgetInstance::markNeedsScissorLayout);
		}
	}

	/**
	 * Called when the element is unmounted from the current parent.
	 */
	@MustBeInvokedByOverriders
	protected void onUnmount() {
		if (Config.CHECKS && this.parent == null) throw new IllegalStateException("Not mounted");
		var needsStencilLayoutParentNotifier = this.needsStencilLayoutParentNotifier;
		if (needsStencilLayoutParentNotifier != null) {
			needsStencilLayoutParentNotifier.close();
			this.needsStencilLayoutParentNotifier = null;
		}
		var needsTraceLayoutParentNotifier = this.needsTraceLayoutParentNotifier;
		if (needsTraceLayoutParentNotifier != null) {
			needsTraceLayoutParentNotifier.close();
			this.needsTraceLayoutParentNotifier = null;
		}
		var needsScissorLayoutParentNotifier = this.needsScissorLayoutParentNotifier;
		if (needsScissorLayoutParentNotifier != null) {
			needsScissorLayoutParentNotifier.close();
			this.needsScissorLayoutParentNotifier = null;
		}
	}

	final void updateMount(IElement parent) {
		var profiler = profiler();
		profiler.push("updateMount");
		if (this.parent != parent) {
			if (this.parent != null) {
				try {
					this.onUnmount();
				} catch (Throwable e) {
					throw GuiException.wrap(FMT."Unmounting \{toStringShort()} from \{parent.toStringShort()}", e);
				}
			}
			this.parent = parent;
			this.guiContext.ensureElementIsManagedOnMount(this);
			try {
				this.onMount();
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Mounting \{toStringShort()} to \{parent.toStringShort()}", e);
			}
		}
		profiler.pop();
	}

	public final void updateMount(WidgetInstance<?> parent) {
		updateMount((IElement) parent);
	}

	/**
	 * Disposes of the element. The element may be reused.
	 */
	public final void dispose() {
		if (Config.CHECKS && this.parent == null) throw new IllegalStateException("Already unmounted");
		onUnmount();
		this.parent = null;
	}

	/**
	 * Computes sizes of element content and children.
	 * <p>
	 * This will usually call
	 * {@link StencilLayoutContext#updateMountAndLayout(StateContext, WidgetInstance, WidgetInstance, BiConsumer)} with
	 * each child and go through a layout procedure wherein {@link WidgetInstance#setAvailableSize(float, float)} and
	 * then {@link WidgetInstance#stencilLayout(StateContext)} are called (in that order) on each child.
	 */
	// TODO: consider using ScopedValue for passing around StateContext
	public final void stencilLayout(StateContext context) {
		try {
			if (Config.CHECKS && (isStencilLayingOut || isTraceLayingOut || isScissorLayingOut)) {
				throw new IllegalStateException("Already laying out: " + this);
			}
			if (!takeNeedsStencilLayout()) {
				if (Config.CHECKS && !hasStencilLaidOut) {
					throw new IllegalStateException("Not marked for layout, but never been laid out: " + this);
				}
				return;
			}
			if (Config.CHECKS) {
				isStencilLayingOut = true;
			}
	//		LOGGER.log(Level.INFO, FMT."Took needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
			var profiler = profiler();
			profiler.push("computeStencilLayout");
			computeStencilLayout(context);
			profiler.pop();
			if (Config.CHECKS) {
				if (!hasStencilLaidOut) hasStencilLaidOut = true;
				isStencilLayingOut = false;
			}
		} catch (Throwable e) {
			throw GuiException.wrap(STR."Stencil laying out \{toStringShort()}", e);
		}
	}

	/**
	 * Sets positions of element content and children.
	 * <p>
	 * This will usually call
	 * {@link StencilLayoutContext#updateMountAndLayout(StateContext, WidgetInstance, WidgetInstance, BiConsumer)} with
	 * each child and go through a layout procedure wherein {@link WidgetInstance#setAvailableSize(float, float)} and
	 * then {@link WidgetInstance#traceLayout()} are called (in that order) on each child.
	 */
	public final void traceLayout() {
		try {
			if (Config.CHECKS && (isTraceLayingOut || isStencilLayingOut || isScissorLayingOut)) {
				throw new IllegalStateException("Already laying out: " + this);
			}
			if (!takeNeedsTraceLayout()) {
				if (Config.CHECKS && !hasTraceLaidOut) {
					throw new IllegalStateException("Not marked for layout, but never been laid out: " + this);
				}
				return;
			}
			if (Config.CHECKS) {
				isTraceLayingOut = true;
			}
	//		LOGGER.log(Level.INFO, FMT."Took needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
			computeTraceLayout();
			if (Config.CHECKS) {
				if (!hasTraceLaidOut) hasTraceLaidOut = true;
				isTraceLayingOut = false;
			}
		} catch (Throwable e) {
			throw GuiException.wrap(STR."Trace laying out \{toStringShort()}", e);
		}
	}

	/**
	 * Clips element content and children.
	 * <p>
	 * This will usually call
	 * {@link StencilLayoutContext#updateMountAndLayout(StateContext, WidgetInstance, WidgetInstance, BiConsumer)} with
	 * each child and go through a layout procedure wherein
	 * {@link WidgetInstance#setClippedRect(float, float, float, float)} and then {@link WidgetInstance#scissorLayout()}
	 * are called (in that order) on each child.
	 */
	public final void scissorLayout() {
		try {
			if (Config.CHECKS && (isTraceLayingOut || isStencilLayingOut || isScissorLayingOut)) {
				throw new IllegalStateException("Already laying out: " + this);
			}
			if (!takeNeedsScissorLayout()) {
				if (Config.CHECKS && !hasScissorLaidOut) {
					throw new IllegalStateException("Not marked for layout, but never been laid out: " + this);
				}
				return;
			}
			if (Config.CHECKS) {
				isScissorLayingOut = true;
			}
			//		LOGGER.log(Level.INFO, FMT."Took needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
			computeScissorLayout();
			if (Config.CHECKS) {
				if (!hasScissorLaidOut) hasScissorLaidOut = true;
				isScissorLayingOut = false;
			}
		} catch (Throwable e) {
			throw GuiException.wrap(STR."Scissor laying out \{toStringShort()}", e);
		}
	}

	public final void mount(MountContext context) {
		if (Config.CHECKS && (!hasStencilLaidOut || !hasTraceLaidOut || !hasScissorLaidOut)) {
			throw new IllegalStateException("Attempted to mount an element that has never been laid out: " + this);
		}
		computeMount(context);
	}

	@Specialize
	public void uSpecMount(MountContext context) {
		context.iSpecMount(this);
	}

	protected static void computeFillLayout(WidgetInstance element) {
		element.requireFiniteAllocatedWidth();
		element.requireFiniteAllocatedHeight();
		element.fillAllocated();
	}

	protected abstract void computeStencilLayout(StateContext context);

	protected abstract void computeTraceLayout();

	protected abstract void computeScissorLayout();

	protected abstract void computeMount(MountContext context);

	protected final void markNeedsStencilLayout() {
//		if (needsLayout.value()) {
//			LOGGER.log(Level.INFO, FMT."Marked needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
//		}
		needsStencilLayout.set();
	}

	protected final void markNeedsTraceLayout() {
//		if (needsLayout.value()) {
//			LOGGER.log(Level.INFO, FMT."Marked needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
//		}
		needsTraceLayout.set();
	}

	protected final void markNeedsScissorLayout() {
//		if (needsLayout.value()) {
//			LOGGER.log(Level.INFO, FMT."Marked needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
//		}
		needsScissorLayout.set();
	}

	/**
	 * @deprecated use {@link #markNeedsStencilLayout()} and/or {@link #markNeedsTraceLayout()} as appropriate instead.
	 */
	@Deprecated
	protected final void markNeedsLayout() {
		markNeedsStencilLayout();
		markNeedsTraceLayout();
		markNeedsScissorLayout();
	}

	/**
	 * This will need to be overridden for any aligned positioning to include {@link #markNeedsTraceLayout()}.
	 */
	protected void propagateNeedsStencilLayoutFromChild() {
		markNeedsStencilLayout();
	}

	protected final boolean takeNeedsStencilLayout() {
		return needsStencilLayout.clearNoNotify();
	}

	protected final boolean takeNeedsTraceLayout() {
		return needsTraceLayout.clearNoNotify();
	}

	protected final boolean takeNeedsScissorLayout() {
		return needsScissorLayout.clearNoNotify();
	}

	public final float allocatedX() {
		if (Config.CHECKS && isStencilLayingOut) {
			throw new IllegalStateException("Attempted to access allocatedX while stencil laying out: " + toStringShort());
		}
		return allocatedX;
	}

	public final float allocatedY() {
		if (Config.CHECKS && isStencilLayingOut) {
			throw new IllegalStateException("Attempted to access allocatedY while stencil laying out: " + toStringShort());
		}
		return allocatedY;
	}

	public final float allocatedWidth() {
		return allocatedWidth;
	}

	public final float allocatedHeight() {
		return allocatedHeight;
	}

	protected final void markNeedsStencilLayoutOnAllocationChange() {
		needsStencilLayout.setNoNotify();
	}

	protected final void markNeedsTraceLayoutOnAllocationChange() {
		needsTraceLayout.setNoNotify();
	}

	protected final void markNeedsScissorLayoutOnAllocationChange() {
		needsScissorLayout.setNoNotify();
	}

	/**
	 * Called by the parent element to set the position of its children.
	 */
	public final void setPosition(float x, float y) {
		if (this.allocatedX != x || this.allocatedY != y) {
			checkFinite(x, "allocatedX");
			checkFinite(y, "allocatedY");
			this.allocatedX = x;
			this.allocatedY = y;
			markNeedsTraceLayoutOnAllocationChange();
			// TODO: move out of here
			markNeedsScissorLayoutOnAllocationChange();
		}
	}

	/**
	 * Called by the parent element to communicate the amount of space available to its children.
	 */
	public final void setAvailableSize(float width, float height) {
		if (this.allocatedWidth != width || this.allocatedHeight != height) {
			checkFinite(width, "allocatedWidth");
			checkFinite(height, "allocatedHeight");
			this.allocatedWidth = Math.max(width, 0f);
			this.allocatedHeight = Math.max(height, 0f);
			markNeedsStencilLayoutOnAllocationChange();
		}
	}

	/**
	 * Copies the {@linkplain #setPosition(float, float) position} from the given {@code other} widget (usually the
	 * parent).
	 */
	public final void copyPositionFrom(WidgetInstance other) {
		setPosition(other.allocatedX, other.allocatedY);
	}

	/**
	 * Copies the {@linkplain #setAvailableSize(float, float) available size} from the given {@code other} widget
	 * (usually the parent).
	 */
	public final void copyAvailableSizeFrom(WidgetInstance other) {
		setAvailableSize(other.allocatedWidth, other.allocatedHeight);
	}

	/**
	 * Copies the {@linkplain #setClippedRect(float, float, float, float) clipped rect} from the given {@code other}
	 * widget (usually the parent).
	 */
	public final void copyClippedRectFrom(WidgetInstance other) {
		var clippedWidth = clipExtent(other.clippedX,
				other.clippedWidth,
				this.allocatedX,
				this.computedWidth);
		var clippedHeight = clipExtent(other.clippedY,
				other.clippedHeight,
				this.allocatedY,
				this.computedHeight);
		setClippedRect(clippedWidth.start(), clippedHeight.start(), clippedWidth.extent(), clippedHeight.extent());
	}

	protected final void requireFiniteAllocatedWidth() {
		if (allocatedWidth() == Float.POSITIVE_INFINITY) {
			throw new GuiException(getClass().getSimpleName() + " cannot be used with infinite allocated width");
		}
	}

	protected final void requireFiniteAllocatedHeight() {
		if (allocatedHeight() == Float.POSITIVE_INFINITY) {
			throw new GuiException(getClass().getSimpleName() + " cannot be used with infinite allocated height");
		}
	}

	public final float computedWidth() {
		return computedWidth;
	}

	public final float computedHeight() {
		return computedHeight;
	}

	public final float clippedX() {
		return clippedX;
	}

	public final float clippedY() {
		return clippedY;
	}

	public final float clippedWidth() {
		return clippedWidth;
	}

	public final float clippedHeight() {
		return clippedHeight;
	}

	protected final void setComputedSize(float width, float height) {
		if (Config.CHECKS && !isStencilLayingOut) {
			throw new IllegalStateException("Not stencil laying out: " + this.toStringShort());
		}
		if (computedWidth != width || computedHeight != height) {
			checkFinite(width, "computedWidth");
			checkPositive(width, "computedWidth");
			checkFinite(height, "computedHeight");
			checkPositive(height, "computedHeight");
			computedWidth = width;
			computedHeight = height;
			// TODO: move out of here
			markNeedsScissorLayoutOnAllocationChange();
		}
	}

	/**
	 * <p>
	 * Widgets may be clipped by parents. To support this, parents must propagate a clipped rect to their children. The
	 * clip is enforced by parents when rendering, but must be enforced by each element when implementing input event
	 * handlers.
	 *
	 * @see #clippedX
	 * @see #clippedY
	 * @see #clippedWidth
	 * @see #clippedHeight
	 */
	public final void setClippedRect(float x, float y, float width, float height) {
		if (clippedX != x || clippedY != y || clippedWidth != width || clippedHeight != height) {
			checkFinite(x, "clippedX");
			checkPositive(x, "clippedX");
			checkFinite(y, "clippedY");
			checkPositive(y, "clippedY");
			checkFinite(width, "clippedWidth");
			checkPositive(width, "clippedWidth");
			checkFinite(height, "clippedHeight");
			checkPositive(height, "clippedHeight");
			clippedX = x;
			clippedY = y;
			clippedWidth = width;
			clippedHeight = height;
			markNeedsScissorLayoutOnAllocationChange();
		}
	}

	private void checkFieldsEqual(float ours, float theirs, String name) {
		if (Config.CHECKS) {
			Preconditions.checkArgument(theirs == ours,
										() -> STR."theirs.\{name} (\{theirs}) != ours.\{name} (\{ours})");
		}
	}

	protected final void copyComputedSizeFrom(WidgetInstance other) {
		setComputedSize(other.computedWidth(), other.computedHeight());
	}

	/**
	 * Sets the computed size to fill the entire allocated rect.
	 */
	protected final void fillAllocated() {
		setComputedSize(allocatedWidth(), allocatedHeight());
	}

	/**
	 * Checks if the given coordinates are inside the element's clipped rectangle.
	 */
	public final boolean isInside(double checkX, double checkY) {
		return isInside((float) checkX, (float) checkY);
	}

	/**
	 * Checks if the given coordinates are inside the element's clipped rectangle.
	 */
	public final boolean isInside(float checkX, float checkY) {
		return CoordUtil.isInside(this.clippedX,
				this.clippedY,
				this.clippedWidth,
				this.clippedHeight,
				checkX,
				checkY);
	}

	public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
		LOGGER.log(() -> STR."attempted to apply an unsupported builtin selector (\{selector}) on \{this}", List.of(WARN));
	}

	public final String toStringShort() {
		if (this instanceof IdWidget.State idWidget) {
			return FMT."\{this.widgetClass.getName()}@%016x\{this.key} #\{idWidget.id()}";
		}
		return FMT."\{this.widgetClass.getName()}@%016x\{this.key}";
	}

	@Override
	public final String toString() {
		return FMT."""
		WidgetInstance[
		  class=\{this.widgetClass.getName()}
		  key=%016x\{this.key}
		  generation=\{this.generation}
		  initialized=\{this.initialized}
		]""";
	}

	private static void checkFinite(float f, String name) {
		if (!Float.isFinite(f)) {
			throw new IllegalArgumentException(STR."Expected a finite value for argument '\{name}', found \{f}");
		}
	}

	private static void checkPositive(float f, String name) {
		if (f < 0) {
			throw new IllegalArgumentException(STR."Expected a positive value for argument '\{name}', found \{f}");
		}
	}

	public final GuiContext guiContext() {
		return guiContext;
	}

	@Override
	public final long key() {
		return key;
	}

	@Override
	public final boolean equals(Object o) {
		return super.equals(o);
	}

	@Override
	public final int hashCode() {
		return super.hashCode();
	}

	@Nullable
	public final IElement parent() {
		return this.parent;
	}
}
