package dev.pfaff.altargui;

import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.util.WidgetInstanceConsumer;
import dev.pfaff.unfettered.UnsafeWrapper;
import it.unimi.dsi.fastutil.objects.ObjectIterators;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Spliterator;
import java.util.function.Function;
import java.util.function.Predicate;

import static dev.pfaff.unfettered.UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET;
import static dev.pfaff.unfettered.UnsafeWrapper.ARRAY_OBJECT_INDEX_SCALE;

public final class WidgetInstanceList implements List<WidgetInstance<?>> {
	private WidgetInstance<?>[] elements;
	private int length;

	public WidgetInstanceList(int initialCapacity) {
		this.elements = new WidgetInstance<?>[initialCapacity];
	}

	public WidgetInstance<?>[] elements() {
		return elements;
	}

	public int size() {
		return length;
	}

	@Override
	public boolean isEmpty() {
		return length == 0;
	}

	@Contract(pure = true)
	@Override
	public boolean contains(Object o) {
		return indexOf(o) != -1;
	}

	@NotNull
	@Override
	public Iterator<WidgetInstance<?>> iterator() {
		return listIterator();
	}

	@Override
	public ListIterator<WidgetInstance<?>> listIterator() {
		return ObjectIterators.wrap(elements, 0, length);
	}

	@Override
	public ListIterator<WidgetInstance<?>> listIterator(int index) {
		return ObjectIterators.wrap(elements, index, length - index);
	}

	@Override
	public Spliterator<WidgetInstance<?>> spliterator() {
		return Arrays.spliterator(elements, 0, length);
	}

	@Override
	public WidgetInstance[] toArray() {
		return Arrays.copyOf(elements, length);
	}

	@Override
	public <T> T[] toArray(@NonNull T[] a) {
		//noinspection unchecked
		return (T[]) Arrays.copyOf(elements, length);
	}

	public void trim() {
		Arrays.fill(elements, length, elements.length, null);
	}

	public void clear() {
		length = 0;
	}

	@Override
	public WidgetInstance<?> get(int index) {
		if (index >= length) throw new IndexOutOfBoundsException(index);
		return elements[index];
	}

	@Override
	public WidgetInstance<?> set(int index, WidgetInstance<?> element) {
		if (index >= length) throw new IndexOutOfBoundsException(index);
		var old = elements[index];
		elements[index] = element;
		return old;
	}

	@Override
	public void add(int index, WidgetInstance<?> element) {
		throw new UnsupportedOperationException();
	}

	@Override
	public WidgetInstance<?> remove(int index) {
		if (index >= length) throw new IndexOutOfBoundsException(index);
		var value = elements[index];
		System.arraycopy(elements, index+1, elements, index, elements.length - index - 1);
		return value;
	}

	@Override
	public void addFirst(WidgetInstance<?> widgetInstance) {
		List.super.addFirst(widgetInstance);
	}

	@Override
	public void addLast(WidgetInstance<?> widgetInstance) {
		List.super.addLast(widgetInstance);
	}

	@Override
	public WidgetInstance<?> getFirst() {
		return List.super.getFirst();
	}

	@Override
	public WidgetInstance<?> getLast() {
		return List.super.getLast();
	}

	@Override
	public WidgetInstance<?> removeFirst() {
		return List.super.removeFirst();
	}

	@Override
	public WidgetInstance<?> removeLast() {
		return List.super.removeLast();
	}

	@Override
	public int indexOf(Object o) {
		var elements = this.elements;
		var upr = length;
		if (upr > elements.length) throw new AssertionError();
		for (int i = 0; i < upr; i++) if (elements[i] == o) return i;
		return -1;
	}

	@Override
	public int lastIndexOf(Object o) {
		var elements = this.elements;
		var i = length;
		if (i > elements.length) throw new AssertionError();
		while (i > 0) if (elements[--i] == o) return i;
		return -1;
	}

	@Override
	public List<WidgetInstance<?>> subList(int fromIndex, int toIndex) {
		throw new UnsupportedOperationException();
	}

	private WidgetInstance<?>[] grow(WidgetInstance<?>[] elements) {
		return grow(elements, elements.length * 2);
	}

	private WidgetInstance<?>[] grow(WidgetInstance<?>[] elements, int minCapacity) {
		var newCapacity = elements.length;
		while (newCapacity < minCapacity) newCapacity *= 2;
		return this.elements = Arrays.copyOf(elements, newCapacity);
	}

	public WidgetInstance<?>[] ensureCapacityAndReturnArray(int capacity) {
		var elements = this.elements;
		if (capacity > elements.length) {
			elements = grow(elements, capacity);
		}
		return elements;
	}

	@Override
	public boolean add(WidgetInstance<?> element) {
		int length = this.length;
		var elements = this.elements;
		if (length >= elements.length) {
			elements = grow(elements);
		}
		if (Config.UNSAFE_OPTIMIZATIONS) {
			UnsafeWrapper.putReference(elements, ARRAY_OBJECT_BASE_OFFSET + (long) length * ARRAY_OBJECT_INDEX_SCALE, element);
		} else {
			elements[length] = element;
		}
		this.length = length + 1;
		return true;
	}

	@Override
	public boolean remove(@NonNull Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		for (var o : c) {
			if (!contains(o)) return false;
		}
		return true;
	}

	@Override
	public boolean addAll(Collection<? extends WidgetInstance<?>> c) {
		for (var element : c) add(element);
		return true;
	}

	@Override
	public boolean addAll(int index, Collection<? extends WidgetInstance<?>> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	public void forEach(WidgetInstanceConsumer consumer) {
		var elements = this.elements;
		var upr = length;
		if (upr > elements.length) throw new AssertionError();
		for (int i = 0; i < upr; ) consumer.accept(elements[i++]);
	}

	public void setSizeAssumeCapacityAndInitialized(int size) {
		assert size <= elements.length;
		length = size;
	}

	/**
	 * Top-down element iteration.
	 */
	public boolean iterateDown(Predicate<WidgetInstance<?>> consumer) {
		var a = this.elements();
		var upr = this.size();
		if (upr > a.length) throw new AssertionError();
		for (int i = upr; i > 0;) {
			var element = a[--i];
			if (consumer.test(element)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Top-down element iteration.
	 */
	public <T> boolean iterateDown(Function<WidgetInstance<?>, @Nullable T> filter, Predicate<T> consumer) {
		var a = this.elements();
		var upr = this.size();
		if (upr > a.length) throw new AssertionError();
		for (int i = upr; i > 0;) {
			// Profiling makes one thing exceedingly clear: null-checking a field whose type is an interface and then
			// calling a function on that interface is **much** faster than instanceof-checking an object
			var element = filter.apply(a[--i]);
			if (element != null && consumer.test(element)) {
				return true;
			}
		}
		return false;
	}

	@Deprecated
	public <T> boolean iterateDown(Class<T> clazz, Predicate<T> consumer) {
		return iterateDown(element -> clazz.isInstance(element) ? (T)element : null, consumer);
	}
}
