package dev.pfaff.altargui.improv;

import dev.pfaff.altargui.ffi.*;
import dev.pfaff.altargui.types.Alignment;

import static dev.pfaff.altargui.ffi.Strength.MEDIUM;
import static dev.pfaff.altargui.ffi.Strength.REQUIRED;

public final class Constraints {
	/**
	 * Returns the constraints that will ensure that the {@code object} is contained within the {@code container}.
	 */
	public static Constraint[] fits(IRect<? extends ToExpression> object, IRect<? extends ToExpression> container, boolean allowOverflow) {
		var s = allowOverflow ? MEDIUM : REQUIRED;
		return new Constraint[]{
				object.x().ge(container.x())
						.withStrength(s),
				object.y().ge(container.y())
						.withStrength(s),
				object.x().minus(container.x())
						.plus(object.width())
						.le(container.width())
						.withStrength(s),
				object.y().minus(container.y())
						.plus(object.height())
						.le(container.height())
						.withStrength(s),
		};
	}

	/**
	 * Returns the expression that will calculate the top-left position of the {@code object} such that it is aligned
	 * according to {@code alignment} along the {@code container}.
	 */
	public static ToExpression alignedTopLeftPosition(Line<? extends ToExpression> object, Line<? extends ToExpression> container, Alignment alignment) {
		return switch (alignment) {
			case Start -> container.position();
			case Center -> container.position()
					.plus(container.length().dividedBy(2.0))
					.minus(object.length().dividedBy(2.0));
			case End -> container.position().plus(container.length()).minus(container.length());
		};
	}

	public static Constraint align(Line<? extends ToExpression> object, Line<? extends ToExpression> container, Alignment alignment) {
		return object.position().eq(alignedTopLeftPosition(object, container, alignment));
	}

	public static Constraint[] align(IRect<? extends ToExpression> object, IRect<? extends ToExpression> container, Alignment xAlignment, Alignment yAlignment) {
		return new Constraint[]{
				align(new Line<>(object.x(), object.width()),
						new Line<>(container.x(), container.width()),
						xAlignment),
				align(new Line<>(object.y(), object.height()),
						new Line<>(container.y(), container.height()),
						yAlignment),
		};
	}

	public static Constraint[] eqlPos(VarRect viewport, double x, double y) {
		return new Constraint[]{viewport.x().eq(x), viewport.y().eq(y)};
	}

	public static Constraint[] eqlPos(VarRect viewport, ToExpression x, ToExpression y) {
		return new Constraint[]{viewport.x().eq(x), viewport.y().eq(y)};
	}

	public static Constraint[] eqlSize(VarRect viewport, double width, double height) {
		return new Constraint[]{viewport.width().eq(width), viewport.height().eq(height)};
	}

	public static Constraint[] eqlSize(VarRect viewport, ToExpression width, ToExpression height) {
		return new Constraint[]{viewport.width().eq(width), viewport.height().eq(height)};
	}
}
