package dev.pfaff.altargui.improv;

import dev.pfaff.altargui.ffi.*;
import org.jetbrains.annotations.MustBeInvokedByOverriders;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * <h2>Lifecycle</h2>
 * <p>
 * The element lifecycle follows this path:
 *
 * <ol>
 *     <li>Create and attach to a {@link Context}</li>
 *     <li>Mount to a {@link #parent()}</li>
 *     <li>Perform layout</li>
 *     <li>Unmount from the {@link #parent()}</li>
 *     <li>Possibly go back to #2</li>
 *     <li>Dispose of the element
 *         <br><strong>NOTE</strong>: After this point the element can no longer be used, and internal state
 *         <strong>may</strong> be invalid.
 *     </li>
 * </ol>
 */
public abstract class Element {
	public static final Element[] EMPTY_ARRAY = new Element[0];
	public static final StackWalker STACK_WALKER = StackWalker.getInstance(StackWalker.Option.RETAIN_CLASS_REFERENCE);

	private final Context context;
	private @Nullable Element parent;
	private final VarRect allocation;
	private @Nullable ExprRect available;
	private boolean isMounted;

	protected Element(Context context) {
		this.context = context;
		this.allocation = context.newVarRect();
	}

	/**
	 * The UI context the element exists in.
	 */
	public final @NotNull Context context() {
		return this.context;
	}

	/**
	 * The element's current parent element, if any.
	 */
	public final @Nullable Element parent() {
		return this.parent;
	}

	/**
	 * @return the element's allocation rect. This return value is final.
	 */
	public final @NotNull VarRect allocation() {
		return this.allocation;
	}

	public final void ensureMounted() {
		if (!this.isMounted) throw new IllegalStateException("The element is not mounted: " + this);
	}

	/**
	 * Do not call this directly, unless you are the {@link Context} implementation.
	 */
	@MustBeInvokedByOverriders
	public void onMount(@Nullable Element parent) {
		if (this.isMounted) throw new IllegalStateException("The element is already mounted");
		this.parent = parent;
		this.isMounted = true;
	}

	/**
	 * Do not call this directly, unless you are the {@link Context} implementation.
	 */
	@MustBeInvokedByOverriders
	public void onUnmount() {
		this.ensureMounted();
		this.isMounted = false;
		this.parent = null;
		if (this.available != null) {
			this.available = null;
			this.onSetAvailable();
		}
	}

	/**
	 * Do not call this directly, unless you are the {@link Context} implementation.
	 */
	@MustBeInvokedByOverriders
	public void onDispose() {
		// TODO: remove element's constraints (will either need to track them for the entire element, or leave it up
		//  to subclasses... probably the latter)
	}

	// TODO: remove this altogether
	@Deprecated
	public final void setAvailable(@NotNull ExprRect available) {
		Objects.requireNonNull(available);
		if (!available.equals(this.available)) {
			this.available = available;
			this.onSetAvailable();
		}
	}

	/**
	 * Override this to invalidate parts of your layout when the "available" variables change. Note that any changes
	 * in this method should be queued, not executed immediately.
	 */
	@MustBeInvokedByOverriders
	protected void onSetAvailable() {
	}

	public final void queueLayout(LayoutMutation mutation) {
		var trace = STACK_WALKER.walk(Stream::toList);
		this.context.submitImmediateEvent(() -> {
			try {
				this.mutateLayoutImmediately(mutation);
			} catch (LayoutException e) {
				var traceString = trace.stream()
						.map(StackWalker.StackFrame::toString)
						.collect(Collectors.joining("\n"));
				throw new RuntimeException("Queued layout failed. Stack trace follows:\n" + traceString, e);
			}
		});
	}

	public final void mutateLayoutImmediately(LayoutMutation mutation) throws LayoutException {
		this.context.checkThread();
		try (var context = new LayoutContextImpl()) {
			mutation.apply(context);
		}
	}

	@Override
	public final boolean equals(Object obj) {
		return super.equals(obj);
	}

	@Override
	public final int hashCode() {
		return super.hashCode();
	}

	private final class LayoutContextImpl implements LayoutContext, AutoCloseable {
		private final FFILayoutContext ffiLayoutContext;
		private final ExprRect available;
		private boolean closed = false;

		private LayoutContextImpl() {
			this.ffiLayoutContext = Element.this.context.ffiEngine().newLayoutContext();
			this.available = Element.this.available;
		}

		private void ensureOpen() {
			if (this.closed) throw new IllegalStateException("The context is closed");
		}

		@Override
		public Context context() {
			this.ensureOpen();
			return Element.this.context();
		}

		@Override
		public FFILayoutContext ffiLayoutContext() {
			this.ensureOpen();
			return this.ffiLayoutContext;
		}

		@Override
		public ConstraintId addConstraint(@Nullable String comment, Constraint constraint) throws LayoutException {
			this.ensureOpen();
			Util.dumpConstraints(comment, new Constraint[] {constraint});
			var id = this.ffiLayoutContext.addConstraint(constraint);
			if (id == null) throw new AddConstraintException(comment, 0, constraint, new NullPointerException());
			return id;
		}

		@Override
		public void removeConstraint(ConstraintId id) throws LayoutException {
			this.ensureOpen();
			if (!this.ffiLayoutContext.removeConstraint(id)) {
				// TODO: log a warning
			}
		}

		@Override
		public ExprRect available() {
			return this.available;
		}

		@Override
		public void close() {
			this.ensureOpen();
			this.closed = true;
		}
	}
}
