package dev.pfaff.altargui.improv;

import dev.pfaff.altargui.types.Axis;
import dev.pfaff.altargui.ffi.ExprRect;
import dev.pfaff.altargui.ffi.Expression;
import dev.pfaff.altargui.ffi.Rect;
import dev.pfaff.altargui.improv.elements.*;
import dev.pfaff.altargui.function.CharProvider;

import java.util.List;
import java.util.function.Supplier;
import java.util.stream.Stream;

public final class ImprovTest {
	private static final class WindowElement {
		private final AbsoluteElement positionElement;
		private final SizedElement sizeElement;

		private WindowElement(Context context) {
			this.positionElement = new AbsoluteElement(context);
			this.sizeElement = new SizedElement(context);
			this.positionElement.setChild(this.sizeElement);
		}

		public void setChild(Element element) {
			this.sizeElement.setChild(element);
		}

		public void setSize(double width, double height) {
			this.sizeElement.setWidth(new Expression(width));
			this.sizeElement.setHeight(new Expression(height));
		}

		public void mount() {
			var context = this.positionElement.context();
			context.mountChild(null, this.positionElement);
			this.positionElement.setAvailable(new ExprRect(context.newVarRect()));
		}

		public void unmount() {
			var context = this.positionElement.context();
			context.unmountChild(null, this.positionElement);
		}
	}

	public static void test() {
		var context = Context.create();

		var window = new WindowElement(context);
		window.setSize(800.0, 600.0);

		window.mount();

		Supplier<PaddingElement> makePaddingElement = () -> {
			var element = new PaddingElement(context);
			element.setStartX(new Expression(4.0));
			element.setStartY(new Expression(4.0));
			element.setEndX(new Expression(4.0));
			element.setEndY(new Expression(4.0));
			return element;
		};

		{
			var textElement = new TextElement(context, CharProvider.of("Hello, World!"));

			{
				window.setChild(textElement);

				// call before checking values
				context.processEvents();

				{
					var laidContext = context.ffiEngine().newLaidOutContext();
					var varRect = textElement.allocation();
					var rect = varRect.getValues(laidContext);
					assert rect.equals(new Rect(0.0, 0.0, 8.0 * "Hello, World!".length(), 8.0)) : "rect: " + rect;
				}

				window.setChild(null);
			}

			var paddedElement = makePaddingElement.get();
			window.setChild(paddedElement);

			{
				paddedElement.setChild(textElement);

				// call before checking values
				context.processEvents();

				{
					var laidContext = context.ffiEngine().newLaidOutContext();
//				laidContext.dump();
					System.out.println("padding rect: " + paddedElement.allocation().getValues(laidContext));
					var varRect = textElement.allocation();
					var rect = varRect.getValues(laidContext);
					assert rect.equals(new Rect(4.0, 4.0, 8.0 * "Hello, World!".length(), 8.0)) : "rect: " + rect;
				}
			}

			{
				var rectElement = new RectElement(context);
				paddedElement.setChild(rectElement);

				context.processEvents();

				{
					var laidContext = context.ffiEngine().newLaidOutContext();
//				laidContext.dump();
					System.out.println("padding rect: " + paddedElement.allocation().getValues(laidContext));
					var varRect = rectElement.allocation();
					var rect = varRect.getValues(laidContext);
					assert rect.equals(new Rect(4.0, 4.0, 800 - 4.0 * 2, 600.0 - 4.0 * 2)) : "rect: " + rect;
				}

				paddedElement.setChild(null);
			}

			{
				var stackElement = new StackElement(context);
				paddedElement.setChild(stackElement);

				var background = new SizedElement(context);
				background.setChild(new RectElement(context));
				background.setWidth(textElement.allocation().width().toExpression());
				background.setHeight(textElement.allocation().height().toExpression());

				stackElement.setChildSupplier(MultiChildSupplier.of(List.of(
						background,
						textElement
				)));

				context.processEvents();

				{
					var laidContext = context.ffiEngine().newLaidOutContext();
//				laidContext.dump();
					var expectedTextRect = new Rect(4.0, 4.0, 8.0 * "Hello, World!".length(), 8.0);
					{
						var varRect = textElement.allocation();
						var rect = varRect.getValues(laidContext);
						assert rect.equals(expectedTextRect) : "rect: " + rect;
					}
					{
						var varRect = background.allocation();
						var rect = varRect.getValues(laidContext);
						assert rect.equals(expectedTextRect) : "rect: " + rect;
					}
					{
						var varRect = stackElement.allocation();
						var rect = varRect.getValues(laidContext);
						assert rect.equals(expectedTextRect) : "rect: " + rect;
					}
				}
			}
		}

		{
			Supplier<Element> makeChild = () -> {
				var stackElement = new StackElement(context);

				var textElement = new TextElement(context, CharProvider.of("Hello, World!"));

				var background = new SizedElement(context);
				background.setChild(new RectElement(context));
				background.setWidth(textElement.allocation().width().toExpression());
				background.setHeight(textElement.allocation().height().toExpression());

				stackElement.setChildSupplier(MultiChildSupplier.of(List.of(
						background,
						textElement
				)));

				return stackElement;
			};

			var trackElement = new TrackElement(context, Axis.X);
			var paddedElement = makePaddingElement.get();
			window.setChild(paddedElement);
			paddedElement.setChild(trackElement);

			int n = 1;
			trackElement.setChildSupplier(MultiChildSupplier.of(Stream.generate(makeChild)
					.limit(n)
					.toList()));

			try {
				context.processEvents();
			} catch (Throwable e) {
				var laidContext = context.ffiEngine().newLaidOutContext();
				laidContext.dump();
				throw e;
			}

			{
				var laidContext = context.ffiEngine().newLaidOutContext();
				laidContext.dump();
				var expectedTextRect = new Rect(4.0, 4.0, 8.0 * "Hello, World!".length(), 8.0);
				{
					var varRect = trackElement.getChild(0).allocation();
					var rect = varRect.getValues(laidContext);
					assert rect.equals(expectedTextRect) : "rect: " + rect + ", vars: " + varRect;
				}
				{
					var varRect = trackElement.getChild(n - 1).allocation();
					var rect = varRect.getValues(laidContext);
					assert rect.equals(new Rect(expectedTextRect.x() + expectedTextRect.width() * (n - 1),
							expectedTextRect.y(),
							expectedTextRect.width(),
							expectedTextRect.height())) : "rect: " + rect;
				}
				{
					var rect = trackElement.allocation().getValues(laidContext);
					assert rect.equals(new Rect(expectedTextRect.x(),
							expectedTextRect.y(),
							expectedTextRect.width() * n,
							expectedTextRect.height())) : "rect: " + rect;
				}
			}
		}
	}
}
