package dev.pfaff.altargui.improv.elements;

import dev.pfaff.altargui.improv.Context;
import dev.pfaff.altargui.improv.Element;
import dev.pfaff.altargui.improv.MultiChildSupplier;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.OpenArrayList;
import dev.pfaff.altargui.util.Sink;
import org.jetbrains.annotations.MustBeInvokedByOverriders;

import java.util.Objects;
import java.util.function.ObjIntConsumer;

public abstract class AbstractMultiChildElement extends Element {
	private final OpenArrayList<Element> currentChildren = OpenArrayList.wrap(Element.EMPTY_ARRAY);

	private MultiChildSupplier supplier;
	private ListenerDisposer supplierListener;

	public AbstractMultiChildElement(Context context) {
		super(context);
	}

	@Override
	public void onDispose() {
		super.onDispose();

		if (this.supplierListener != null) {
			this.supplierListener.close();
			this.supplierListener = null;
		}
	}

	public void setChildSupplier(MultiChildSupplier supplier) {
		Objects.requireNonNull(supplier);
		if (!supplier.equals(this.supplier)) {
			this.supplier = supplier;
			if (this.supplierListener != null) this.supplierListener.close();
			this.supplierListener = supplier.listenable().listenAndInvoke(() -> {
				supplier.forEach(new ElementSink());
			});
		}
	}

	public Element getChild(int index) {
		return this.currentChildren.get(index);
	}

	public int childCount() {
		return this.currentChildren.size();
	}

	public void forEachChild(ObjIntConsumer<Element> consumer) {
		var a = this.currentChildren.a();
		var upr = this.currentChildren.size();
		if (upr <= a.length) {
			for (int i = 0; i < upr; i++) {
				consumer.accept(a[i], i);
			}
		}
	}

	@MustBeInvokedByOverriders
	protected void onMountChild(int index, Element child, boolean moved) {
	}

	@MustBeInvokedByOverriders
	protected void onUnmountChild(int index, Element child) {
	}

	/**
	 * Called after {@link #onMountChild(int, Element, boolean)} and {@link #onUnmountChild(int, Element)}.
 	 * @param count the number of child elements
	 */
	@MustBeInvokedByOverriders
	protected void onUpdateChildCount(int count) {
	}

	private class ElementSink implements Sink<Element> {
		private int childCount = 0;

		@Override
		public void begin(long sizeL) {
			if (sizeL > Integer.MAX_VALUE) throw new IllegalArgumentException("too many children");
			if (sizeL != -1) {
				int size = (int) sizeL;
				// preallocate enough capacity
				// note that if any elements are replaced, a reallocation may be necessary later
				AbstractMultiChildElement.this.currentChildren.ensureCapacity(size);
			}
		}

		@Override
		public void end() {
			Element[] a = AbstractMultiChildElement.this.currentChildren.a();
			int upr = AbstractMultiChildElement.this.currentChildren.size();
			if (upr <= a.length) {
				for (int i = this.childCount; i < upr; i++) {
					var element = a[i];
					AbstractMultiChildElement.this.context().unmountChild(AbstractMultiChildElement.this, element);
					AbstractMultiChildElement.this.onUnmountChild(i, element);
				}
			}
			AbstractMultiChildElement.this.currentChildren.size(this.childCount);
			AbstractMultiChildElement.this.onUpdateChildCount(this.childCount);
		}

		@Override
		public void accept(Element element) {
			var currentChildren = AbstractMultiChildElement.this.currentChildren;
			int currentI = currentChildren.indexOf(element);
			int desiredI = this.childCount++;
			if (currentI != desiredI) { // not in desired order
				if (currentI == -1) { // new child
					if (desiredI == currentChildren.size()) {
						AbstractMultiChildElement.this.currentChildren.add(element);
					} else {
						assert desiredI < currentChildren.size();
						var wrongElement = currentChildren.set(desiredI, element);
						currentChildren.add(wrongElement);
					}
					AbstractMultiChildElement.this.context().mountChildImmediately(AbstractMultiChildElement.this,
							element);
					AbstractMultiChildElement.this.onMountChild(desiredI, element, false);
				} else if (currentI < desiredI) {
					throw new IllegalArgumentException("Duplicate child element:"
							+ "\n  parent: " + AbstractMultiChildElement.this
							+ "\n  child: " + element);
				} else { // swap wrong element (at desired position) with element at current position
					var wrongElement = currentChildren.get(desiredI);
					currentChildren.set(desiredI, element);
					currentChildren.set(currentI, wrongElement);
					AbstractMultiChildElement.this.onMountChild(desiredI, element, true);
				}
			}
		}
	}
}
