package dev.pfaff.altargui.improv.elements;

import dev.pfaff.altargui.types.Axis;
import dev.pfaff.altargui.ffi.ExprRect;
import dev.pfaff.altargui.ffi.Expression;
import dev.pfaff.altargui.ffi.Strength;
import dev.pfaff.altargui.ffi.Variable;
import dev.pfaff.altargui.improv.ConstraintGroup;
import dev.pfaff.altargui.improv.Context;
import dev.pfaff.altargui.improv.Element;
import dev.pfaff.altargui.util.OpenArrayList;
import dev.pfaff.altargui.types.Tuple;
import org.jetbrains.annotations.NotNull;

public final class TrackElement extends AbstractMultiChildElement {
	private final OpenArrayList<Entry> entries = OpenArrayList.wrap(new Entry[0]);

	private record ChildPosAvailable(Variable mainPosition, Variable crossPosition, Expression mainAvailable, Variable crossAvailable) {
	}

	@NotNull
	private static Variable getChildAvailable(TrackElement trackElement) {
		var allocation = trackElement.allocation();
		if (!(trackElement.axis.select(allocation.width(), allocation.height()) instanceof Tuple(
				_, var crossAvailable
		))) throw new AssertionError();
		return crossAvailable;
	}

	@NotNull
	private ChildPosAvailable getChildPosAvailable(int index) {
		var allocation = this.allocation();
		if (!(this.axis.select(allocation.x(), allocation.y()) instanceof Tuple(
				var firstMainPosition, var crossPosition
		))) throw new AssertionError();
		var mainPosition = index < 1 ? firstMainPosition : this.entries.get(index - 1).nextMainPosition;
		if (!(this.axis.select(allocation.width(), allocation.height()) instanceof Tuple(
				var fullMainAvailable, var crossAvailable
		))) throw new AssertionError();
		return new ChildPosAvailable(mainPosition,
				crossPosition,
				fullMainAvailable.minus(mainPosition).toExpression(),
				crossAvailable);
	}

	private static final class Entry {
		/**
		 * Depends on {@link TrackElement#axis} and {@link TrackElement#getChild(int)}.
		 */
		public final ConstraintGroup constraintGroup;
		public final Variable nextMainPosition;

		private Entry(TrackElement trackElement, int index) {
			var nextChildMainPosTag = "Track[child[" + index + "]::next_main_pos]";
			var childPosTag = "Track[child[" + index + "]::pos]";
			var minimizeCrossTag = "Track[self::minimize_cross_size_to_child[" + index + "]]";
			var allocation = trackElement.allocation();
			this.nextMainPosition = trackElement.context().newVariable();
			this.constraintGroup = new ConstraintGroup(trackElement, context -> {
				var child = trackElement.getChild(index);
				if (child == null) return;

				// gather variables
				ChildPosAvailable desiredChildPosAvailable = trackElement.getChildPosAvailable(index);
				var childAllocation = child.allocation();
				if (!(trackElement.axis.select(childAllocation.x(), childAllocation.y()) instanceof Tuple(
						var childPosMain, var childPosCross
				))) throw new AssertionError();
				if (!(trackElement.axis.select(childAllocation.width(), childAllocation.height()) instanceof Tuple(
						var childSizeMain, var childSizeCross
				))) throw new AssertionError();

				context.addConstraints(nextChildMainPosTag,
						childPosMain.plus(childSizeMain).eq(this.nextMainPosition));

				context.addConstraints(childPosTag,
						childPosMain.eq(desiredChildPosAvailable.mainPosition()),
						childPosCross.eq(desiredChildPosAvailable.crossPosition()));

				context.addConstraints(minimizeCrossTag,
						childSizeCross.eq(desiredChildPosAvailable.crossAvailable()).withStrength(Strength.WEAK));
			});
		}
	}

	private Axis axis;

	public TrackElement(Context context, Axis axis) {
		super(context);
		this.axis = axis;
	}

	private final ConstraintGroup groupAvailable = ExpandElement.makeAvailableGroup(this, "Track");

	private final ConstraintGroup groupMinimize = new ConstraintGroup(this, context -> {
		var tag = "Track[self::lock_main_size_to_sum_of_children]";

		var allocation = this.allocation();
		if (!(this.axis.select(allocation.width(), allocation.height()) instanceof Tuple(
				var mainSize, _
		))) throw new AssertionError();

		if (this.entries.isEmpty()) {
			context.addConstraint(tag, mainSize.eq(0.0));
		}
		if (!(this.axis.select(allocation.x(), allocation.y()) instanceof Tuple(
				var mainPosition, _
		))) throw new AssertionError();

		context.addConstraint(tag, this.entries.getLast().nextMainPosition.minus(mainPosition).eq(mainSize));
	});

	public void setAxis(Axis axis) {
		if (this.axis != axis) {
			this.axis = axis;
			this.entries.forEach(entry -> entry.constraintGroup.queueUpdate());
			this.groupMinimize.queueUpdate();
			this.forEachChild((e, i) -> this.updateChildAvailable(i, e));
		}
	}

	private void updateChildAvailable(int index, Element child) {
		var childPosition = this.getChildPosAvailable(index);
		var xy = this.axis.select(childPosition.mainPosition(), childPosition.crossPosition());
		var widthHeight = this.axis.select(childPosition.mainAvailable(), childPosition.crossAvailable());
		child.setAvailable(new ExprRect(xy.a(), xy.b(), widthHeight.a(), widthHeight.b()));
	}

	@Override
	protected void onMountChild(int index, Element child, boolean moved) {
		super.onMountChild(index, child, moved);
		if (index >= this.entries.size()) {
			if (index != this.entries.size()) throw new IllegalStateException("onMountChild invariant violation");
			this.entries.add(new Entry(this, index));
		}
		this.updateChildAvailable(index, child);
		this.entries.get(index).constraintGroup.queueUpdate();
	}

	@Override
	protected void onUnmountChild(int index, Element child) {
		super.onUnmountChild(index, child);
		this.entries.get(index).constraintGroup.queueReset();
	}

	@Override
	protected void onUpdateChildCount(int count) {
		super.onUpdateChildCount(count);
		this.groupMinimize.queueUpdate();
		var entries = this.entries.a();
		int upr = this.entries.size();
		if (upr <= entries.length) {
			for (int i = count; i < upr;) {
				entries[i++].constraintGroup.queueReset();
			}
		}
		this.entries.size(count);
	}

	@Override
	protected void onSetAvailable() {
		super.onSetAvailable();
		this.groupAvailable.queueUpdate();
	}
}
