package dev.pfaff.altargui.inspector;

import dev.pfaff.altargui.Inspector;
import dev.pfaff.altargui.Key;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.widgets.FrameBuilderWidget;
import dev.pfaff.altargui.widgets.TextWidget;
import dev.pfaff.altargui.widgets.TrackingWidgets;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.*;
import java.util.*;

import static dev.pfaff.altargui.Inspector.withInspectorFontSize;
import static dev.pfaff.altargui.Key.localKey;
import static dev.pfaff.altargui.util.MethodHandleUtil.findStatic;
import static dev.pfaff.altargui.widgets.TextWidgets.text;
import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.methodType;

public sealed interface DebugPropertyValue
	permits DetailedRawObjectValue, NonDetailedRawObjectValue, StaticWidgetValue {
	/**
	 * Returns a one-line view of the value.
	 */
	@NotNull Widget buildPreview(Inspector inspector);

	DebugPropertyValue NULL = ofLiteral("null");
	DebugPropertyValue TRUE = ofLiteral("true");
	DebugPropertyValue FALSE = ofLiteral("false");

	MethodHandle MH_ofBoolean = makeMH_of(boolean.class);
	MethodHandle MH_ofByte = makeMH_of(byte.class);
	MethodHandle MH_ofShort = makeMH_of(short.class);
	MethodHandle MH_ofChar = makeMH_of(char.class);
	MethodHandle MH_ofInt = makeMH_of(int.class);
	MethodHandle MH_ofLong = makeMH_of(long.class);
	MethodHandle MH_ofFloat = makeMH_of(float.class);
	MethodHandle MH_ofDouble = makeMH_of(double.class);
	MethodHandle MH_ofObject = makeMH_of(Object.class);

	static MethodHandle getMH_of(Class clazz) {
		if (clazz == boolean.class) return MH_ofBoolean;
		if (clazz == byte.class) return MH_ofByte;
		if (clazz == short.class) return MH_ofShort;
		if (clazz == char.class) return MH_ofChar;
		if (clazz == int.class) return MH_ofInt;
		if (clazz == long.class) return MH_ofLong;
		if (clazz == float.class) return MH_ofFloat;
		if (clazz == double.class) return MH_ofDouble;
		return MH_ofObject;
	}

	private static MethodHandle makeMH_of(Class<?> clazz) {
		var l = MethodHandles.lookup();
		return findStatic(l, DebugPropertyValue.class, "of", methodType(DebugPropertyValue.class, clazz));
	}

	static DebugPropertyValue of(boolean value) {
		return value ? TRUE : FALSE;
	}

	static DebugPropertyValue of(byte value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (byte v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(short value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (short v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(char value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (char v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(int value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (int v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(long value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (long v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(float value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (float v) -> withInspectorFontSize(text(v))));
	}

	static DebugPropertyValue of(double value) {
		return new StaticWidgetValue(TrackingWidgets.tracking(value, (double v) -> withInspectorFontSize(text(v))));
	}

	static <T> DebugPropertyValue ofObjectRaw(@NotNull T value) {
		var preview = Objects.requireNonNull(ValueRenderers.forValue(value, ValueRenderers.Kind.Preview));
		var detailed = ValueRenderers.forValue(value, ValueRenderers.Kind.Detailed);
		if (detailed == null) return new NonDetailedRawObjectValue<>(preview, value);
		return new DetailedRawObjectValue<>(detailed, value, preview);
	}

	static DebugPropertyValue of(@Nullable Object value) {
		return switch (value) {
			case null -> NULL;
			// should be impossible to get a recursive Optional, right?
			case Optional<?> o -> of(o.orElse(null));
			case OptionalDouble o -> o.isPresent() ? of(o.getAsDouble()) : NULL;
			case OptionalInt o -> o.isPresent() ? of(o.getAsInt()) : NULL;
			case OptionalLong o -> o.isPresent() ? of(o.getAsLong()) : NULL;
			case Boolean b -> of((boolean) b);
			case Byte b -> of((byte) b);
			case Short s -> of((short) s);
			case Character c -> of((char) c);
			case Integer i -> of((int) i);
			case Long l -> of((long) l);
			case Float f -> of((float) f);
			case Double d -> of((double) d);
			default -> ofObjectRaw(value);
		};
	}

	public record FieldInfo(String name,
							MethodHandle namedPreviewBuilder,
							Key.CallSite key) {
		private static final MethodHandle MH_buildKeyValueEntry;
		private static final MethodHandle MH_buildValuePreview;
		private static final MethodHandle MH_buildEveryFrame;
		public static final MethodType MT_W_WID = methodType(Widget.class,
				Widget.class,
				Inspector.class,
				DebugPropertyValue.class);

		static {
			var l = MethodHandles.lookup();
			MH_buildKeyValueEntry = findStatic(l,
					Inspector.class,
					"buildKeyValueEntry",
					methodType(Widget.class, Widget.class, Widget.class));
			try {
				MH_buildValuePreview = l.findVirtual(Inspector.class,
						"buildValuePreview",
						methodType(Widget.class, DebugPropertyValue.class));

				MH_buildEveryFrame = l.findStatic(FieldInfo.class,
						"buildEveryFrame",
						methodType(Widget.class, MethodHandle.class, Inspector.class, Object.class));
			} catch (NoSuchMethodException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
		}

		public FieldInfo(String name, @Nullable String tooltip, MethodHandle fieldGetter, boolean isFinal) {
			this(name, makeNamedPreviewBuilder(makeNameWidget(name, tooltip), fieldGetter, isFinal), localKey());
		}

		private static @NotNull Widget makeNameWidget(String name, @Nullable String tooltip) {
			Widget widget = withInspectorFontSize(TextWidget.of(name));
			if (tooltip != null) {
				widget = widget.withTooltip(withInspectorFontSize(TextWidget.of(tooltip)));
			}
//			widget = row(List.of(ButtonWidget.inlineStyle(CLICK_TO_INSPECT_FIELD_TEXT, () -> {
//				Logger.log(() -> "TODO: inspect Field", List.of(WARN));
//			}), widget));
			return widget;
		}

		private static Widget buildEveryFrame(MethodHandle builder, Inspector inspector, Object value) {
			return new FrameBuilderWidget(_ -> {
				try {
					return (Widget) builder.invokeExact(inspector, value);
				} catch (Throwable e) {
					throw new RuntimeException(e);
				}
			});
		}

		private static MethodHandle makeNamedPreviewBuilder(@NotNull Widget nameWidget, MethodHandle fieldGetter, boolean isFinal) {
			var getter = makeDebugPropertyGetter(fieldGetter);
			// bind the nameWidget
			var MH = insertArguments(dropArguments(MH_buildKeyValueEntry, 0, Inspector.class), 1, nameWidget);
			MH = permuteArguments(MH, MT_W_WID, 1, 0);
			MH = foldArguments(MH, MH_buildValuePreview);
			MH = filterArguments(MH, 1, getter);
			if (isFinal) return MH;

			var builder = MH.asType(methodType(Widget.class, Inspector.class, Object.class));

//			try {
//				var lambdaFactory = LambdaMetafactory.metafactory(MethodHandles.lookup(),
//								"build",
//								methodType(builder.type().parameterType(0)),
//								FrameBuilderWidget.Builder.MT_build,
//								dropArguments(builder, 1, long.class),
//								FrameBuilderWidget.Builder.MT_build)
//						.dynamicInvoker();
//				return filterReturnValue(lambdaFactory, MH_newFrameBuilderWidget);
//			} catch (LambdaConversionException e) {
//				throw new RuntimeException(e);
//			}
			return MH_buildEveryFrame.bindTo(builder);
		}

		private static MethodHandle makeDebugPropertyGetter(MethodHandle fieldGetter) {
			var debugPropertyValueConstructor = getMH_of(fieldGetter.type().returnType())
				.asType(methodType(DebugPropertyValue.class, fieldGetter.type().returnType()));
			return MethodHandles.filterArguments(debugPropertyValueConstructor, 0, fieldGetter)
								.asType(methodType(DebugPropertyValue.class, Object.class));
		}
	}

	static @NotNull DebugPropertyValue ofLiteral(@NotNull String value) {
		return new StaticWidgetValue(withInspectorFontSize(TextWidget.of(value)));
	}
}
