package dev.pfaff.altargui.inspector;

import dev.pfaff.altargui.Inspector;
import dev.pfaff.altargui.LoadableTex2d;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.inspector.renderers.*;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.widgets.TextWidget;
import dev.pfaff.altargui.widgets.TrackingWidgets;
import dev.pfaff.log4truth.Logger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.invoke.VarHandle;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.file.Path;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Stream;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.Inspector.OBJECT_INSPECTOR_Y_SPACING;
import static dev.pfaff.altargui.Inspector.withInspectorFontSize;
import static dev.pfaff.altargui.util.MethodHandleUtil.findStatic;
import static dev.pfaff.altargui.widgets.MultiChildWidget.col;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.WARN;
import static java.lang.invoke.MethodType.methodType;

/**
 * Registry of value renderers.
 */
public final class ValueRenderers {
    private static final Map<Class, DefaultRenderer> DEFAULT_RENDERERS = new IdentityHashMap<>();
    private static final Map<Class, @Nullable ValueRenderer> EXACT_CLASS_DETAILED_RENDERERS = new IdentityHashMap<>();
    private static final Map<Class, @NotNull ValueRenderer> EXACT_CLASS_PREVIEW_RENDERERS = new IdentityHashMap<>();
    private static final Map<Class, Function<Class, ValueRenderer>> SUPER_CLASS_DETAILED_RENDERERS = new IdentityHashMap<>();
    private static final Map<Class, Function<Class, ValueRenderer>> SUPER_CLASS_PREVIEW_RENDERERS = new IdentityHashMap<>();

    public enum Kind {
        Preview,
        Detailed,
    }

    public static <T> @Nullable ValueRenderer<T> forClass(Class<T> clazz, @NotNull Kind kind) {
        if (clazz.isPrimitive()) throw new IllegalArgumentException("Primitives are handled elsewhere");
        return switch (kind) {
            case Preview -> EXACT_CLASS_PREVIEW_RENDERERS.computeIfAbsent(clazz, c -> makeRenderer(c, Kind.Preview, null));
            case Detailed -> EXACT_CLASS_DETAILED_RENDERERS.computeIfAbsent(clazz, c -> makeRenderer(c, Kind.Detailed, null));
        };
    }

    public static <T> @Nullable ValueRenderer<T> forValue(T value, @NotNull Kind kind) {
        return (ValueRenderer<T>) forClass((Class<? extends T>) value.getClass(), kind);
    }

    public static <T> ValueRenderer<T> defaultForClass(Class<? extends T> clazz, @NotNull Kind kind) {
        if (clazz.isPrimitive()) throw new IllegalArgumentException("Primitives are handled elsewhere");
        var renderer = DEFAULT_RENDERERS.computeIfAbsent(clazz, ValueRenderers::makeDefaultRenderer);
        return switch (kind) {
            case Preview -> renderer.preview();
            case Detailed -> renderer.detailed();
        };
    }

    public static <T> ValueRenderer<T> defaultForValue(T value, @NotNull Kind kind) {
        return (ValueRenderer<T>) defaultForClass(value.getClass(), kind);
    }

    private static <T extends S, S> @Nullable ValueRenderer<T> getExactSuperClassRenderer(Class<T> clazz, Class<S> superClass, @NotNull Kind kind) {
        LOGGER.log(() -> STR."checking for super class renderer on \{superClass}", List.of(DEBUG));
        var superClassRenderer = switch (kind) {
            case Preview -> SUPER_CLASS_PREVIEW_RENDERERS.get(superClass);
            case Detailed -> SUPER_CLASS_DETAILED_RENDERERS.get(superClass);
        };
        if (superClassRenderer != null) {
            LOGGER.log(() -> STR."found super class renderer on \{superClass}", List.of(DEBUG));
            //noinspection unchecked
            return superClassRenderer.apply(clazz);
        }
        return null;
    }

    private static <T extends S, S> @Nullable ValueRenderer<T> findSuperClassRenderer(Class<T> clazz, Class<S> superClass, @NotNull Kind kind) {
        var renderer = ValueRenderers.getExactSuperClassRenderer(clazz, superClass, kind);
        if (renderer != null) return renderer;

        var superSuperClass = superClass.getSuperclass();
        if (superSuperClass != null) {
            renderer = ValueRenderers.getExactSuperClassRenderer(clazz, superSuperClass, kind);
            if (renderer != null) return renderer;
        }

        for (Class<?> superInterface : superClass.getInterfaces()) {
            renderer = ValueRenderers.getExactSuperClassRenderer(clazz, (Class<? super T>) superInterface, kind);
            if (renderer != null) return renderer;
        }

        // yes, this will duplicate the above 2 cases. I can split the method to eliminate that later.
        if (superSuperClass != null) {
            renderer = ValueRenderers.findSuperClassRenderer(clazz, superSuperClass, kind);
            if (renderer != null) return renderer;
        }

        for (Class<?> superInterface : superClass.getInterfaces()) {
            renderer = ValueRenderers.findSuperClassRenderer(clazz, (Class<? super T>) superInterface, kind);
            if (renderer != null) return renderer;
        }
        return null;
    }

    private static <T> ValueRenderer<? super T> makeRenderer(Class<T> clazz, @NotNull Kind kind, @Nullable ValueRenderer<? super T> override) {
        LOGGER.log(() -> STR."preparing renderer for \{clazz}", List.of(DEBUG));
        if (override == null) {
            override = findSuperClassRenderer(clazz, clazz, kind);
        }
        if (override == null) return ValueRenderers.defaultForClass(clazz, kind);
        return switch (kind) {
            case Preview -> override;
            // TODO: build a custom class for each one
            case Detailed -> new CustomDetailsValueRenderer<>(override, ValueRenderers.defaultForClass(clazz, kind));
        };
    }

    private static final Widget FIELDS_WARNING = withInspectorFontSize(TextWidget.of("* Some fields may be omitted due to access restrictions"))
            .withVerticalMargin(OBJECT_INSPECTOR_Y_SPACING);

    private static <T> DefaultRenderer<T> makeDefaultRenderer(Class<T> clazz) {
        var className = clazz.getSimpleName();
        int lambdaI = className.indexOf('/');
        MethodType lambdaType = null;
        if (lambdaI != -1) {
            var methods = Arrays.stream(clazz.getDeclaredMethods())
                                .filter(method -> !Modifier.isStatic(method.getModifiers()))
                                .toList();
            if (methods.size() == 1) {
                var method = methods.get(0);
                lambdaType = methodType(method.getReturnType(), method.getParameterTypes());
            } else {
                LOGGER.log(() -> STR."Lambda has more than one instance method: \{methods}", List.of(WARN));
            }
        }

        if (clazz.isArray()) {
            return new DefaultRenderer<>(CollectionValueRenderers.arrayPreview(clazz),
                                         CollectionValueRenderers.arrayDetailed(clazz));
        }

        var displayName = withInspectorFontSize(TextWidget.of(className));
        var display = lambdaType != null ? withInspectorFontSize(TextWidget.of(MethodTypeValueRenderer.format(lambdaType)))
                                                             .withTooltip(displayName) : displayName;

        Stream<DebugPropertyValue.FieldInfo> fieldsStream = Stream.empty();
        Class superClass = clazz;
        while (superClass != null) {
            fieldsStream = Stream.concat(fieldsStream, getFields(superClass));
            superClass = superClass.getSuperclass();
        }
        var fields = fieldsStream
                .sorted(Comparator.comparing(DebugPropertyValue.FieldInfo::name))
                .map(DebugPropertyValue.FieldInfo::namedPreviewBuilder)
                .toList();

        var impl = MH_defaultRendererDetailedImpl
                .asType(MH_defaultRendererDetailedImpl.type().changeParameterType(2, clazz))
                .asType(MH_defaultRendererDetailedImpl.type())
                .bindTo(fields);
        return new DefaultRenderer<>((inspector, value) -> display, (inspector, value) -> {
			try {
				return (Widget) impl.invokeExact(inspector, value);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		});
    }

    private static final MethodHandle MH_defaultRendererDetailedImpl;

    static {
        var l = MethodHandles.lookup();
        MH_defaultRendererDetailedImpl = findStatic(l, ValueRenderers.class, "defaultRendererDetailedImpl", methodType(Widget.class, List.class, Inspector.class, Object.class));
    }

    @NotNull
    private static <T> Widget defaultRendererDetailedImpl(List<MethodHandle> fields, Inspector inspector, T value) {
        // TODO: use countedLoop for building the fields widget list
//        MethodHandles.countedLoop(MethodHandles.constant(int.class, fields.size()), MethodHandles.constant())
        return col(List.of(
                FIELDS_WARNING,
                Inspector.labeledCol("Fields", fields.stream().map(field -> {
                    try {
                        return (Widget) field.invoke(inspector, value);
                    } catch (Throwable e) {
                        throw new RuntimeException(e);
                    }
                }).toList())
        ));
    }

    private static Stream<DebugPropertyValue.FieldInfo> getFields(Class<?> clazz) {
        MethodHandles.Lookup l = MethodHandles.lookup();
        boolean publicOnly = false;
        try {
            l = MethodHandles.privateLookupIn(clazz, l);
        } catch (IllegalAccessException e1) {
            // ignore, use fallback lookup
            publicOnly = true;
			try {
				l.accessClass(clazz);
			} catch (IllegalAccessException e2) {
                e2.addSuppressed(e1);
                l = MethodHandles.publicLookup();
                try {
                    l.accessClass(clazz);
                } catch (IllegalAccessException e3) {
                    e3.addSuppressed(e2);
                    Logger.log(() -> STR."Unable to access fields of \{clazz} due to access restrictions", e3, List.of(WARN));
                    return Stream.empty();
                }
			}
		}
        var publicOnly_ = publicOnly;
        var l_ = l;
		if (clazz.isRecord()) {
			return Arrays.stream(clazz.getRecordComponents())
                .map(comp -> {
                    var meta = comp.getAnnotation(InspectorMeta.class);
                    MethodHandle mh;
                    try {
                        mh = l_.unreflect(comp.getAccessor());
                    } catch (IllegalAccessException e) {
                        throw new RuntimeException(e);
                    }
                    // the warning saying that meta is always non-null is wrong.
                    return new DebugPropertyValue.FieldInfo(comp.getName(),
                            meta != null ? meta.tooltip() : null,
                            mh,
                            true);
                });
		}
		return Arrays.stream(clazz.getDeclaredFields())
            .filter(field -> {
                if (Modifier.isStatic(field.getModifiers())) return false;
                if (publicOnly_ && !Modifier.isPublic(field.getModifiers())) return false;
                return true;
            })
            .map(field -> {
                var meta = field.getAnnotation(InspectorMeta.class);
                VarHandle vh;
                try {
                    vh = l_.unreflectVarHandle(field);
                } catch (IllegalAccessException e) {
                    throw new RuntimeException(e);
                }
                // the warning saying that meta is always non-null is wrong.
                return new DebugPropertyValue.FieldInfo(field.getName(),
                        meta != null ? meta.tooltip() : null,
                        vh.toMethodHandle(VarHandle.AccessMode.GET),
                        Modifier.isFinal(field.getModifiers()));
            });
	}

    public static <T> void registerExactRenderer(Class<? extends T> clazz, @NotNull Kind kind, @NotNull ValueRenderer<T> override) {
        Objects.requireNonNull(override);
        var map = switch (kind) {
            case Preview -> EXACT_CLASS_PREVIEW_RENDERERS;
            case Detailed -> EXACT_CLASS_DETAILED_RENDERERS;
        };
        if (map.putIfAbsent(clazz, makeRenderer(clazz, kind, override)) != null) {
            throw new IllegalStateException(STR."Renderer for \{clazz.getName()} exactly is already registered");
        }
    }

    public static <T, R> void registerExactRendererBy(Class<? extends T> clazz, @NotNull Kind kind, @NotNull Function<T, R> func) {
        registerExactRenderer(clazz, kind, ValueRenderer.by(func, kind));
    }

    public static <T> void registerSuperClassRenderer(Class<T> clazz, @NotNull Kind kind, @NotNull Function<Class<? extends T>, @NotNull ValueRenderer<T>> override) {
        Objects.requireNonNull(override);
        var map = switch (kind) {
            case Preview -> SUPER_CLASS_PREVIEW_RENDERERS;
            case Detailed -> SUPER_CLASS_DETAILED_RENDERERS;
        };
        if (map.putIfAbsent(clazz, (Function) override) != null) {
            throw new IllegalStateException(STR."Renderer for \{clazz.getName()} and sub-classes is already registered");
        }
    }

    private record DefaultRenderer<T>(ValueRenderer<T> preview, ValueRenderer<T> detailed) {}

    static {
        CollectionValueRenderers.register();
        registerExactRenderer(String.class, Kind.Preview, (inspector, value) -> {
			return TrackingWidgets.tracking(value, s -> {
				return withInspectorFontSize(TextWidget.of('"' + s.replace("\"", "\\\"") + '"'));
			});
		});
        registerSuperClassRenderer(Enum.class, Kind.Preview, c -> EnumValueRenderer.preview(c));
        registerExactRenderer(MethodType.class, Kind.Preview, MethodTypeValueRenderer.PREVIEW);
        registerSuperClassRenderer(Font.class, Kind.Detailed, c -> FontValueRenderer.DETAILED);
        registerExactRenderer(LoadableTex2d.class, Kind.Detailed, Tex2dValueRenderer.DETAILED);
        registerExactRenderer(Class.class, Kind.Preview, ClassValueRenderer.PREVIEW);
        registerExactRenderer(Class.class, Kind.Detailed, ClassValueRenderer.DETAILED);
        registerExactRenderer(Method.class, Kind.Preview, MethodValueRenderer.PREVIEW);
        registerExactRenderer(Method.class, Kind.Detailed, MethodValueRenderer.DETAILED);
        registerExactRenderer(Field.class, Kind.Preview, FieldValueRenderer.PREVIEW);
        registerExactRenderer(Field.class, Kind.Detailed, FieldValueRenderer.DETAILED);
        registerSuperClassRenderer(Path.class, Kind.Preview, c -> PathValueRenderer.PREVIEW);
    }
}
