package dev.pfaff.altargui.inspector.renderers;

import dev.pfaff.altargui.Inspector;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.inspector.DebugPropertyValue;
import dev.pfaff.altargui.inspector.ValueRenderer;
import dev.pfaff.altargui.inspector.ValueRenderers;
import dev.pfaff.altargui.util.MethodHandleUtil;
import dev.pfaff.altargui.widgets.TextFragmentWidget;
import dev.pfaff.altargui.widgets.TextWidget;
import dev.pfaff.altargui.widgets.TrackingWidgets;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.StringConcatException;
import java.util.Map;

import static dev.pfaff.altargui.Inspector.withInspectorFontSize;
import static dev.pfaff.altargui.Key.intKey;
import static dev.pfaff.altargui.Key.valueKey;
import static dev.pfaff.altargui.util.MethodHandleUtil.*;
import static dev.pfaff.altargui.widgets.TextWidgets.text;
import static java.lang.invoke.MethodHandles.arrayLength;
import static java.lang.invoke.MethodHandles.filterReturnValue;
import static java.lang.invoke.MethodType.methodType;
import static java.lang.invoke.StringConcatFactory.makeConcatWithConstants;

public final class CollectionValueRenderers {
    public static final ValueRenderer<Iterable> ITERABLE_DETAILED = (inspector, value) -> Inspector.labeledCol("Elements", mounter -> {
        int i = 0;
        for (var element : value) {
            var elementValue = DebugPropertyValue.of(element);
            mounter.layout(intKey(i), inspector.buildKeyValueEntry(withInspectorFontSize(text(i++)), elementValue));
        }
    });

    public static final ValueRenderer<Map> MAP_DETAILED = (inspector, value) -> Inspector.labeledCol("Entries", mounter -> {
        for (var entry : ((Map<?, ?>) value).entrySet()) {
            var key = entry.getKey();
            var entryValue = DebugPropertyValue.of(entry.getValue());
            mounter.layout(valueKey(key), inspector.buildKeyValueEntry(TrackingWidgets.tracking(key, k -> withInspectorFontSize(TextWidget.of(String.valueOf(k)))), entryValue));
        }
    });

    public static void register() {
        ValueRenderers.registerSuperClassRenderer(Iterable.class, ValueRenderers.Kind.Detailed, c -> ITERABLE_DETAILED);
        ValueRenderers.registerSuperClassRenderer(Map.class, ValueRenderers.Kind.Detailed, c -> MAP_DETAILED);
    }

    private static final MethodHandle MH_withInspectorFontSize = findStatic(MethodHandles.lookup(),
                                                                                             Inspector.class,
                                                                                             "withInspectorFontSize",
                                                                                             methodType(TextWidget.class, TextWidget.class));

    private static <T> MethodHandle makeBuildArrayPreviewStringMH(Class<T> arrayClass) {
        var name = arrayClass.componentType().getName();
        var l = MethodHandles.lookup();
        try {
            var MH = filterReturnValue(arrayLength(arrayClass),
                                       makeConcatWithConstants(l,
                                                               "buildArrayPreview",
                                                               methodType(String.class, int.class),
                                                               name + "[\1]").dynamicInvoker());
            MH = filterReturnValue(MH, TextWidget.MH_ofString);
            MH = filterReturnValue(MH,
                    MH_withInspectorFontSize.asType(MH_withInspectorFontSize.type().changeParameterType(0,
                            TextWidget.MH_ofString.type().returnType())));
            return MH;
        } catch (StringConcatException e) {
            throw new RuntimeException(e);
        }
    }

    public static <T> ValueRenderer<T> arrayPreview(Class<T> c) {
        var MH_buildArrayStringPreview = makeBuildArrayPreviewStringMH(c);
        return (inspector, value) -> TrackingWidgets.trackingIdentity(value, a -> {
            try {
                return (Widget) MH_buildArrayStringPreview.invoke(a);
            } catch (Throwable e) {
                throw new RuntimeException(e);
            }
        });
    }

    public static <T> ValueRenderer<T> arrayDetailed(Class<T> c) {
        var lengthGetter = arrayLength(c).asType(methodType(int.class, Object.class));
        var elementGetter = MethodHandles.arrayElementGetter(c);
        var elementType = elementGetter.type().returnType();
        var debugPropertyValueBuilder = DebugPropertyValue.getMH_of(elementType);
        var getterAndValueBuilder = MethodHandles.filterReturnValue(
                elementGetter, debugPropertyValueBuilder.asType(methodType(DebugPropertyValue.class, elementType))
                ).asType(methodType(DebugPropertyValue.class, Object.class, int.class));
        return (inspector, value) -> TrackingWidgets.trackingIdentity(value, arr -> {
            return Inspector.labeledCol("Elements", mounter -> {
                int l;
                try {
                    l = (int) lengthGetter.invokeExact(arr);
                } catch (Throwable e) {
                    throw new RuntimeException(e);
                }
                for (int i = 0; i < l; i++) {
                    DebugPropertyValue elementValue;
                    try {
                        elementValue = (DebugPropertyValue) getterAndValueBuilder.invokeExact(arr, i);
                    } catch (Throwable e) {
                        throw new RuntimeException(e);
                    }
                    int i_ = i;
                    mounter.layout(intKey(i), TrackingWidgets.tracking(elementValue, v -> {
                        return inspector.buildKeyValueEntry(withInspectorFontSize(text(i_)), v);
                    }));
                }
            });
        });
    }
}
