package dev.pfaff.altargui.io;

import java.io.IOException;
import java.lang.foreign.MemorySegment;
import java.nio.channels.ByteChannel;
import java.nio.channels.SeekableByteChannel;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.Set;

public final class IOUtil {
	private static final Set<OpenOption> DEFAULT_WRITE_OPEN_OPTIONS = Set.of(StandardOpenOption.CREATE,
			StandardOpenOption.TRUNCATE_EXISTING,
			StandardOpenOption.WRITE);

	private static final int BUFFER_SIZE = 8192;

	public static MemorySegment readAllBytes(Reallocator allocator, Path path) throws IOException {
		try (SeekableByteChannel sbc = Files.newByteChannel(path)) {
			return read(allocator, sbc, sbc.size());
		}
	}

	public static void write(Path path, MemorySegment data, OpenOption... options) throws IOException {
		try (SeekableByteChannel sbc = newByteChannelForWrite(path, options)) {
			writeToChannel(data, sbc);
		}
	}

	public static void write(Path path, MemorySegment data, Set<OpenOption> options) throws IOException {
		try (SeekableByteChannel sbc = newByteChannelForWrite(path, options)) {
			writeToChannel(data, sbc);
		}
	}

	private static SeekableByteChannel newByteChannelForWrite(Path path, Set<OpenOption> options) throws IOException {
		return Files.newByteChannel(path, options.isEmpty() ? DEFAULT_WRITE_OPEN_OPTIONS : options);
	}

	private static SeekableByteChannel newByteChannelForWrite(Path path, OpenOption... options) throws IOException {
		return options.length == 0
				? newByteChannelForWrite(path, DEFAULT_WRITE_OPEN_OPTIONS)
				: Files.newByteChannel(path, options);
	}

	private static void writeToChannel(MemorySegment data, SeekableByteChannel sbc) throws IOException {
		var buf = data.asByteBuffer();
		while (buf.hasRemaining()) sbc.write(buf);
	}

	private static MemorySegment read(Reallocator allocator, ByteChannel source, long initialSize) throws IOException {
		long capacity = initialSize;
		MemorySegment buf = allocator.reallocate(MemorySegment.NULL, capacity);
		long nread = 0;
		int n;
		// read to EOF which may read more or less than initialSize (eg: file
		// is truncated while we are reading)

		while ((n = source.read(buf.asSlice(nread, Math.min(buf.byteSize(), Integer.MAX_VALUE)).asByteBuffer())) != -1) {
			nread += n;
			if (nread == buf.byteSize()) {
				// need to allocate a larger buffer
				capacity = Math.max(newLength(capacity,
								1,       /* minimum growth */
								capacity /* preferred growth */),
						BUFFER_SIZE);
				buf = allocator.reallocate(buf, capacity);
			}
		}
		return buf.asSlice(nread);
	}

	private static long newLength(long oldLength, long minGrowth, long prefGrowth) {
		// preconditions not checked because of inlining
		// assert oldLength >= 0
		// assert minGrowth > 0

		long prefLength = oldLength + Math.max(minGrowth, prefGrowth); // might overflow
		if (prefLength >= 0) {
			return prefLength;
		} else {
			throw new OutOfMemoryError(STR."Required array length \{oldLength} + \{minGrowth} is too large");
		}
	}
}
