package dev.pfaff.altargui.io;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SegmentAllocator;
import java.util.Objects;

@FunctionalInterface
public interface Reallocator {
	/**
	 * Reallocates the given segment to be the specified size. Upon returning, the original segment may be invalid.
	 *
	 * @param segment       the segment to reallocate. This must be either {@link MemorySegment#NULL} or the return
	 *                      value
	 *                      of a prior invocation to this method on the same {@link Reallocator} instance.
	 * @param byteSize      the size (in bytes) of the reallocated segment
	 * @param byteAlignment the alignment (in bytes) of the reallocated segment
	 * @return the new segment
	 */
	MemorySegment reallocate(MemorySegment segment, long byteSize, long byteAlignment);

	/**
	 * Like {@link #reallocate(MemorySegment, long, long)} but the {@code byteAlignment} is at least {@code 1}. This
	 * alignment may be increased by a particular {@link Reallocator} implementation).
	 */
	default MemorySegment reallocate(MemorySegment segment, long byteSize) {
		return reallocate(segment, byteSize, 1);
	}

	default Reallocator withAlignment(long byteAlignment) {
		return new AlignedReallocator(this, byteAlignment);
	}

	static Reallocator ofAuto() {
		return (segment, byteSize, byteAlignment) -> Arena.ofAuto().allocate(byteSize, byteAlignment).copyFrom(segment);
	}

	static Reallocator usingArena(Arena arena) {
		return (segment, byteSize, byteAlignment) -> arena.allocate(byteSize, byteAlignment).copyFrom(segment);
	}

	static Reallocator usingAllocator(SegmentAllocator allocator) {
		return (segment, byteSize, byteAlignment) -> allocator.allocate(byteSize, byteAlignment).copyFrom(segment);
	}

	record AlignedReallocator(Reallocator delegate, long byteAlignment) implements Reallocator {
		public AlignedReallocator {
			Objects.requireNonNull(delegate);
			if (Long.bitCount(byteAlignment) != 1 || byteAlignment < 0) {
				throw new IllegalArgumentException("byteAlignment must be a positive power of two");
			}
		}

		@Override
		public MemorySegment reallocate(MemorySegment segment, long byteSize, long byteAlignment) {
			return this.delegate.reallocate(segment, byteSize, Math.max(byteAlignment, this.byteAlignment));
		}

		@Override
		public MemorySegment reallocate(MemorySegment segment, long byteSize) {
			return this.delegate.reallocate(segment, byteSize, this.byteAlignment);
		}
	}
}
