package dev.pfaff.altargui.pathfinder;

import dev.pfaff.glass.Interop;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.function.Consumer;

import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static dev.pfaff.glass.Interop.ADDRESS_INT;
import static dev.pfaff.glass.Interop.MH_cast_usize_to_long;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.methodType;

public final class Pathfinder {
	private static final Linker LINKER = Linker.nativeLinker();
	private static final SymbolLookup LOOKUP = SymbolLookup.libraryLookup(Path.of("libpathfinder.dylib"),
			Arena.ofAuto());
	private static final MethodHandle MH_lookupGLFunction;
	private static final FunctionDescriptor FD_lookupGLFunction = FunctionDescriptor.of(ADDRESS,
			ADDRESS,
			ADDRESS);
	public static final FunctionDescriptor FD_PFGLLoadWith = FunctionDescriptor.ofVoid(ADDRESS,
			ADDRESS);
	public static final ValueLayout ML_PFResult = JAVA_INT.withName("result");

	private static final StructLayout ML_TypeLayout = structLayout(ADDRESS_INT.withName("size"),
			ADDRESS_INT.withName("align")).withName("TypeLayout");

	private static final MethodHandle MH_TypeLayout_size = filterReturnValue(withoutBase(ML_TypeLayout
					.varHandle(groupElement("size"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.GET),
			MH_cast_usize_to_long);
	private static final MethodHandle MH_TypeLayout_align = filterReturnValue(withoutBase(ML_TypeLayout
					.varHandle(groupElement("align"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.GET),
			MH_cast_usize_to_long);

	private static MemoryLayout getOpaqueTypeLayout(String name) {
		var layout = LOOKUP.find("TYPE_LAYOUT_" + name)
				.get()
				.reinterpret(ML_TypeLayout.byteSize())
				.asSlice(0, ML_TypeLayout);
		try {
			var size = (long) MH_TypeLayout_size.invokeExact(layout);
			var align = (long) MH_TypeLayout_align.invokeExact(layout);
			return structLayout(JAVA_BYTE.withByteAlignment(align), sequenceLayout(size - 1, JAVA_BYTE)).withName(name);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static final StructLayout ML_PFVector2I = structLayout(JAVA_INT, JAVA_INT);
	public static final StructLayout ML_PFVector2F = structLayout(JAVA_FLOAT,
			JAVA_FLOAT);
	public static final StructLayout ML_PFRectF = structLayout(ML_PFVector2F, ML_PFVector2F);
	public static final StructLayout ML_PFRectI = structLayout(ML_PFVector2I, ML_PFVector2I);
	public static final AddressLayout ML_PFDestFramebufferRef = ADDRESS.withName("PFDestFramebufferRef");
	public static final MemoryLayout ML_PFGLDevice = getOpaqueTypeLayout("PFGLDevice");
	public static final AddressLayout ML_PFGLDeviceRef = ADDRESS.withTargetLayout(ML_PFGLDevice);
	public static final MemoryLayout ML_PFResourceLoader = getOpaqueTypeLayout("PFResourceLoader");
	public static final AddressLayout ML_PFResourceLoaderRef = ADDRESS.withTargetLayout(ML_PFResourceLoader);
	public static final MemoryLayout ML_PFGLRenderer = getOpaqueTypeLayout("PFGLRenderer");
	public static final AddressLayout ML_PFGLRendererRef = ADDRESS.withTargetLayout(ML_PFGLRenderer);
	public static final MemoryLayout ML_PFCanvas = getOpaqueTypeLayout("PFCanvas");
	public static final AddressLayout ML_PFCanvasRef = ADDRESS.withTargetLayout(ML_PFCanvas);
	public static final MemoryLayout ML_PFFillStyle = getOpaqueTypeLayout("PFFillStyle");
	public static final AddressLayout ML_PFFillStyleRef = ADDRESS.withTargetLayout(ML_PFFillStyle);
	public static final MemoryLayout ML_PFScene = getOpaqueTypeLayout("PFScene");
	public static final AddressLayout ML_PFSceneRef = ADDRESS.withTargetLayout(ML_PFScene);
	public static final MemoryLayout ML_PFSceneProxy = getOpaqueTypeLayout("PFSceneProxy");
	public static final AddressLayout ML_PFSceneProxyRef = ADDRESS.withTargetLayout(ML_PFSceneProxy);
	public static final MemoryLayout ML_PFCanvasFontContext = getOpaqueTypeLayout("PFCanvasFontContext");
	public static final AddressLayout ML_PFCanvasFontContextRef = ADDRESS.withTargetLayout(ML_PFCanvasFontContext);
	public static final MemoryLayout ML_PFColorF = structLayout(JAVA_FLOAT.withName("r"),
			JAVA_FLOAT.withName("g"),
			JAVA_FLOAT.withName("b"),
			JAVA_FLOAT.withName("a"));
	public static final MemoryLayout ML_PFColorU = structLayout(JAVA_BYTE.withName("r"),
			JAVA_BYTE.withName("g"),
			JAVA_BYTE.withName("b"),
			JAVA_BYTE.withName("a"));
	public static final MemoryLayout ML_PFTextMetrics = structLayout(JAVA_FLOAT.withName("width")).withName(
			"PFTextMetrics");
	public static final VarHandle VH_PFTextMetrics_width =
			withoutBase(ML_PFTextMetrics.varHandle(groupElement(
							"width"))
					.withInvokeExactBehavior());
	public static final ValueLayout.OfByte ML_PFRendererLevel = JAVA_BYTE.withName("PFRendererLevel");
	public static final StructLayout ML_PFRendererMode = structLayout(ML_PFRendererLevel).withName("PFRendererMode");
	public static final MemoryLayout ML_PFGLDestFramebuffer = getOpaqueTypeLayout("PFGLDestFramebuffer");
	public static final MemoryLayout ML_PFinternalRenderTransform = getOpaqueTypeLayout("PFinternalRenderTransform");
	public static final MemoryLayout ML_PFinternalVector2F = getOpaqueTypeLayout("PFinternalVector2F");
	public static final StructLayout ML_PFBuildOptions = structLayout(ML_PFinternalRenderTransform,
			ML_PFinternalVector2F.withName("dilation"),
			JAVA_BOOLEAN.withName("subpixel_aa_enabled"));
	public static final AddressLayout ML_PFBuildOptionsRef = ADDRESS.withTargetLayout(ML_PFBuildOptions);

	public static final StructLayout ML_PFRendererOptions = structLayout(ML_PFGLDestFramebuffer.withName("dest"),
			ML_PFColorF.withName("background_color"),
			JAVA_BYTE.withName("flags"));
	public static final AddressLayout ML_PFRendererOptionsRef = ADDRESS.withTargetLayout(ML_PFRendererOptions);

	public static final MemoryLayout ML_FKFont = getOpaqueTypeLayout("FKFont");
	public static final AddressLayout ML_FKFontRef = ADDRESS.withTargetLayout(ML_FKFont);
	public static final MemoryLayout ML_FKCanvasFormat = JAVA_BYTE.withName("FKCanvasFormat");
	public static final MemoryLayout ML_FKCanvasRasterizationOptions = JAVA_BYTE.withName("FKCanvasRasterizationOptions");

	public static final MethodHandle MH_PFGLDestFramebufferInitFullWindow = LINKER.downcallHandle(LOOKUP.find(
					"PFGLDestFramebufferInitFullWindow").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFDestFramebufferRef, ML_PFVector2I));
	public static final MethodHandle MH_PFGLDeviceInit = LINKER.downcallHandle(LOOKUP.find(
					"PFGLDeviceInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFGLDeviceRef, JAVA_BYTE.withName("PFGLVersion"), JAVA_INT));
	public static final MethodHandle MH_PFGLDeviceDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFGLDeviceDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFGLDeviceRef));
	public static final MethodHandle MH_PFEmbeddedResourceLoaderInit = LINKER.downcallHandle(LOOKUP.find(
					"PFEmbeddedResourceLoaderInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFResourceLoaderRef));
	public static final MethodHandle MH_PFFilesystemResourceLoaderLocateInit = LINKER.downcallHandle(LOOKUP.find(
					"PFFilesystemResourceLoaderLocateInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFResourceLoaderRef));
	public static final MethodHandle MH_PFFilesystemResourceLoaderFromPathInit = LINKER.downcallHandle(LOOKUP.find(
					"PFFilesystemResourceLoaderFromPathInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFResourceLoaderRef, ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static final MethodHandle MH_PFResourceLoaderDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFResourceLoaderDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFResourceLoaderRef));
	public static final MethodHandle MH_PFGLRendererInit = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFGLRendererRef,
					ML_PFGLDeviceRef,
					ML_PFResourceLoaderRef,
					ML_PFRendererMode,
					ML_PFRendererOptionsRef));
	public static final MethodHandle MH_PFGLRendererDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFGLRendererRef));
	public static final MethodHandle MH_PFGLRendererDestFramebufferSizeChanged = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererDestFramebufferSizeChanged").get(),
			FunctionDescriptor.ofVoid(ML_PFGLRendererRef));
	public static final MethodHandle MH_PFGLRendererSetViewport = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererSetViewport").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFGLRendererRef, ML_PFRectI.withName("viewport")));
	public static final MethodHandle MH_PFGLRendererSetWindowSize = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererSetWindowSize").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFGLRendererRef, ML_PFVector2I.withName("window_size")));
	public static final MethodHandle MH_PFCanvasFontContextInitWithSystemSource = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFontContextInitWithSystemSource").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFCanvasFontContextRef));
	public static final MethodHandle MH_PFCanvasFontContextDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFontContextDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasFontContextRef));
	public static final MethodHandle MH_PFCanvasInit = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFCanvasRef, ML_PFCanvasFontContextRef));
	public static final MethodHandle MH_PFCanvasSetSize = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetSize").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFVector2F));
	public static final MethodHandle MH_PFCanvasDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef));
	public static final MethodHandle MH_PFCanvasFontContextGetFontByPostScriptName;
	public static final MethodHandle MH_PFCanvasSetFont = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetFont").get(),
			FunctionDescriptor.of(ML_PFResult,
					ML_PFCanvasRef,
					ML_FKFontRef));

	public static final MethodHandle MH_FKFontDeinit = LINKER.downcallHandle(LOOKUP.find(
					"FKFontDeinit").get(),
			FunctionDescriptor.ofVoid(ML_FKFontRef));
	public static final MethodHandle MH_FKFontGlyphForCodepoint = LINKER.downcallHandle(LOOKUP.find(
					"FKFontGlyphForCodepoint").get(),
			FunctionDescriptor.of(JAVA_LONG, ML_FKFontRef, JAVA_INT));
	public static final MethodHandle MH_FKFontRasterizeGlyph = filterArguments(LINKER.downcallHandle(LOOKUP.find(
					"FKFontRasterizeGlyph").get(),
			FunctionDescriptor.of(JAVA_INT,
					ML_FKFontRef,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ADDRESS_INT,
					ML_PFVector2I,
					ML_FKCanvasFormat,
					JAVA_INT.withName("glyph_id"),
					JAVA_FLOAT.withName("point_size"),
					ML_FKCanvasRasterizationOptions)),
			2,
			Interop.MH_cast_long_to_usize);

	static {
		var MH = LINKER.downcallHandle(LOOKUP.find(
						"PFCanvasFontContextGetFontByPostScriptName").get(),
				FunctionDescriptor.of(ML_PFResult,
						ML_PFCanvasFontContextRef,
						ADDRESS.withTargetLayout(JAVA_BYTE),
						ADDRESS_INT,
						ML_FKFontRef));
		MH_PFCanvasFontContextGetFontByPostScriptName = explicitCastArguments(MH,
				MH.type().changeParameterType(2, long.class));
	}

	public static final MethodHandle MH_PFCanvasSetFontSize = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetFontSize").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef,
					JAVA_FLOAT));

	public static final MethodHandle MH_PFCanvasFillText = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFillText").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ADDRESS_INT,
					ML_PFVector2F));
	public static final MethodHandle MH_PFCanvasStrokeText = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasStrokeText").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ADDRESS_INT,
					ML_PFVector2F));
	public static final MethodHandle MH_PFCanvasMeasureText = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasMeasureText").get(),
			FunctionDescriptor.of(ML_PFTextMetrics,
					ML_PFCanvasRef,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ADDRESS_INT));
	public static final MethodHandle MH_PFCanvasFillRect = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFillRect").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFRectF));
	public static final MethodHandle MH_PFCanvasSetFillStyle = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetFillStyle").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFFillStyleRef));
	public static final MethodHandle MH_PFCanvasSetStrokeStyle = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetStrokeStyle").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFFillStyleRef));
	public static final MethodHandle MH_PFCanvasClear = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasClear").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef));

	public static final MethodHandle MH_PFCanvasTakeScene = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasTakeScene").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFSceneRef));
	public static final MethodHandle MH_PFSceneDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneRef));
	public static final MethodHandle MH_PFSceneProxyInitWithSequentialExecutor = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyInitWithSequentialExecutor").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFSceneProxyRef, ML_PFRendererLevel));
	public static final MethodHandle MH_PFSceneProxyInitWithRayonExecutor = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyInitWithRayonExecutor").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFSceneProxyRef, ML_PFRendererLevel));
	public static final MethodHandle MH_PFSceneProxyReplaceScene = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyReplaceScene").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneProxyRef, ML_PFSceneRef));
	public static final MethodHandle MH_PFSceneProxyDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneProxyRef));
	public static final MethodHandle MH_PFSceneProxyBuildAndRenderGL = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyBuildAndRenderGL").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneProxyRef,
					ML_PFGLRendererRef,
					ML_PFBuildOptionsRef.withName("build_options")));
	public static final MethodHandle MH_PFBuildOptionsInit = LINKER.downcallHandle(LOOKUP.find(
					"PFBuildOptionsInit").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFBuildOptionsRef.withName("build_options")));

	public static final MethodHandle MH_PFFillStyleInitColor = LINKER.downcallHandle(LOOKUP.find(
					"PFFillStyleInitColor").get(),
			FunctionDescriptor.of(ML_PFResult, ML_PFFillStyleRef, ML_PFColorU));
	public static final MethodHandle MH_PFFillStyleDeinit = LINKER.downcallHandle(LOOKUP.find(
					"PFFillStyleDeinit").get(),
			FunctionDescriptor.ofVoid(ML_PFFillStyleRef));

	static {
		var l = MethodHandles.lookup();
		try {
			MH_lookupGLFunction = dropArguments(l.findStatic(Pathfinder.class,
					"lookupGLFunction",
					methodType(MemorySegment.class, SymbolLookup.class, MemorySegment.class)), 2, MemorySegment.class);
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}

		try {
			LINKER.downcallHandle(FunctionDescriptor.ofVoid())
					.invokeExact(LOOKUP.find("init_logging").get());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private static MemorySegment lookupGLFunction(SymbolLookup symbolLookup, MemorySegment namePtr) {
		// set a reasonable max length, because the segment we get has length 0
		var name = namePtr.reinterpret(128).getString(0, StandardCharsets.US_ASCII);
		return symbolLookup.find(name).orElse(MemorySegment.NULL);
	}

	private Pathfinder() {}

	public static void initialize(SymbolLookup openGLSymbolLookup) {
		try (var arena = Arena.ofConfined()) {
			var glContextStub = LINKER.upcallStub(MH_lookupGLFunction.bindTo(openGLSymbolLookup),
					FD_lookupGLFunction,
					arena);
			var pfglLoadWith = LINKER.downcallHandle(FD_PFGLLoadWith);
			pfglLoadWith.invokeExact(LOOKUP.find("PFGLLoadWith").get(), glContextStub, MemorySegment.NULL);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	@FunctionalInterface
	public interface Constructor {
		int construct(MemorySegment segment) throws Throwable;
	}

	@FunctionalInterface
	public interface Destructor extends Consumer<MemorySegment> {
		void destruct(MemorySegment segment) throws Throwable;

		@Override
		default void accept(MemorySegment segment) {
			try {
				destruct(segment);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}
	}

	public static MemorySegment constructNativeObject(MemoryLayout layout, Constructor constructor,
													  Destructor destructor) {
		var arena = Arena.ofAuto();
		var segment = arena.allocate(layout);
		int result;
		try {
			result = constructor.construct(segment);
		} catch (Throwable e) {
			throw new RuntimeException(STR."Unable to construct \{layout}", e);
		}
		if (result != 0) throw new RuntimeException(STR."Unable to construct \{layout}: error code \{result}");
		return manageNativeObject(arena, segment, destructor);
	}

	public static MemorySegment manageNativeObject(Arena arena, MemorySegment segment, Destructor destructor) {
		return segment.reinterpret(arena, destructor);
	}

	public static PFGLDeviceRef createDevice(PFGLVersion glVersion) {
		return new PFGLDeviceRef(constructNativeObject(ML_PFGLDevice,
				ptr -> (int) MH_PFGLDeviceInit.invokeExact(ptr, glVersion.value(), 0),
				ptr -> MH_PFGLDeviceDeinit.invokeExact(ptr)));
	}

	public static PFResourceLoaderRef createEmbeddedResourceLoader() {
		return new PFResourceLoaderRef(constructNativeObject(ML_PFResourceLoader,
				ptr -> (int) MH_PFEmbeddedResourceLoaderInit.invokeExact(ptr),
				ptr -> MH_PFResourceLoaderDeinit.invokeExact(ptr)));
	}

	public static PFResourceLoaderRef createFilesystemResourceLoaderLocate() {
		return new PFResourceLoaderRef(constructNativeObject(ML_PFResourceLoader,
				ptr -> (int) MH_PFFilesystemResourceLoaderLocateInit.invokeExact(ptr),
				ptr -> MH_PFResourceLoaderDeinit.invokeExact(ptr)));
	}

	public static PFResourceLoaderRef createFilesystemResourceLoaderFromPath(Path path) {
		return new PFResourceLoaderRef(constructNativeObject(ML_PFResourceLoader,
				ptr -> {
					try (var arena = Arena.ofConfined()) {
						var pathPtr = arena.allocateFrom(path.toString());
						return (int) MH_PFFilesystemResourceLoaderFromPathInit.invokeExact(ptr, pathPtr);
					} catch (Throwable e) {
						throw new RuntimeException(e);
					}
				},
				ptr -> MH_PFResourceLoaderDeinit.invokeExact(ptr)));
	}

	public static PFCanvasFontContextRef createSystemFontContext() {
		try {
			return new PFCanvasFontContextRef(constructNativeObject(ML_PFCanvasFontContext,
					ptr -> (int) MH_PFCanvasFontContextInitWithSystemSource.invokeExact(ptr),
					ptr -> MH_PFCanvasFontContextDeinit.invokeExact(ptr)));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static PFCanvasRef createCanvas(PFCanvasFontContextRef fontContext) {
		return new PFCanvasRef(constructNativeObject(ML_PFCanvas,
				ptr -> (int) MH_PFCanvasInit.invokeExact(ptr, fontContext.ptr()),
				ptr -> MH_PFCanvasDeinit.invokeExact(ptr)));
	}

	public static PFBuildOptionsRef createBuildOptions() {
		return new PFBuildOptionsRef(constructNativeObject(ML_PFBuildOptions,
				ptr -> (int) MH_PFBuildOptionsInit.invokeExact(ptr),
				_ -> {
				}));
	}
}
