package dev.pfaff.altargui.render;

import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.shader.BuiltInstanceLayout;
import dev.pfaff.altargui.uploadables.Uploadable;

import java.io.Closeable;

import static dev.pfaff.altargui.internal.Config.GL_CHECKS;

public final class Buffer implements Closeable {
	public final RawBuffer raw;
	public final BuiltInstanceLayout instanceDef;

	public Buffer(BuiltInstanceLayout instanceDef, RawBuffer raw) {
		this.instanceDef = instanceDef;
		this.raw = raw;
	}

	/**
	 * @param capacity the required capacity in instances.
	 * @param vao      the shader of the VAO to bind if the buffer is recreated, or {@code -1} if binding is not
	 *                    necessary.
	 */
	public void reset(int capacity, int vao) {
		raw.reset(capacity * instanceDef.stride(), () -> {
			if (vao != -1) GLUtil.glBindVertexArray(vao);
			try {
				instanceDef.bindAttribs();
			} finally {
				if (Config.GL_ALWAYS_UNBIND && vao != -1) GLUtil.glxUnbindVertexArray();
			}
		});
	}

	/**
	 * Begins uploading to the buffer. Attempts to upload more than {@code capacity} might result in a
	 * {@link BufferOverflowException} being thrown.
	 * <p>
	 * The buffer's contents are undefined after this operation.
	 */
	public void beginUploading() {
		raw.beginUploading();
	}

	public void finishUploading() {
		raw.finishUploading();
	}

	/**
	 * @param addit the number of additional bytes.
	 */
	private void checkUpload(int addit) {
		// check that the additional bytes form whole elements (no trailing bytes)
		if (GL_CHECKS) {
			int additElements = addit / instanceDef.stride();
			int additElementsBytes = additElements * instanceDef.stride();
			if (additElementsBytes != addit) {
				throw new IllegalArgumentException(STR."Buffer has trailing bytes: \{addit} bytes forms \{additElements} elements but leaves \{addit - additElementsBytes} trailing bytes");
			}
		}
	}

	public void uploadAt(long offset, Uploadable uploadable) {
		checkUpload(uploadable.size());
		raw.upload(uploadable);
	}

	public void upload(Uploadable uploadable) {
		checkUpload(uploadable.size());
		raw.upload(uploadable);
	}

	public int instanceCount() {
		return raw.length() / instanceDef.stride();
	}

	public int instanceCapacity() {
		return raw.capacity() / instanceDef.stride();
	}

	@Override
	public String toString() {
		return STR."\{getClass().getSimpleName()}[instanceDef=\{instanceDef}, raw=\{raw}]";
	}

	@Override
	public void close() {
		raw.close();
	}
}
