package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.internal.util.GrowableMemorySegment;
import dev.pfaff.altargui.shader.Program;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.GLException;
import dev.pfaff.glass.opengl.enums.GLErrorCode;
import dev.pfaff.glass.opengl.enums.GLPrimitiveType;

import java.lang.foreign.MemorySegment;
import java.lang.foreign.StructLayout;
import java.util.List;
import java.util.stream.Stream;

import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.JAVA_INT;

/**
 * Uses {@link dev.pfaff.glass.opengl.GL#glDrawArraysIndirect(GLPrimitiveType, ConstPointer)}.
 */
public final class GL42DrawArraysIndirectLayeredRenderBuffer extends LayeredRenderBuffer {
	private static final StructLayout ML_DrawArraysIndirectCommand = structLayout(JAVA_INT.withName("count"),
			JAVA_INT.withName("instanceCount"),
			JAVA_INT.withName("first"),
			JAVA_INT.withName("baseInstance"));

	// TODO: use explicit attribute locations so that we can use a single VAO for everything!
	private final List<VAO> vaos = Stream.generate(VAO::new).limit(this.programs.size()).toList();

	private final GrowableMemorySegment indirectBuffer = new GrowableMemorySegment(1024);
	private long indirectBufferWriteOffset = 0L;
	private long indirectBufferReadOffset = 0L;

	public GL42DrawArraysIndirectLayeredRenderBuffer() {
	}

	@Override
	public void beginUploading(int capacity) {
		super.beginUploading(capacity);
		this.indirectBufferWriteOffset = 0L;
		this.indirectBufferReadOffset = 0L;
	}

	@Override
	protected void onRecreate() {
		var vaosIter = this.vaos.iterator();
		var layoutsIter = this.instanceLayouts.iterator();
		while (vaosIter.hasNext()) {
			@SuppressWarnings("resource") var vao = vaosIter.next();
			var layout = layoutsIter.next();
			try (var _ = vao.bound()) {
				layout.bindAttribs();
			}
		}
	}

	@Override
	public void close() {
		this.vaos.forEach(AbstractVertexArrayObject::close);
		super.close();
	}

	@Override
	protected void putDataForEntry(MemorySegment ms, int programId, int start, int end) {
		int instanceCount = end - start;
		var layerInfo = this.programs.get(programId);
		var offset = this.indirectBufferWriteOffset;
		this.indirectBuffer.ensureTotalCapacity(this.indirectBufferWriteOffset += ML_DrawArraysIndirectCommand.byteSize());
		var indirectCmd = this.indirectBuffer.ms().asSlice(offset, ML_DrawArraysIndirectCommand);
		indirectCmd.setAtIndex(JAVA_INT, 0, 0);
		indirectCmd.setAtIndex(JAVA_INT, 1, layerInfo.templateVertexCount());
		indirectCmd.setAtIndex(JAVA_INT, 2, instanceCount);
		indirectCmd.setAtIndex(JAVA_INT, 3, start);
	}

	@Override
	protected VAO getVAOForEntry(MemorySegment ms, int programId) {
		return this.vaos.get(programId);
	}

	@Override
	protected void drawEntry(MemorySegment ms, int programId, Program program) {
		var drawMode = program.drawMode();
		if (this.indirectBufferReadOffset >= this.indirectBufferWriteOffset) {
			throw new BufferOverflowException(STR."\{this.indirectBufferReadOffset} >= \{this.indirectBufferWriteOffset}");
		}
		try {
			var offset = this.indirectBufferReadOffset;
			this.indirectBufferReadOffset = offset + ML_DrawArraysIndirectCommand.byteSize();
			// while this is supported on OpenGL 4.0, baseInstance is not until OpenGL 4.2
			GuiGlobals.GL.glDrawArraysIndirect(drawMode.glPrimitiveType(),
					new ConstPointer<>(this.indirectBuffer.ms().asSlice(offset)));
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_OPERATION) {
				throw new RuntimeException(STR."Either a geometry shader is active or one of the buffers (\{this}) is mapped and not persistent", e);
			} else {
				throw e;
			}
		}
	}
}
