package dev.pfaff.altargui.render;

import dev.pfaff.altargui.host.ContextForker;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.glass.opengl.GL;

import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.stream.IntStream;

import static dev.pfaff.altargui.GuiGlobals.HOST;
import static dev.pfaff.altargui.internal.Config.GL_CONTEXT_POOL_SIZE;

/**
 * A pool of OpenGL worker threads, each with their own context with sharing enabled with the main context.
 */
public abstract sealed class GLContextPool {
	public static ScopedValue<GL> POOL_ACQUIRED_GL = ScopedValue.newInstance();

	public abstract Executor acquireExecutor();

	/**
	 * The global context pool.
	 */
	public static final GLContextPool GL_CONTEXT_POOL;

	static {
		var forker = HOST.contextForker();
		if (forker != null) {
			GL_CONTEXT_POOL = new PerCore(forker);
		} else {
			GL_CONTEXT_POOL = new OnMainThread();
		}
	}

	private static final class OnMainThread extends GLContextPool {
		@Override
		public Executor acquireExecutor() {
			return GLThread::submitToGLThread;
		}
	}

	private static final class PerCore extends GLContextPool {
		private final List<LinkedBlockingQueue<Runnable>> threadTasks;

		public PerCore(ContextForker forker) {
			this.threadTasks = IntStream.range(0, GL_CONTEXT_POOL_SIZE)
					.mapToObj(i -> {
						var queue = new LinkedBlockingQueue<Runnable>();
						Thread.ofPlatform().name(STR."gl-worker-\{i}").start(() -> {
							try (var ctx = forker.forkContext()) {
								ScopedValue.runWhere(POOL_ACQUIRED_GL, ctx.gl(), () -> {
									for (; ; ) {
										Runnable task;
										try {
											task = queue.take();
										} catch (InterruptedException e) {
											throw new RuntimeException(e);
										}
										try {
											task.run();
										} catch (Throwable e) {

										}
									}
								});
							}
						});
						return queue;
					})
					.toList();
		}

		@Override
		public Executor acquireExecutor() {
			LinkedBlockingQueue<Runnable> queue = Objects.requireNonNull(this.threadTasks.stream()
					.unordered()
					.min(Comparator.comparing(LinkedBlockingQueue::size))
					.orElse(null));
			return e -> {
				try {
					queue.put(e);
				} catch (InterruptedException ex) {
					throw new RuntimeException(ex);
				}
			};
		}
	}
}
