package dev.pfaff.altargui.render;

import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.text.StyleFlags;

import java.lang.foreign.StructLayout;
import java.lang.invoke.VarHandle;
import java.nio.ByteOrder;
import java.util.List;

import static dev.pfaff.altargui.render.BufferUtil.emitShaderColor;
import static dev.pfaff.altargui.render.BufferUtil.emitVec2;
import static dev.pfaff.altargui.render.BufferUtil.emitVec2AsNormUByte;
import static dev.pfaff.altargui.text.Font.CHARS_PER_ROWCOL;
import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.JAVA_BYTE;
import static java.lang.foreign.ValueLayout.JAVA_FLOAT;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class TextRenderLayerUtil {
	public static final float DEFAULT_SIZE = 8f;
	public static final Color DEFAULT_COLOR = Color.fromRgba(0xf0f0f0_ff);
	public static final int DEFAULT_STYLE = StyleFlags.REGULAR;

	public static final StructLayout TEXT_CHAR_LAYOUT = structLayout(
			structLayout(JAVA_FLOAT.withName("x"), JAVA_FLOAT.withName("y")).withName("pos"),
			JAVA_FLOAT.withName("size"),
			// bvec4 must be laid out as RGBA. To write an RGBA int and have it be in the correct order, we must write
			// it in BIG_ENDIAN order.
			JAVA_INT.withName("colorRgba").withOrder(ByteOrder.BIG_ENDIAN),
			JAVA_BYTE.withName("char"),
			JAVA_BYTE.withName("flags"),
			JAVA_BYTE.withName("UBase"),
			JAVA_BYTE.withName("VBase")
			);
	public static final VarHandle VH_posX = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("pos"), groupElement("x"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_posY = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("pos"), groupElement("y"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_size = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("size"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_color = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("colorRgba"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_char = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("char"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_flags = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("flags"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_uBase = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
			groupElement("UBase"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_vBase = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("VBase"))
			.withInvokeExactBehavior());

	public static final float BOLD_ADVANCE = 1.0f / Font.PIXELS_PER_CHAR;
	public static final float SPACING = 1.0f / Font.PIXELS_PER_CHAR;

	public static final List<Character> TEX_U;
	public static final List<Character> TEX_V;

	static {
		Character[] u = new Character[16];
		Character[] v = new Character[16];

		for (int i = 0; i < 256; i++) {
			int ui = i % CHARS_PER_ROWCOL;
			int vi = i / CHARS_PER_ROWCOL;
			var uf = (float)ui / CHARS_PER_ROWCOL;
			var vf = (float)vi / CHARS_PER_ROWCOL;
			u[ui] = (char)(uf * Character.MAX_VALUE);
			v[vi] = (char)(vf * Character.MAX_VALUE);
		}

		TEX_U = List.of(u);
		TEX_V = List.of(v);
	}

	public static char getU(byte character) {
		return TEX_U.get(Byte.toUnsignedInt(character) % CHARS_PER_ROWCOL);
	}

	public static char getV(byte character) {
		return TEX_V.get(Byte.toUnsignedInt(character) / CHARS_PER_ROWCOL);
	}

	public static void emitTextInstance(Sink sink, float x, float y, float size, byte character, Color color, int flags, float uBase, float vBase) {
		emitVec2(sink, x, y);
		sink.accept(size);
		emitShaderColor(sink, color);
		sink.accept(character);
		sink.accept((byte)flags);
		emitVec2AsNormUByte(sink, uBase, vBase);
	}
}
