package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.shader.InstanceLayout;
import dev.pfaff.altargui.shader.ShaderProg;
import dev.pfaff.altargui.shader.VertexAttribSpec;
import dev.pfaff.altargui.uploadables.Uploadable;
import dev.pfaff.glass.opengl.enums.*;

import java.nio.ByteOrder;
import java.util.List;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.Shaders.GENERIC_INSTANCED_SHADER;
import static dev.pfaff.altargui.shader.VertexAttrib.DIVISOR_PER_VERTEX;
import static dev.pfaff.log4truth.StandardTags.DEBUG;

/**
 * A reusable buffer for unified rendering. Every rendered element is made up of triangles. The
 * entire scene is emitted to a single buffer and then rendered with a single draw call.
 */
public final class UnifiedRenderBuffer extends AbstractVertexArrayObject {
	public static final InstanceLayout VERTEX_SPEC = new InstanceLayout(List.of(
			new VertexAttribSpec("Position",
					GLVertexAttribPointerType.FLOAT,
					2)
			, new VertexAttribSpec("UV",
					GLVertexAttribPointerType.UNSIGNED_SHORT,
					2)
					.withNormalized()
			, new VertexAttribSpec("Layer",
					GLVertexAttribPointerType.UNSIGNED_INT,
					1)
			, new VertexAttribSpec("Color",
					GLVertexAttribPointerType.UNSIGNED_BYTE,
					4)
					.withNormalized().withByteOrder(ByteOrder.BIG_ENDIAN)
	), DIVISOR_PER_VERTEX);

	static {
		if (VERTEX_SPEC.stride() != Uploadable.VERTEX_LAYOUT.byteSize()) {
			throw new AssertionError(STR."\{VERTEX_SPEC.stride()} != \{Uploadable.VERTEX_LAYOUT.byteSize()}:\n  spec: \{VERTEX_SPEC}\n  layout: \{Uploadable.VERTEX_LAYOUT}");
		}
	}

	private final ShaderProg shader;
	public final Buffer verticesBuffer;

	public UnifiedRenderBuffer() {
		this.shader = GENERIC_INSTANCED_SHADER.get();
		bind();
		try {
			this.verticesBuffer = new Buffer(VERTEX_SPEC.build(shader), RawBuffer.create(GLBufferTarget.ARRAY_BUFFER, true));
		} finally {
			unbind();
		}

		LOGGER.log(() -> STR."built vertices buffer: \{this.verticesBuffer}", List.of(DEBUG));
	}

//	/**
//	 * Appends a new template of the given float offsets.
//	 * @param offsets a sequence of pairs of floats (X,Y).
//	 * @return the index of the start of the template.
//	 */
//	public int uploadTemplate(float[] offsets) {
//		if (offsets.length % 2 != 0) {
//			throw new IllegalArgumentException("Expected an array of xy pairs, found an extra float");
//		}
//		templateBuffer.beginUploading();
//		int i = templateBuffer.length();
//		templateBuffer.upload(offsets.length, sink -> {
//			sink.accept(offsets);
//		});
//		templateBuffer.finishUploading();
//		return i;
//	}

	public void beginUploadingInstances(int capacity) {
		verticesBuffer.reset(capacity, vertexArrayId);
		verticesBuffer.beginUploading();
	}

	public void finishUploadingInstances() {
		verticesBuffer.finishUploading();
	}

	/**
	 * Uploads an instance to the buffer. {@link #beginUploadingInstances(int)} must have been
	 * called already.
	 */
	public void uploadInstance(Uploadable uploadable) {
		verticesBuffer.upload(uploadable);
	}

	public void preDraw() {
		GL.glEnable(GLEnableCap.BLEND);
		GL.glBlendFuncSeparate(GLBlendingFactor.SRC_ALPHA,
				GLBlendingFactor.ONE_MINUS_SRC_ALPHA,
				GLBlendingFactor.ONE,
				GLBlendingFactor.ZERO);
		GL.glDisable(GLEnableCap.DEPTH_TEST);

		this.bind();

		// TODO: texture arrays
//			GL.glxBindTexture(GL13.GL_TEXTURE0, Resources.MINECRAFT_ASCII_FONT_TEXTURE);

		this.shader.bind();
		GuiGlobals.GL.checkError();
	}

	public void postDraw() {
		this.unbind();
	}

	/**
	 * This calls {@link #preDraw()} and {@link #postDraw()} for you.
	 */
	public void draw() {
		this.preDraw();
		this.drawRange(0, this.verticesBuffer.instanceCount());
		this.postDraw();
	}

	public void drawRange(int first, int count) {
		GLUtil.glDrawArrays(GLPrimitiveType.TRIANGLES, first, count);
	}

	@Override
	public void close() {
		super.close();
		verticesBuffer.close();
	}

	public int instanceCount() {
		return verticesBuffer.instanceCount();
	}

	public int byteCount() {
		return verticesBuffer.instanceCount() * verticesBuffer.instanceDef.stride();
	}
}
