package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.render.*;
import dev.pfaff.glass.opengl.enums.GLBufferAccess;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;

import java.lang.foreign.MemoryLayout;
import java.util.Objects;
import java.util.function.Consumer;

/**
 * Might be optimal for large amounts of data. Testing will be needed to confirm this and establish a rough threshold.
 */
public final class StaticGPUBuffer extends RenderObject {
	private final int program;
	private final int id;
	private final MemoryLayout instanceLayout;
	private final int instanceCount;

	public StaticGPUBuffer(int program, int id, MemoryLayout instanceLayout, int instanceCount) {
		this.program = program;
		this.id = id;
		this.instanceLayout = instanceLayout;
		this.instanceCount = instanceCount;
	}

	public static StaticGPUBuffer build(int layer, MemoryLayout instanceLayout, int instanceCount, Consumer<? super Sink> builder) {
		int length = (int) instanceLayout.byteSize() * instanceCount;

		int id = GLUtil.glGenBuffers();
		GuiGlobals.GL.glBindBuffer(GLBufferTarget.ARRAY_BUFFER, id);

		var buffer = GLUtil.glMapBuffer(GLBufferTarget.ARRAY_BUFFER, GLBufferAccess.WRITE_ONLY, length);
		var sink = new MemorySegmentSink(buffer);

		builder.accept(sink);

		if (sink.offset() != length) {
			throw new RuntimeException(STR."Buffer size not equal to allocated size: expected \{length}, found \{sink.offset()}");
		}

		GLUtil.glUnmapBuffer(GLBufferTarget.ARRAY_BUFFER);

		if (Config.GL_ALWAYS_UNBIND) GuiGlobals.GL.glBindBuffer(GLBufferTarget.ARRAY_BUFFER, 0);

		return new StaticGPUBuffer(layer, id, instanceLayout, instanceCount);
	}

	@Override
	public void close() {
		GLUtil.glDeleteBuffers(id);
	}

	@Override
	public void uploadTo(Sink sink) {
		sink.accept(this);
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return program;
	}

	public int id() {
		return id;
	}

	@Override
	public MemoryLayout instanceLayout() {
		return instanceLayout;
	}

	@Override
	public int instanceCount() {
		return instanceCount;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == this) return true;
		if (obj == null || obj.getClass() != this.getClass()) return false;
		var that = (StaticGPUBuffer) obj;
		return this.program == that.program &&
			   this.id == that.id &&
			   Objects.equals(this.instanceLayout, that.instanceLayout) &&
			   this.instanceCount == that.instanceCount;
	}

	@Override
	public int hashCode() {
		int result = 1;
		result = 31 * result + Integer.hashCode(program);
		result = 31 * result + Integer.hashCode(id);
		result = 31 * result + instanceLayout.hashCode();
		result = 31 * result + Integer.hashCode(instanceCount);
		return result;
	}

	@Override
	public String toString() {
		return STR."StaticGPUBuffer[program=\{program}, id=\{id}, instanceLayout=\{instanceLayout}, instanceCount=\{instanceCount}]";
	}

}
