package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.render.BufferOverflowException;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.RenderObject;
import dev.pfaff.altargui.render.ViaIntermediateBuffer;
import dev.pfaff.altargui.texture.TextureAllocator;

import java.lang.foreign.Arena;
import java.lang.foreign.MemoryLayout;
import java.lang.foreign.MemorySegment;
import java.util.function.Consumer;

import static dev.pfaff.altargui.render.TextRenderLayerUtil.*;
import static dev.pfaff.altargui.shader.Programs.textRenderLayer;

/**
 * A multi-character text render object with individual styling.
 */
public final class TextRenderObject extends RenderObject {
	private float x, y;
	private MemorySegment ms;
	private int length;

	public TextRenderObject(int capacity) {
		this.allocate(capacity);
	}

	private void allocate(int capacity) {
		capacity = Integer.highestOneBit((capacity - 1) | 1) << 1;
		this.ms = Arena.ofAuto().allocate(TEXT_CHAR_LAYOUT, capacity);
	}

	public void updatePosition(float x, float y) {
		var offX = x - this.x;
		var offY = y - this.y;
		this.x = x;
		this.y = y;
		ms.spliterator(TEXT_CHAR_LAYOUT).forEachRemaining(charMs -> {
			VH_posX.set(charMs, 0L, (float)VH_posX.get(charMs, 0L) + offX);
			VH_posY.set(charMs, 0L, (float)VH_posY.get(charMs, 0L) + offY);
		});
	}

	public void updateTex(float u, float v) {
//		Logger.log(() -> FMT."Updating tex coords to \{u}, \{v}");

		// the shader accepts base coordinates [0, 256)
		final var shaderTexUnit = 256;
		// this is set elsewhere
		var textureSize = TextureAllocator.MAX_TEXTURE_SIZE;
		var requiredBaseAlignment = Math.divideExact(textureSize, shaderTexUnit);
		var actualBaseAlignment = TextureAllocator.UNIT;
		if (requiredBaseAlignment > actualBaseAlignment) {
			throw new AssertionError(STR."base alignment is less than required: \{actualBaseAlignment} < \{requiredBaseAlignment}");
		}
		assert (int) (u * shaderTexUnit) == u * shaderTexUnit;
		assert (int) (v * shaderTexUnit) == v * shaderTexUnit;

		// convert from [0, 1) to [0, 256)
		var uB = (byte)(int)(u * (float)shaderTexUnit);
		var vB = (byte)(int)(v * (float)shaderTexUnit);

		ms.spliterator(TEXT_CHAR_LAYOUT).forEachRemaining(charMs -> {
			VH_uBase.set(charMs, 0L, uB);
			VH_vBase.set(charMs, 0L, vB);
		});
	}

	public void updateCharacters(int length, Provider provider) {
		if (length > ms.byteSize() / TEXT_CHAR_LAYOUT.byteSize()) allocate(length);
		var ms = this.ms;

		var xBase = this.x;
		var yBase = this.y;
		var sink = new Sink() {
			private int i;

			@Override
			public void accept(byte character, int color, int flags, float x, float y, float size) {
				int i = this.i;
				if (i >= length) throw new BufferOverflowException("Overflowed specified length");
				VH_posX.set(ms, (long)i, xBase + x);
				VH_posY.set(ms, (long)i, yBase + y);
				VH_char.set(ms, (long)i, character);
//				Logger.log(() -> FMT."Accepted '\{(char)Byte.toUnsignedInt(character)}' (0x%02x\{Byte.toUnsignedInt(character)})");
				VH_color.set(ms, (long)i, color);
				VH_flags.set(ms, (long)i, (byte)flags);
				VH_size.set(ms, (long)i, size);
				this.i = i + 1;
			}
		};
		provider.accept(sink);

		this.length = length;
	}

	@Override
	public int instanceCount() {
		return length;
	}

	@Override
	public int size() {
		return super.size();
	}

	@Override
	public MemoryLayout instanceLayout() {
		return TEXT_CHAR_LAYOUT;
	}

	@Override
	public void uploadTo(dev.pfaff.altargui.render.Sink sink) {
		sink.accept(ms.asSlice(0, TEXT_CHAR_LAYOUT.byteSize() * length));
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return textRenderLayer();
	}

	@Override
	public void close() {
	}

	@Override
	public String toString() {
		var sb = new StringBuilder("TextRenderObject[x=");
		sb.append(x);
		sb.append(", y=");
		sb.append(y);
		sb.append(", length=");
		sb.append(length);
		sb.append(", ms=");
		sb.append(ms);
		sb.append(", characters=[");
		var upr = ms.byteSize() / TEXT_CHAR_LAYOUT.byteSize();
		for (int i = 0; i < upr; i++) {
			if (i == 0) {
				sb.append('\'');

			}
			sb.append((char)Byte.toUnsignedInt((byte)VH_char.get(ms, (long)i)));
			if (i == upr - 1) {
				sb.append('\'');
			} else {
				sb.append("', '");
			}
		}
		sb.append("]]");
		return sb.toString();
	}

	@FunctionalInterface
	public interface Provider extends Consumer<Sink> {
		@Override
		void accept(Sink sink);
	}

	@FunctionalInterface
	public interface Sink {
		/**
		 * @param color RGBA
		 * @param flags {@link dev.pfaff.altargui.text.StyleFlags style flags}
		 * @param x the x-offset of the character relative to the render object's position.
		 * @param y the y-offset of the character relative to the render object's position.
		 * @param size the size of the character.
		 */
		void accept(byte character, int color, int flags, float x, float y, float size);
	}
}
