package dev.pfaff.altargui.selector;

import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.util.OpenArrayList;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.util.StringCursor;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.util.List;
import java.util.function.Consumer;
import java.util.function.Function;

/**
 * A selector for an element in the tree. Similar to CSS selectors.
 */
public interface Selector {
	/**
	 * @param element the element that the selector is being applied to.
	 */
	void select(@Nullable WidgetInstance element, Consumer<@NonNull WidgetInstance> consumer);

	default @Nullable WidgetInstance selectFirst(@Nullable WidgetInstance element) {
		var consumer = new Consumer<WidgetInstance>() {
			private @Nullable WidgetInstance selected;

			@Override
			public void accept(WidgetInstance o) {
				if (selected == null) selected = o;
			}
		};
		select(element, consumer);
		return consumer.selected;
	}

	default @Nullable WidgetInstance selectSingle(@Nullable WidgetInstance element) {
		var consumer = new Consumer<WidgetInstance>() {
			private boolean hasSelected;
			private @Nullable WidgetInstance<?> selected;

			@Override
			public void accept(WidgetInstance o) {
				if (this.hasSelected) {
					this.selected = null;
					return;
				}
				this.selected = o;
				this.hasSelected = true;
			}
		};
		select(element, consumer);
		return consumer.selected;
	}

	/**
	 * Selects all matching elements. The returned list is mutable and owned by the caller.
	 */
	default List<@NonNull WidgetInstance<?>> selectAll(@Nullable WidgetInstance<?> element) {
		var selected = OpenArrayList.<WidgetInstance<?>>wrap(WidgetInstance.EMPTY_ARRAY);
		select(element, selected::add);
		return selected;
	}

	static Selector nthElement(int i, @Nullable Selector next) {
		return new NthElement(i, next);
	}

	static Selector nthElement(int i) {
		return nthElement(i, null);
	}

	static Selector children(@Nullable Selector next) {
		return new Children(next);
	}

	static Selector children() {
		return children(null);
	}

	static Selector childrenRecursive(@Nullable Selector next) {
		return new ChildrenRecursive(next);
	}

	static Selector childrenRecursive() {
		return childrenRecursive(null);
	}

	static Selector single(@Nullable Selector next) {
		return new Single(next);
	}

	static Selector onlyChild(@Nullable Selector next) {
		return single(children(next));
	}

	static Selector onlyChild() {
		return onlyChild(null);
	}

	static Selector isClass(@NonNull Class<? extends Widget> clazz, @Nullable Selector next) {
		return new IsClass(clazz, next);
	}

	static Selector isClass(@NonNull Class<? extends Widget> clazz) {
		return isClass(clazz, null);
	}

	private static @NonNull String parseFunction(StringCursor cursor) throws SyntaxError {
		if (!cursor.advanceMatches("(")) throw new SyntaxError("Expected '(' at " + cursor);
		var s = cursor.advanceUntilAndSkip(')');
		if (s == null) throw new SyntaxError("Expected arguments (if any) followed by ')'");
		return s;
	}

	static Selector parse(StringCursor cursor) throws SyntaxError {
		// builder stack
		OpenArrayList<Function<@Nullable Selector, Selector>> stack = OpenArrayList.wrap(new Function[0]);

		while (cursor.hasRemaining()) {
			cursor.skipWhitespace();
			if (cursor.matches("{")) break;

			if (cursor.advanceMatches(">>")) {
				cursor.skipWhitespace();
				stack.push(Selector::childrenRecursive);
			} else if (cursor.advanceMatches(">")) {
				cursor.skipWhitespace();
				stack.push(Selector::children);
			} else if (cursor.advanceMatches("#")) {
				var id = cursor.advanceUntilOrEnd(c -> c == ' ' || c == ':');
				stack.push(next -> new HasId(id, next));
			} else if (cursor.advanceMatches(".")) {
				throw new UnsupportedOperationException("TODO: parse tag selector");
			} else if (cursor.advanceMatches(":")) {
				switch (cursor.advanceUntilOrEnd(c -> c == ' ' || c == '(' || c == ':')) {
					case null -> throw new SyntaxError("Expected a selector name at " + cursor);
					default -> throw new SyntaxError("Unrecognized selector name at " + cursor);
				}
			} else {
				var name = cursor.advanceUntilOrEnd(c -> c == ' ' || c == '(' || c == ':');
				switch (name) {
					case "nth-child" -> {
						var s = parseFunction(cursor);
						int n;
						try {
							n = Integer.parseInt(s);
						} catch (NumberFormatException e) {
							throw new SyntaxError("Expected a valid integer", e);
						}
						stack.push(next -> nthElement(n, next));
					}
					case "is-class" -> {
						var s = parseFunction(cursor);
						Class<?> clazz;
						try {
							clazz = Class.forName(s);
						} catch (ClassNotFoundException e) {
							throw new SyntaxError("Expected a valid class name at " + cursor, e);
						}
						if (!Widget.class.isAssignableFrom(clazz)) {
							throw new SyntaxError("Expected a Widget class at " + cursor);
						}
						//noinspection unchecked
						stack.push(next -> isClass((Class<? extends Widget>) clazz, next));
					}
					case "single" -> {
						stack.push(Selector::single);
					}
					case null -> throw new SyntaxError("Expected a selector name at " + cursor);
					default -> throw new SyntaxError("Unrecognized selector name at " + cursor);
				}
			}
		}

		if (stack.isEmpty()) {
			throw new SyntaxError("Expected a selector name at " + cursor);
		}

		Selector selector = null;
		do {
			selector = stack.pop().apply(selector);
		} while (!stack.isEmpty());
		return selector;
	}
}
