package dev.pfaff.altargui.shader;

import dev.pfaff.altargui.resource.ResourceLoader;
import dev.pfaff.altargui.texture.TextureAllocator;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.GLVersion;
import dev.pfaff.glass.opengl.MutPointer;
import dev.pfaff.glass.opengl.enums.GLShaderParameterName;
import dev.pfaff.glass.opengl.enums.GLShaderType;
import dev.pfaff.log4truth.Logger;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.nio.charset.StandardCharsets;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.regex.MatchResult;
import java.util.regex.Pattern;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.internal.InternalGlobals.GL_FEATURES;
import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static dev.pfaff.altargui.util.ThreadFuture.spawnVirtualTask;
import static java.lang.foreign.ValueLayout.*;

/**
 * Loads shader source code, handling any necessary preprocessing.
 */
public final class ShaderLoader {
	private static final String IMPORT_PATTERN_PREFIX = "#import <";
	private static final String IMPORT_PATTERN_SUFFIX = ">";
	private static final Pattern IMPORT_PATTERN = Pattern.compile(IMPORT_PATTERN_PREFIX + "[^>]+?" + IMPORT_PATTERN_SUFFIX);
	private static final Pattern AUTO_VERSION_PATTERN = Pattern.compile("#auto_version");
	private static final Pattern ATLAS_SIZE_PATTERN = Pattern.compile("#atlas_size");

	private final ResourceLoader loader;
	private final ConcurrentHashMap<String, BasicFuture<String>> cache = new ConcurrentHashMap<>();

	public ShaderLoader(ResourceLoader loader) {
		this.loader = loader;
	}

	public void clearCache() {
		this.cache.clear();
	}

	public Shader loadSync(String name, GLShaderType type) throws IOException {
		var source = loadSource(name).join();
		return executeOn(() -> {
			var shader = new Shader(GL.glCreateShader(type));
			try (var arena = Arena.ofConfined()) {
				var sourcePtr = arena.allocateFrom(source, StandardCharsets.US_ASCII);
				var sourcesPtr = arena.allocateFrom(ADDRESS, sourcePtr);
				var lenPtr = arena.allocateFrom(JAVA_INT, (int) sourcePtr.byteSize());
				GL.glShaderSource(shader.handle(), 1, new ConstPointer<>(sourcesPtr), new ConstPointer<>(lenPtr));
			}
			GL.glCompileShader(shader.handle());
			try (var arena = Arena.ofConfined()) {
				var result = arena.allocate(JAVA_INT);
				GL.glGetShaderiv(shader.handle(), GLShaderParameterName.COMPILE_STATUS, new MutPointer<>(result));
				if (result.get(JAVA_INT, 0) == 0) {
					var buf = arena.allocate(JAVA_BYTE, 32768);
					var len = arena.allocate(JAVA_INT);
					GL.glGetShaderInfoLog(shader.handle(),
							(int) buf.byteSize(),
							new MutPointer<>(len),
							new MutPointer<>(buf));
					var infoLog = buf.getString(0, StandardCharsets.US_ASCII);
					var sb = new StringBuilder()
							.append("Couldn't compile ")
							.append(type)
							.append(" shader (")
							.append(name)
							.append("):\n")
							.append(infoLog)
							.append("\nSource code:");
					var start = 0;
					var lineN = 0;
					var max = (int)source.chars().filter(c -> c == '\n').count() + 1;
					var lineNPadding = Integer.toString(max-1).length();
					while (start < source.length()) {
						var end = source.indexOf('\n', start);
						if (end == -1) end = source.length();
						var lineNStr = Integer.toString(++lineN);
						sb.append("\n")
								.repeat(' ', lineNPadding - lineNStr.length())
								.append(lineNStr)
								.append(" │ ")
								.append(source, start, end);
						start = end + 1;
					}
					throw new IOException(sb.toString());
				}
			}
			return shader;
		}, GLThread::submitToGLThread);
	}

	public BasicFuture<String> loadSource(String name) {
		return this.cache.computeIfAbsent(name, name_ -> spawnVirtualTask(() -> {
			Logger.log(() -> STR."Loading shader '\{name_}'");
			try {
				var resource = Objects.requireNonNull(this.loader.resolveResource(name_));
				var source = resource.readString(StandardCharsets.UTF_8);
				source = preprocess(source);
				return source;
			} catch (Throwable e) {
				throw new RuntimeException(STR."Failed to load shader '\{name}'", e);
			}
		}));
	}

	private String preprocess(String source) {
		source = AUTO_VERSION_PATTERN.matcher(source).replaceAll(GL_FEATURES.glVersion.api() == GLVersion.API.OpenGL
				? "#version 150"
				: """
						#version 310 es
						
						precision mediump float;
						precision mediump int;
						precision mediump sampler2D;
						precision mediump sampler2DArray;""");
		source = ATLAS_SIZE_PATTERN.matcher(source).replaceAll(Integer.toString(TextureAllocator.MAX_TEXTURE_SIZE));

		// the final source code prior to expanding imports
		final var finalSource = source;

		var matcher = IMPORT_PATTERN.matcher(finalSource);
		var tasks = matcher.results()
				.map(result -> loadSource(finalSource.substring(result.start() + IMPORT_PATTERN_PREFIX.length(),
						result.end() - IMPORT_PATTERN_SUFFIX.length())))
				.toList();
		return matcher.replaceAll(new Function<>() {
			private int i = 0;

			@Override
			public String apply(MatchResult result) {
				return tasks.get(this.i++).join();
			}
		});
	}
}
