package dev.pfaff.altargui.text;

import dev.pfaff.altargui.host.Host;
import dev.pfaff.altargui.resource.ClassResourceLoadableResource;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.texture.ResourceLoadableTexture;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SequenceLayout;
import java.lang.foreign.StructLayout;
import java.lang.invoke.VarHandle;
import java.util.Objects;

import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.PathElement.sequenceElement;
import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.JAVA_FLOAT;

public final class MinecraftBitmapFont implements Font {
	private static final StructLayout DATA_ENTRY_LAYOUT = structLayout(
			JAVA_FLOAT.withName("minX"),
			JAVA_FLOAT.withName("maxX")
	);
	private static final SequenceLayout DATA_LAYOUT = sequenceLayout(256, DATA_ENTRY_LAYOUT);
	private static final VarHandle VH_minX = withoutBase(DATA_LAYOUT.varHandle(sequenceElement(),
			groupElement("minX"))
			.withInvokeExactBehavior());
	private static final VarHandle VH_maxX = withoutBase(DATA_LAYOUT.varHandle(sequenceElement(),
					groupElement("maxX"))
			.withInvokeExactBehavior());

	public static final MinecraftBitmapFont MINECRAFT_ASCII =
			new MinecraftBitmapFont(new ResourceLoadableTexture(new ClassResourceLoadableResource(
			Host.class,
			"/assets/altar-gui/fonts/minecraft_ascii.png")));

	private final @NotNull LoadableTexture texture;
	private final MemorySegment data = Arena.ofAuto().allocate(DATA_LAYOUT);
	private boolean loaded;

	public MinecraftBitmapFont(@NotNull LoadableTexture texture) {
		this.texture = Objects.requireNonNull(texture);
	}

	@Override
	public LoadableTexture texture() {
		return this.texture;
	}

	@Override
	public float getMinX(char character) {
		byte b = (byte) character;
		if (Byte.toUnsignedInt(b) != character) b = FALLBACK_CHAR;
		return (float) VH_minX.get(this.data, Byte.toUnsignedLong(b));
	}

	@Override
	public float getMaxX(char character) {
		byte b = (byte) character;
		if (Byte.toUnsignedInt(b) != character) b = FALLBACK_CHAR;
		return (float) VH_maxX.get(this.data, Byte.toUnsignedLong(b));
	}

	@Override
	public void reload() throws IOException {
		// TODO: define an sdf file format and read that instead of converting every time
		var imageData = this.texture.load();
		try (var arena = Arena.ofConfined()) {
			var sdf = SDFFontUtil.transparentAndWhiteToSDF(imageData, arena);
			int pixelsPerChar = Math.divideExact(sdf.width(), 16);
			for (int i = 0; i < 256; i++) {
				int startX = (i % 16) * pixelsPerChar;
				int startY = (i / 16) * pixelsPerChar;
				// plus 1 because we want the *end* of the pixel.
				// return value is relative to startX
				float endX = sdf.findCharacterEndX(pixelsPerChar, pixelsPerChar, startX, startY) +
						   ((float) pixelsPerChar / PIXELS_PER_CHAR);
				float minX = 0f;
				float maxX = i == ' ' ? 0.5f : endX / pixelsPerChar;
				VH_minX.set(this.data, (long)i, minX);
				VH_maxX.set(this.data, (long)i, maxX);
			}
		}
		if (!this.loaded) this.loaded = true;
	}

	@Override
	public boolean isLoaded() {
		return this.loaded;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == this) return true;
		if (obj == null || obj.getClass() != this.getClass()) return false;
		var that = (MinecraftBitmapFont) obj;
		return this.texture.equals(that.texture);
	}

	@Override
	public int hashCode() {
		return this.texture.hashCode();
	}

	@Override
	public String toString() {
		return STR."MinecraftByteFont[texture=\{this.texture}]";
	}
}
