package dev.pfaff.altargui.uploadables;

import dev.pfaff.altargui.render.Sink;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.Bits;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SequenceLayout;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.VarHandle;
import java.util.List;

import static dev.pfaff.altargui.render.BufferUtil.packChars;
import static dev.pfaff.altargui.render.BufferUtil.packFloats;
import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static java.lang.foreign.MemoryLayout.*;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.PathElement.sequenceElement;
import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.*;

public final class UploadableRect implements Uploadable {
	public static final List<Float> OFFSETS = List.of(
			0f, 1f, // Top-left
			1f, 1f, // Top-right
			1f, 0f, // Bottom-right

			1f, 0f, // Bottom-right
			0f, 0f, // Bottom-left
			0f, 1f // Top-left
	);
	public static final char OFFSET_BITS;

	static {
		char offsetBits = 0;
		for (int i = OFFSETS.size(); i >= 0; i--) {
			offsetBits <<= 1;
			offsetBits |= 1;
		}
		OFFSET_BITS = offsetBits;
	}

	public static final int TRIANGLES = OFFSETS.size() / 2;

	private static final SequenceLayout LAYOUT = sequenceLayout(TRIANGLES, VERTEX_LAYOUT)
			.withName("Rect")
			.withByteAlignment(Long.BYTES);
	private static final MethodHandle MH_setVertexPosition_I = explicitCastArguments(
			withoutBase(LAYOUT.varHandle(sequenceElement(),
							groupElement("position"),
							groupElement("packed"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.SET),
			methodType(void.class, MemorySegment.class, int.class, long.class)
	);
	private static final MethodHandle MH_setVertexUV_I = explicitCastArguments(
			withoutBase(LAYOUT.varHandle(sequenceElement(),
							groupElement("uv"),
							groupElement("packed"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.SET),
			methodType(void.class, MemorySegment.class, int.class, int.class)
	);
	private static final MethodHandle MH_setVertexTexLayer_I = explicitCastArguments(
			withoutBase(LAYOUT.varHandle(sequenceElement(),
							groupElement("texLayer"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.SET),
			methodType(void.class, MemorySegment.class, int.class, char.class)
	);
	private static final MethodHandle MH_setVertexColor_I = explicitCastArguments(
			withoutBase(LAYOUT.varHandle(sequenceElement(),
							groupElement("colorRgba"))
					.withInvokeExactBehavior())
					.toMethodHandle(VarHandle.AccessMode.SET),
			methodType(void.class, MemorySegment.class, int.class, int.class)
	);

	private final MemorySegment ms;

	public UploadableRect(Arena arena) {
		this.ms = arena.allocate(LAYOUT);
	}

	public void updateCorners(float x1, float y1, float x2, float y2) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexPosition_I.invokeExact(ms, i, packFloats(
						Bits.bit(OFFSET_BITS, i * 2) ? x2 : x1,
						Bits.bit(OFFSET_BITS, i * 2 + 1) ? y2 : y1
				));
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void updatePositionSize(float x, float y, float width, float height) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexPosition_I.invokeExact(ms, i, packFloats(
						Math.fma(OFFSETS.get(i * 2), width, x),
						Math.fma(OFFSETS.get(i * 2 + 1), height, y)
				));
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void updateUVExtent(float u, float v, float extentU, float extentV) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexUV_I.invokeExact(ms, i, packChars(
						(char) (Math.fma(OFFSETS.get(i * 2), extentU, u) * Character.MAX_VALUE),
						(char) (Math.fma(OFFSETS.get(i * 2 + 1), extentV, v) * Character.MAX_VALUE)
				));
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void updateUV(float u1, float v1, float u2, float v2) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexUV_I.invokeExact(ms, i, packChars(
						(char) ((Bits.bit(OFFSET_BITS, i * 2) ? u2 : u1) * Character.MAX_VALUE),
						(char) ((Bits.bit(OFFSET_BITS, i * 2 + 1) ? v2 : v1) * Character.MAX_VALUE)
				));
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void updateTexLayer(int layer) {
		char layerC = (char) layer;
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexTexLayer_I.invokeExact(ms, i, layerC);
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @param color RGBA
	 */
	public void updateColor(Color color) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				MH_setVertexColor_I.invokeExact(ms, i, color.rgba());
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @param startColor RGBA
	 * @param endColor   RGBA
	 */
	// TODO: make this more flexible
	public void updateColorVerticalGradient(int startColor, int endColor) {
		try {
			for (int i = 0; i < TRIANGLES; i++) {
				boolean isEndY = Bits.bit(OFFSET_BITS, i * 2 + 1);
				MH_setVertexColor_I.invokeExact(ms, i, isEndY ? endColor : startColor);
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public int size() {
		return (int) LAYOUT.byteSize();
	}

	@Override
	public void uploadTo(Sink sink) {
		sink.accept(ms);
	}
}
