package dev.pfaff.altargui.util;

import java.util.Objects;
import java.util.concurrent.*;
import java.util.function.Consumer;
import java.util.function.Function;

import static dev.pfaff.altargui.util.ThrowUtil.throwUnchecked;

public interface BasicFuture<V> {
	static BasicFuture COMPLETED_NULL = () -> null;

	V join() throws CompletionException, CancellationException;

	default void cancel() {}

	default BasicFuture<V> cancelOnInterrupt() {
		var self = this;
		return new CancelOnInterruptFuture<>(self);
	}

	/**
	 * Spawns a new virtual thread that joins this future and does not attempt to catch exceptions. If one is thrown,
	 * the default uncaught exception handler for the VM will be invoked.
	 *
	 * @return this future.
	 */
	default BasicFuture<V> uncatch() {
		Thread.startVirtualThread(this::join);
		return this;
	}

	default <U> BasicFuture<U> map(Function<V, U> mapper) {
		Objects.requireNonNull(mapper);
		return ThreadFuture.spawnVirtualTask(() -> mapper.apply(join()));
	}

	default <U> BasicFuture<U> flatMap(Function<V, BasicFuture<U>> mapper) {
		Objects.requireNonNull(mapper);
		return ThreadFuture.spawnVirtualTask(() -> mapper.apply(join()).cancelOnInterrupt().join());
	}

	default BasicFuture<Void> thenAccept(Consumer<V> consumer) {
		Objects.requireNonNull(consumer);
		return ThreadFuture.spawnVirtualTask(() -> consumer.accept(join()));
	}

	public static <V> BasicFuture<V> completedTask(V value) {
		// this nullability warning is wrong
		//noinspection ConstantValue
		if (value == null) {
			//noinspection unchecked
			return COMPLETED_NULL;
		}
		return () -> value;
	}

	public static <V> BasicFuture<V> failedTask(Throwable exception) {
		Objects.requireNonNull(exception);
		return () -> {
			throw new CompletionException(exception);
		};
	}

	public static <V> BasicFuture<V> wrap(Future<V> future) {
		Objects.requireNonNull(future);
		return new BasicFuture<>() {
			@Override
			public V join() throws CompletionException, CancellationException {
				try {
					return future.get();
				} catch (InterruptedException e) {
					throw throwUnchecked(e);
				} catch (ExecutionException e) {
					throw new CompletionException(e.getCause());
				}
			}

			@Override
			public void cancel() {
				future.cancel(true);
			}
		};
	}

	public static <V> BasicFuture<V> wrap(CompletableFuture<V> future) {
		return new BasicFuture<V>() {
			@Override
			public V join() throws CompletionException, CancellationException {
				return future.join();
			}

			@Override
			public void cancel() {
				future.cancel(true);
			}
		};
	}

	public static <V> BasicFuture<V> spawnTaskOn(Callable<V> callable, Executor executor) {
		Objects.requireNonNull(callable);
		return wrap(CompletableFuture.supplyAsync(() -> {
			try {
				return callable.call();
			} catch (RuntimeException | Error e) {
				throw e;
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}, executor));
	}

	public static <V> V executeOn(Callable<V> callable, Executor executor) {
		Objects.requireNonNull(callable);
		return CompletableFuture.supplyAsync(() -> {
			try {
				return callable.call();
			} catch (RuntimeException | Error e) {
				throw e;
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}, executor).join();
	}

	public static void executeOn(Runnable runnable, Executor executor) {
		Objects.requireNonNull(runnable);
		CompletableFuture.runAsync(() -> {
			try {
				runnable.run();
			} catch (RuntimeException | Error e) {
				throw e;
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}, executor).join();
	}
}
