package dev.pfaff.altargui.util;

import java.lang.invoke.CallSite;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MutableCallSite;

import static java.lang.invoke.MethodHandles.constant;

/**
 * A wrapper for a boolean value that is stored and accessed via a {@link MutableCallSite}. The JVM should be able to
 * optimize uses of the value more than if it were simply a non-final field.
 * <p>
 * From my testing, it appears that the Hotspot JVM has more aggressive speculative optimization of {@link CallSite}s
 * than it does non-final fields. That knowledge led to the creation of this wrapper class. Note that this technique
 * will <strong>only</strong> be effective if the Hotspot JVM is run with {@code -XX:+TrustFinalNonStaticFields}.
 */
public final class FakeConstantBoolean {
	private static final MethodHandle MH_false = constant(boolean.class, false);
	private static final MethodHandle MH_true = constant(boolean.class, true);

	private final MutableCallSite CS_value;

	/**
	 * The initial value will be {@code false}.
	 */
	public FakeConstantBoolean() {
		this(false);
	}

	/**
	 * @param value the initial value
	 */
	public FakeConstantBoolean(boolean value) {
		this.CS_value = new MutableCallSite(value ? MH_true : MH_false);
	}

	/**
	 * Stores a new value. The call-site will only be modified if the new value differs from the stored value.
	 *
	 * @param value the new value
	 */
	public void set(boolean value) {
		try {
			// guarded because the cost of updating a call-site *likely* far outweighs the cost of the check.
			if (value != this.get()) {
				this.CS_value.setTarget(value ? MH_true : MH_false);
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @return the stored value
	 */
	public boolean get() {
		try {
			return (boolean) this.CS_value.getTarget().invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
