package dev.pfaff.altargui.util;

import dev.pfaff.altargui.WidgetInstance;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.HashCommon;
import it.unimi.dsi.fastutil.longs.*;
import it.unimi.dsi.fastutil.objects.AbstractObjectCollection;
import it.unimi.dsi.fastutil.objects.AbstractObjectSet;
import it.unimi.dsi.fastutil.objects.ObjectCollection;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectSpliterator;
import it.unimi.dsi.fastutil.objects.ObjectSpliterators;

import java.util.Arrays;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.LongConsumer;

import static it.unimi.dsi.fastutil.HashCommon.arraySize;
import static it.unimi.dsi.fastutil.HashCommon.maxFill;

public final class NonNullLong2WidgetInstanceHashMap extends AbstractLong2ObjectMap<WidgetInstance<?>>
	implements Cloneable, Hash {

	/** The array of keys. */
	private long[] key;
	/** The array of values. */
	private WidgetInstance<?>[] value;
	/** The mask for wrapping a position counter. */
	private int mask;
	/** Whether this map contains the key zero. */
	private boolean containsNullKey;
	/** The current table size. */
	private int n;
	/**
	 * Threshold after which we rehash. It must be the table size times {@link #DEFAULT_LOAD_FACTOR}.
	 */
	private int maxFill;
	/**
	 * We never resize below this threshold, which is the construction-time {#n}.
	 */
	private final int minN;
	/** Number of entries in the set (including the key zero, if present). */
	private int size;
	/** Cached set of entries. */
	private final MapEntrySet entries = new MapEntrySet();
	/** Cached set of keys. */
	private final KeySet keys = new KeySet();
	/** Cached collection of values. */
	private final ValueCollection values = new ValueCollection();
	/**
	 * Creates a new hash map.
	 *
	 * <p>
	 * The actual table size will be the least power of two greater than
	 * {@code expected}/{@link Hash#DEFAULT_LOAD_FACTOR DEFAULT_LOAD_FACTOR}.
	 *
	 * @param expected
	 *            the expected number of elements in the hash map.
	 */
	@SuppressWarnings("unchecked")
	public NonNullLong2WidgetInstanceHashMap(final int expected) {
		if (expected < 0)
			throw new IllegalArgumentException("The expected number of elements must be nonnegative");
		minN = n = arraySize(expected, DEFAULT_LOAD_FACTOR);
		mask = n - 1;
		maxFill = maxFill(n, DEFAULT_LOAD_FACTOR);
		key = new long[n + 1];
		value = new WidgetInstance[n + 1];
	}

	private int realSize() {
		return containsNullKey ? size - 1 : size;
	}
	private void ensureCapacity(final int capacity) {
		final int needed = arraySize(capacity, DEFAULT_LOAD_FACTOR);
		if (needed > n)
			rehash(needed);
	}
	private void tryCapacity(final long capacity) {
		final int needed = (int) Math.min(1 << 30,
										  Math.max(2, HashCommon.nextPowerOfTwo((long) Math.ceil(capacity / DEFAULT_LOAD_FACTOR))));
		if (needed > n)
			rehash(needed);
	}
	private WidgetInstance<?> removeEntry(final int pos) {
		var oldValue = value[pos];
		value[pos] = null;
		size--;
		shiftKeys(pos);
		if (n > minN && size < maxFill / 4 && n > DEFAULT_INITIAL_SIZE)
			rehash(n / 2);
		return oldValue;
	}
	private WidgetInstance<?> removeNullEntry() {
		containsNullKey = false;
		var oldValue = value[n];
		value[n] = null;
		size--;
		if (n > minN && size < maxFill / 4 && n > DEFAULT_INITIAL_SIZE)
			rehash(n / 2);
		return oldValue;
	}
	@Override
	public void putAll(Map<? extends Long, ? extends WidgetInstance<?>> m) {
		if (DEFAULT_LOAD_FACTOR <= .5)
			ensureCapacity(m.size()); // The resulting map will be sized for m.size() elements
		else
			tryCapacity(size() + m.size()); // The resulting map will be tentatively sized for size() + m.size()
		// elements
		super.putAll(m);
	}

	private int find(final long k) {
		if (((k) == (0)))
			return containsNullKey ? n : -(n + 1);
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return -(pos + 1);
		if (((k) == (curr)))
			return pos;
		// There's always an unused entry.
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return -(pos + 1);
			if (((k) == (curr)))
				return pos;
		}
	}
	private void insert(final int pos, final long k, final WidgetInstance<?> v) {
		if (pos == n)
			containsNullKey = true;
		key[pos] = k;
		value[pos] = v;
		if (size++ >= maxFill)
			rehash(arraySize(size + 1, DEFAULT_LOAD_FACTOR));
	}
	@Override
	public WidgetInstance<?> put(final long k, final WidgetInstance<?> v) {
		final int pos = find(k);
		if (pos < 0) {
			insert(-pos - 1, k, v);
			return defRetValue;
		}
		var oldValue = value[pos];
		value[pos] = v;
		return oldValue;
	}
	/**
	 * Shifts left entries with the specified hash code, starting at the specified
	 * position, and empties the resulting free entry.
	 *
	 * @param pos
	 *            a starting position.
	 */
	protected final void shiftKeys(int pos) {
		// Shift entries with the same hash.
		int last, slot;
		long curr;
		final long[] key = this.key;
		for (;;) {
			pos = ((last = pos) + 1) & mask;
			for (;;) {
				if (((curr = key[pos]) == (0))) {
					key[last] = (0);
					value[last] = null;
					return;
				}
				slot = (int) it.unimi.dsi.fastutil.HashCommon.mix((curr)) & mask;
				if (last <= pos ? last >= slot || slot > pos : last >= slot && slot > pos)
					break;
				pos = (pos + 1) & mask;
			}
			key[last] = curr;
			value[last] = value[pos];
		}
	}
	@Override

	public WidgetInstance<?> remove(final long k) {
		if (((k) == (0))) {
			if (containsNullKey)
				return removeNullEntry();
			return defRetValue;
		}
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return defRetValue;
		if (((k) == (curr)))
			return removeEntry(pos);
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return defRetValue;
			if (((k) == (curr)))
				return removeEntry(pos);
		}
	}
	@Override

	public WidgetInstance<?> get(final long k) {
		if (((k) == (0)))
			return containsNullKey ? value[n] : defRetValue;
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return defRetValue;
		if (((k) == (curr)))
			return value[pos];
		// There's always an unused entry.
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return defRetValue;
			if (((k) == (curr)))
				return value[pos];
		}
	}
	@Override

	public boolean containsKey(final long k) {
		if (((k) == (0)))
			return containsNullKey;
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return false;
		if (((k) == (curr)))
			return true;
		// There's always an unused entry.
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return false;
			if (((k) == (curr)))
				return true;
		}
	}
	@Override
	public boolean containsValue(final Object v) {
		var value = this.value;
		var key = this.key;
		if (containsNullKey && Objects.equals(value[n], v))
			return true;
		for (int i = n; i-- != 0;)
			if (!((key[i]) == (0)) && Objects.equals(value[i], v))
				return true;
		return false;
	}
	/** {@inheritDoc} */
	@Override

	public WidgetInstance<?> getOrDefault(final long k, final WidgetInstance<?> defaultValue) {
		if (((k) == (0)))
			return containsNullKey ? value[n] : defaultValue;
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return defaultValue;
		if (((k) == (curr)))
			return value[pos];
		// There's always an unused entry.
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return defaultValue;
			if (((k) == (curr)))
				return value[pos];
		}
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> putIfAbsent(final long k, final WidgetInstance<?> v) {
		final int pos = find(k);
		if (pos >= 0)
			return value[pos];
		insert(-pos - 1, k, v);
		return defRetValue;
	}
	/** {@inheritDoc} */
	@Override

	public boolean remove(final long k, final Object v) {
		if (((k) == (0))) {
			if (containsNullKey && Objects.equals(v, value[n])) {
				removeNullEntry();
				return true;
			}
			return false;
		}
		long curr;
		final long[] key = this.key;
		int pos;
		// The starting point.
		if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
			return false;
		if (((k) == (curr)) && Objects.equals(v, value[pos])) {
			removeEntry(pos);
			return true;
		}
		while (true) {
			if (((curr = key[pos = (pos + 1) & mask]) == (0)))
				return false;
			if (((k) == (curr)) && Objects.equals(v, value[pos])) {
				removeEntry(pos);
				return true;
			}
		}
	}
	/** {@inheritDoc} */
	@Override
	public boolean replace(final long k, final WidgetInstance<?> oldValue, final WidgetInstance<?> v) {
		final int pos = find(k);
		if (pos < 0 || !Objects.equals(oldValue, value[pos]))
			return false;
		value[pos] = v;
		return true;
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> replace(final long k, final WidgetInstance<?> v) {
		final int pos = find(k);
		if (pos < 0)
			return defRetValue;
		var oldValue = value[pos];
		value[pos] = v;
		return oldValue;
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> computeIfAbsent(final long k, final java.util.function.LongFunction<? extends WidgetInstance<?>> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		final int pos = find(k);
		if (pos >= 0)
			return value[pos];
		var newValue = mappingFunction.apply(k);
		insert(-pos - 1, k, newValue);
		return newValue;
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> computeIfAbsent(final long key, final Long2ObjectFunction<? extends WidgetInstance<?>> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		final int pos = find(key);
		if (pos >= 0)
			return value[pos];
		if (!mappingFunction.containsKey(key))
			return defRetValue;
		var newValue = mappingFunction.get(key);
		insert(-pos - 1, key, newValue);
		return newValue;
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> computeIfPresent(final long k,
							  final java.util.function.BiFunction<? super Long, ? super WidgetInstance<?>, ? extends WidgetInstance<?>> remappingFunction) {
		Objects.requireNonNull(remappingFunction);
		final int pos = find(k);
		if (pos < 0)
			return defRetValue;
		var newValue = remappingFunction.apply(Long.valueOf(k), (value[pos]));
		if (newValue == null) {
			if (((k) == (0)))
				removeNullEntry();
			else
				removeEntry(pos);
			return defRetValue;
		}
		return value[pos] = (newValue);
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> compute(final long k,
					 final java.util.function.BiFunction<? super Long, ? super WidgetInstance<?>, ? extends WidgetInstance<?>> remappingFunction) {
		Objects.requireNonNull(remappingFunction);
		final int pos = find(k);
		var newValue = remappingFunction.apply(Long.valueOf(k), pos >= 0 ? (value[pos]) : null);
		if (newValue == null) {
			if (pos >= 0) {
				if (((k) == (0)))
					removeNullEntry();
				else
					removeEntry(pos);
			}
			return defRetValue;
		}
		if (pos < 0) {
			insert(-pos - 1, k, newValue);
			return newValue;
		}
		return value[pos] = newValue;
	}
	/** {@inheritDoc} */
	@Override
	public WidgetInstance<?> merge(final long k, final WidgetInstance<?> v,
				   final java.util.function.BiFunction<? super WidgetInstance<?>, ? super WidgetInstance<?>, ? extends WidgetInstance<?>> remappingFunction) {
		Objects.requireNonNull(remappingFunction);
		final int pos = find(k);
		if (pos < 0 || value[pos] == null) {
			if (v == null)
				return defRetValue;
			insert(-pos - 1, k, v);
			return v;
		}
		final WidgetInstance<?> newValue = remappingFunction.apply((value[pos]), (v));
		if (newValue == null) {
			if (((k) == (0)))
				removeNullEntry();
			else
				removeEntry(pos);
			return defRetValue;
		}
		return value[pos] = (newValue);
	}
	/*
	 * Removes all elements from this map.
	 *
	 * <p>To increase object reuse, this method does not change the table size. If
	 * you want to reduce the table size, you must use {@link #trim()}.
	 *
	 */
	@Override
	public void clear() {
		if (size == 0)
			return;
		size = 0;
		containsNullKey = false;
		Arrays.fill(key, (0));
		Arrays.fill(value, null);
	}
	@Override
	public int size() {
		return size;
	}
	@Override
	public boolean isEmpty() {
		return size == 0;
	}

	public void forEachValue(WidgetInstanceConsumer consumer) {
		for (var value : this.value) {
			if (value != null) {
				consumer.accept(value);
			}
		}
	}

	/**
	 * The entry class for a hash map does not record key and value, but rather the
	 * position in the hash table of the corresponding entry. This is necessary so
	 * that calls to {@link java.util.Map.Entry#setValue(Object)} are reflected in
	 * the map
	 */
	final class MapEntry implements Long2ObjectMap.Entry<WidgetInstance<?>>, Map.Entry<Long, WidgetInstance<?>>, LongObjectPair<WidgetInstance<?>> {
		// The table index this entry refers to, or -1 if this entry has been deleted.
		int index;
		MapEntry(final int index) {
			this.index = index;
		}
		MapEntry() {
		}
		@Override
		public long getLongKey() {
			return key[index];
		}
		@Override
		public long leftLong() {
			return key[index];
		}
		@Override
		public WidgetInstance<?> getValue() {
			return value[index];
		}
		@Override
		public WidgetInstance<?> right() {
			return value[index];
		}
		@Override
		public WidgetInstance<?> setValue(final WidgetInstance<?> v) {
			var oldValue = value[index];
			value[index] = v;
			return oldValue;
		}
		@Override
		public LongObjectPair<WidgetInstance<?>> right(final WidgetInstance<?> v) {
			value[index] = v;
			return this;
		}
		/**
		 * {@inheritDoc}
		 *
		 * @deprecated Please use the corresponding type-specific method instead.
		 */
		@Deprecated
		@Override
		public Long getKey() {
			return Long.valueOf(key[index]);
		}
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals(final Object o) {
			if (!(o instanceof Map.Entry))
				return false;
			Map.Entry<Long, WidgetInstance<?>> e = (Map.Entry<Long, WidgetInstance<?>>) o;
			return ((key[index]) == ((e.getKey()).longValue()))
				   && Objects.equals(value[index], (e.getValue()));
		}
		@Override
		public int hashCode() {
			return it.unimi.dsi.fastutil.HashCommon.long2int(key[index])
				   ^ ((value[index]) == null ? 0 : (value[index]).hashCode());
		}
		@Override
		public String toString() {
			return key[index] + "=>" + value[index];
		}
	}
	/** An iterator over a hash map. */
	private abstract class MapIterator<ConsumerType> {
		/**
		 * The index of the last entry returned, if positive or zero; initially,
		 * {@link #n}. If negative, the last entry returned was that of the key of index
		 * {@code - pos - 1} from the {@link #wrapped} list.
		 */
		int pos = n;
		/**
		 * The index of the last entry that has been returned (more precisely, the value
		 * of {@link #pos} if {@link #pos} is positive, or {@link Integer#MIN_VALUE} if
		 * {@link #pos} is negative). It is -1 if either we did not return an entry yet,
		 * or the last returned entry has been removed.
		 */
		int last = -1;
		/** A downward counter measuring how many entries must still be returned. */
		int c = size;
		/**
		 * A boolean telling us whether we should return the entry with the null key.
		 */
		boolean mustReturnNullKey = NonNullLong2WidgetInstanceHashMap.this.containsNullKey;
		/**
		 * A lazily allocated list containing keys of entries that have wrapped around
		 * the table because of removals.
		 */
		LongArrayList wrapped;
		@SuppressWarnings("unused")
		abstract void acceptOnIndex(final ConsumerType action, final int index);
		public boolean hasNext() {
			return c != 0;
		}
		public int nextEntry() {
			if (!hasNext())
				throw new NoSuchElementException();
			c--;
			if (mustReturnNullKey) {
				mustReturnNullKey = false;
				return last = n;
			}
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			for (;;) {
				if (--pos < 0) {
					// We are just enumerating elements from the wrapped list.
					last = Integer.MIN_VALUE;
					final long k = wrapped.getLong(-pos - 1);
					int p = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask;
					while (!((k) == (key[p])))
						p = (p + 1) & mask;
					return p;
				}
				if (!((key[pos]) == (0)))
					return last = pos;
			}
		}
		public void forEachRemaining(final ConsumerType action) {
			if (mustReturnNullKey) {
				mustReturnNullKey = false;
				acceptOnIndex(action, last = n);
				c--;
			}
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			while (c != 0) {
				if (--pos < 0) {
					// We are just enumerating elements from the wrapped list.
					last = Integer.MIN_VALUE;
					final long k = wrapped.getLong(-pos - 1);
					int p = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask;
					while (!((k) == (key[p])))
						p = (p + 1) & mask;
					acceptOnIndex(action, p);
					c--;
				} else if (!((key[pos]) == (0))) {
					acceptOnIndex(action, last = pos);
					c--;
				}
			}
		}
		/**
		 * Shifts left entries with the specified hash code, starting at the specified
		 * position, and empties the resulting free entry.
		 *
		 * @param pos
		 *            a starting position.
		 */
		private void shiftKeys(int pos) {
			// Shift entries with the same hash.
			int last, slot;
			long curr;
			final long[] key = NonNullLong2WidgetInstanceHashMap.this.key;
			for (;;) {
				pos = ((last = pos) + 1) & mask;
				for (;;) {
					if (((curr = key[pos]) == (0))) {
						key[last] = (0);
						value[last] = null;
						return;
					}
					slot = (int) it.unimi.dsi.fastutil.HashCommon.mix((curr)) & mask;
					if (last <= pos ? last >= slot || slot > pos : last >= slot && slot > pos)
						break;
					pos = (pos + 1) & mask;
				}
				if (pos < last) { // Wrapped entry.
					if (wrapped == null)
						wrapped = new LongArrayList(2);
					wrapped.add(key[pos]);
				}
				key[last] = curr;
				value[last] = value[pos];
			}
		}
		public void remove() {
			if (last == -1)
				throw new IllegalStateException();
			if (last == n) {
				containsNullKey = false;
				value[n] = null;
			} else if (pos >= 0)
				shiftKeys(last);
			else {
				// We're removing wrapped entries.
				NonNullLong2WidgetInstanceHashMap.this.remove(wrapped.getLong(-pos - 1));
				last = -1; // Note that we must not decrement size
				return;
			}
			size--;
			last = -1; // You can no longer remove this entry.
		}
		public int skip(final int n) {
			int i = n;
			while (i-- != 0 && hasNext())
				nextEntry();
			return n - i - 1;
		}
	}
	private final class EntryIterator extends NonNullLong2WidgetInstanceHashMap.MapIterator<Consumer<? super Entry<WidgetInstance<?>>>>
		implements
		ObjectIterator<Entry<WidgetInstance<?>>> {
		private NonNullLong2WidgetInstanceHashMap.MapEntry entry;
		@Override
		public NonNullLong2WidgetInstanceHashMap.MapEntry next() {
			return entry = new NonNullLong2WidgetInstanceHashMap.MapEntry(nextEntry());
		}
		// forEachRemaining inherited from MapIterator superclass.
		@Override
		final void acceptOnIndex(final Consumer<? super Long2ObjectMap.Entry<WidgetInstance<?>>> action, final int index) {
			action.accept(entry = new NonNullLong2WidgetInstanceHashMap.MapEntry(index));
		}
		@Override
		public void remove() {
			super.remove();
			entry.index = -1; // You cannot use a deleted entry.
		}
	}
	private final class FastEntryIterator extends NonNullLong2WidgetInstanceHashMap.MapIterator<Consumer<? super Entry<WidgetInstance<?>>>>
		implements
		ObjectIterator<Long2ObjectMap.Entry<WidgetInstance<?>>> {
		private final NonNullLong2WidgetInstanceHashMap.MapEntry entry = new NonNullLong2WidgetInstanceHashMap.MapEntry();
		@Override
		public NonNullLong2WidgetInstanceHashMap.MapEntry next() {
			entry.index = nextEntry();
			return entry;
		}
		// forEachRemaining inherited from MapIterator superclass.
		@Override
		final void acceptOnIndex(final Consumer<? super Long2ObjectMap.Entry<WidgetInstance<?>>> action, final int index) {
			entry.index = index;
			action.accept(entry);
		}
	}
	private abstract class MapSpliterator<ConsumerType, SplitType extends NonNullLong2WidgetInstanceHashMap.MapSpliterator<ConsumerType, SplitType>> {
		/**
		 * The index (which bucket) of the next item to give to the action. Unlike
		 * {@link SetIterator}, this counts up instead of down.
		 */
		int pos = 0;
		/** The maximum bucket (exclusive) to iterate to */
		int max = n;
		/** An upwards counter counting how many we have given */
		int c = 0;
		/** A boolean telling us whether we should return the null key. */
		boolean mustReturnNull = NonNullLong2WidgetInstanceHashMap.this.containsNullKey;
		boolean hasSplit = false;
		MapSpliterator() {
		}
		MapSpliterator(int pos, int max, boolean mustReturnNull, boolean hasSplit) {
			this.pos = pos;
			this.max = max;
			this.mustReturnNull = mustReturnNull;
			this.hasSplit = hasSplit;
		}
		abstract void acceptOnIndex(final ConsumerType action, final int index);
		abstract SplitType makeForSplit(int pos, int max, boolean mustReturnNull);
		public boolean tryAdvance(final ConsumerType action) {
			if (mustReturnNull) {
				mustReturnNull = false;
				++c;
				acceptOnIndex(action, n);
				return true;
			}
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			while (pos < max) {
				if (!((key[pos]) == (0))) {
					++c;
					acceptOnIndex(action, pos++);
					return true;
				}
				++pos;
			}
			return false;
		}
		public void forEachRemaining(final ConsumerType action) {
			if (mustReturnNull) {
				mustReturnNull = false;
				++c;
				acceptOnIndex(action, n);
			}
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			while (pos < max) {
				if (!((key[pos]) == (0))) {
					acceptOnIndex(action, pos);
					++c;
				}
				++pos;
			}
		}
		public long estimateSize() {
			if (!hasSplit) {
				// Root spliterator; we know how many are remaining.
				return size - c;
			} else {
				// After we split, we can no longer know exactly how many we have (or at least
				// not efficiently).
				// (size / n) * (max - pos) aka currentTableDensity * numberOfBucketsLeft seems
				// like a good estimate.
				return Math.min(size - c, (long) (((double) realSize() / n) * (max - pos)) + (mustReturnNull ? 1 : 0));
			}
		}
		public SplitType trySplit() {
			if (pos >= max - 1)
				return null;
			int retLen = (max - pos) >> 1;
			if (retLen <= 1)
				return null;
			int myNewPos = pos + retLen;
			int retPos = pos;
			int retMax = myNewPos;
			// Since null is returned first, and the convention is that the returned split
			// is the prefix of elements,
			// the split will take care of returning null (if needed), and we won't return
			// it anymore.
			SplitType split = makeForSplit(retPos, retMax, mustReturnNull);
			this.pos = myNewPos;
			this.mustReturnNull = false;
			this.hasSplit = true;
			return split;
		}
		public long skip(long n) {
			if (n < 0)
				throw new IllegalArgumentException("Argument must be nonnegative: " + n);
			if (n == 0)
				return 0;
			long skipped = 0;
			if (mustReturnNull) {
				mustReturnNull = false;
				++skipped;
				--n;
			}
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			while (pos < max && n > 0) {
				if (!((key[pos++]) == (0))) {
					++skipped;
					--n;
				}
			}
			return skipped;
		}
	}
	private final class EntrySpliterator
		extends
		NonNullLong2WidgetInstanceHashMap.MapSpliterator<Consumer<? super Entry<WidgetInstance<?>>>, NonNullLong2WidgetInstanceHashMap.EntrySpliterator>
		implements
		ObjectSpliterator<Entry<WidgetInstance<?>>> {
		private static final int POST_SPLIT_CHARACTERISTICS = ObjectSpliterators.SET_SPLITERATOR_CHARACTERISTICS
															  & ~java.util.Spliterator.SIZED;
		EntrySpliterator() {
		}
		EntrySpliterator(int pos, int max, boolean mustReturnNull, boolean hasSplit) {
			super(pos, max, mustReturnNull, hasSplit);
		}
		@Override
		public int characteristics() {
			return hasSplit ? POST_SPLIT_CHARACTERISTICS : ObjectSpliterators.SET_SPLITERATOR_CHARACTERISTICS;
		}
		@Override
		final void acceptOnIndex(final Consumer<? super Long2ObjectMap.Entry<WidgetInstance<?>>> action, final int index) {
			action.accept(new NonNullLong2WidgetInstanceHashMap.MapEntry(index));
		}
		@Override
		final NonNullLong2WidgetInstanceHashMap.EntrySpliterator makeForSplit(int pos, int max, boolean mustReturnNull) {
			return new NonNullLong2WidgetInstanceHashMap.EntrySpliterator(pos, max, mustReturnNull, true);
		}
	}
	private final class MapEntrySet extends AbstractObjectSet<Entry<WidgetInstance<?>>> implements FastEntrySet<WidgetInstance<?>> {
		@Override
		public ObjectIterator<Long2ObjectMap.Entry<WidgetInstance<?>>> iterator() {
			return new NonNullLong2WidgetInstanceHashMap.EntryIterator();
		}
		@Override
		public ObjectIterator<Long2ObjectMap.Entry<WidgetInstance<?>>> fastIterator() {
			return new NonNullLong2WidgetInstanceHashMap.FastEntryIterator();
		}
		@Override
		public ObjectSpliterator<Long2ObjectMap.Entry<WidgetInstance<?>>> spliterator() {
			return new NonNullLong2WidgetInstanceHashMap.EntrySpliterator();
		}
		//
		@Override
		@SuppressWarnings("unchecked")
		public boolean contains(final Object o) {
			if (!(o instanceof Map.Entry<?, ?> e))
				return false;
			if (!(e.getKey() instanceof Long k_))
				return false;
			final long k = k_;
			if (!(e.getValue() instanceof WidgetInstance<?> v))
				return false;
			if (((k) == (0)))
				return NonNullLong2WidgetInstanceHashMap.this.containsNullKey && Objects.equals(value[n], v);
			long curr;
			var key = NonNullLong2WidgetInstanceHashMap.this.key;
			int pos;
			// The starting point.
			if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
				return false;
			if (((k) == (curr)))
				return Objects.equals(value[pos], v);
			// There's always an unused entry.
			while (true) {
				if (((curr = key[pos = (pos + 1) & mask]) == (0)))
					return false;
				if (((k) == (curr)))
					return Objects.equals(value[pos], v);
			}
		}
		@Override
		@SuppressWarnings("unchecked")
		public boolean remove(final Object o) {
			if (!(o instanceof Map.Entry<?, ?> e))
				return false;
			if (!(e.getKey() instanceof Long k_))
				return false;
			final long k = k_;
			if (!(e.getValue() instanceof WidgetInstance<?> v))
				return false;
			if (((k) == (0))) {
				if (containsNullKey && Objects.equals(value[n], v)) {
					removeNullEntry();
					return true;
				}
				return false;
			}
			long curr;
			final long[] key = NonNullLong2WidgetInstanceHashMap.this.key;
			int pos;
			// The starting point.
			if (((curr = key[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((k)) & mask]) == (0)))
				return false;
			if (((curr) == (k))) {
				if (Objects.equals(value[pos], v)) {
					removeEntry(pos);
					return true;
				}
				return false;
			}
			while (true) {
				if (((curr = key[pos = (pos + 1) & mask]) == (0)))
					return false;
				if (((curr) == (k))) {
					if (Objects.equals(value[pos], v)) {
						removeEntry(pos);
						return true;
					}
				}
			}
		}
		@Override
		public int size() {
			return size;
		}
		@Override
		public void clear() {
			NonNullLong2WidgetInstanceHashMap.this.clear();
		}
		/** {@inheritDoc} */
		@Override
		public void forEach(final Consumer<? super Long2ObjectMap.Entry<WidgetInstance<?>>> consumer) {
			if (containsNullKey)
				consumer.accept(new AbstractLong2ObjectMap.BasicEntry<>(key[n], value[n]));
			for (int pos = n; pos-- != 0;)
				if (!((key[pos]) == (0)))
					consumer.accept(new AbstractLong2ObjectMap.BasicEntry<>(key[pos], value[pos]));
		}
		/** {@inheritDoc} */
		@Override
		public void fastForEach(final Consumer<? super Long2ObjectMap.Entry<WidgetInstance<?>>> consumer) {
			var entry = new MapEntry();
			if (containsNullKey) {
				entry.index = n;
				consumer.accept(entry);
			}
			for (int pos = n; pos-- != 0;)
				if (!((key[pos]) == (0))) {
					entry.index = pos;
					consumer.accept(entry);
				}
		}
	}
	@Override
	public FastEntrySet<WidgetInstance<?>> long2ObjectEntrySet() {
		return entries;
	}
	/**
	 * An iterator on keys.
	 *
	 * <p>
	 * We simply override the
	 * {@link java.util.ListIterator#next()}/{@link java.util.ListIterator#previous()}
	 * methods (and possibly their type-specific counterparts) so that they return
	 * keys instead of entries.
	 */
	private final class KeyIterator extends NonNullLong2WidgetInstanceHashMap.MapIterator<LongConsumer> implements LongIterator {
		public KeyIterator() {
			super();
		}
		// forEachRemaining inherited from MapIterator superclass.
		// Despite the superclass declared with generics, the way Java inherits and
		// generates bridge methods avoids the boxing/unboxing
		@Override
		final void acceptOnIndex(final java.util.function.LongConsumer action, final int index) {
			action.accept(key[index]);
		}
		@Override
		public long nextLong() {
			return key[nextEntry()];
		}
	}
	private final class KeySpliterator extends NonNullLong2WidgetInstanceHashMap.MapSpliterator<LongConsumer, NonNullLong2WidgetInstanceHashMap.KeySpliterator>
		implements
		LongSpliterator {
		private static final int POST_SPLIT_CHARACTERISTICS = LongSpliterators.SET_SPLITERATOR_CHARACTERISTICS
															  & ~java.util.Spliterator.SIZED;
		KeySpliterator() {
		}
		KeySpliterator(int pos, int max, boolean mustReturnNull, boolean hasSplit) {
			super(pos, max, mustReturnNull, hasSplit);
		}
		@Override
		public int characteristics() {
			//noinspection MagicConstant
			return hasSplit ? POST_SPLIT_CHARACTERISTICS : LongSpliterators.SET_SPLITERATOR_CHARACTERISTICS;
		}
		@Override
		final void acceptOnIndex(final java.util.function.LongConsumer action, final int index) {
			action.accept(key[index]);
		}
		@Override
		final NonNullLong2WidgetInstanceHashMap.KeySpliterator makeForSplit(int pos, int max, boolean mustReturnNull) {
			return new NonNullLong2WidgetInstanceHashMap.KeySpliterator(pos, max, mustReturnNull, true);
		}
	}
	private final class KeySet extends AbstractLongSet {
		@Override
		public LongIterator iterator() {
			return new NonNullLong2WidgetInstanceHashMap.KeyIterator();
		}
		@Override
		public LongSpliterator spliterator() {
			return new NonNullLong2WidgetInstanceHashMap.KeySpliterator();
		}
		/** {@inheritDoc} */
		@Override
		public void forEach(final java.util.function.LongConsumer consumer) {
			if (containsNullKey)
				consumer.accept(key[n]);
			for (int pos = n; pos-- != 0;) {
				final long k = key[pos];
				if (!((k) == (0)))
					consumer.accept(k);
			}
		}
		@Override
		public int size() {
			return size;
		}
		@Override
		public boolean contains(long k) {
			return containsKey(k);
		}
		@Override
		public boolean remove(long k) {
			final int oldSize = size;
			NonNullLong2WidgetInstanceHashMap.this.remove(k);
			return size != oldSize;
		}
		@Override
		public void clear() {
			NonNullLong2WidgetInstanceHashMap.this.clear();
		}
	}
	@Override
	public LongSet keySet() {
		return keys;
	}
	/**
	 * An iterator on values.
	 *
	 * <p>
	 * We simply override the
	 * {@link java.util.ListIterator#next()}/{@link java.util.ListIterator#previous()}
	 * methods (and possibly their type-specific counterparts) so that they return
	 * values instead of entries.
	 */
	private final class ValueIterator extends NonNullLong2WidgetInstanceHashMap.MapIterator<WidgetInstanceConsumer>
		implements ObjectIterator<WidgetInstance<?>> {
		public ValueIterator() {
			super();
		}
		// forEachRemaining inherited from MapIterator superclass.
		// Despite the superclass declared with generics, the way Java inherits and
		// generates bridge methods avoids the boxing/unboxing
		@Override
		final void acceptOnIndex(final WidgetInstanceConsumer action, final int index) {
			action.accept(value[index]);
		}
		@Override
		public WidgetInstance<?> next() {
			return value[nextEntry()];
		}
	}
	private final class ValueSpliterator extends NonNullLong2WidgetInstanceHashMap.MapSpliterator<Consumer<? super WidgetInstance<?>>, NonNullLong2WidgetInstanceHashMap.ValueSpliterator>
		implements
		ObjectSpliterator<WidgetInstance<?>> {
		private static final int POST_SPLIT_CHARACTERISTICS = ObjectSpliterators.COLLECTION_SPLITERATOR_CHARACTERISTICS
															  & ~java.util.Spliterator.SIZED;
		ValueSpliterator() {
		}
		ValueSpliterator(int pos, int max, boolean mustReturnNull, boolean hasSplit) {
			super(pos, max, mustReturnNull, hasSplit);
		}
		@Override
		public int characteristics() {
			return hasSplit ? POST_SPLIT_CHARACTERISTICS : ObjectSpliterators.COLLECTION_SPLITERATOR_CHARACTERISTICS;
		}
		@Override
		final void acceptOnIndex(final Consumer<? super WidgetInstance<?>> action, final int index) {
			action.accept(value[index]);
		}
		@Override
		final NonNullLong2WidgetInstanceHashMap.ValueSpliterator makeForSplit(int pos, int max, boolean mustReturnNull) {
			return new NonNullLong2WidgetInstanceHashMap.ValueSpliterator(pos, max, mustReturnNull, true);
		}
	}
	@Override
	public ObjectCollection<WidgetInstance<?>> values() {
		return values;
	}
	/**
	 * Rehashes the map, making the table as small as possible.
	 *
	 * <p>
	 * This method rehashes the table to the smallest size satisfying the load
	 * factor. It can be used when the set will not be changed anymore, so to
	 * optimize access speed and size.
	 *
	 * <p>
	 * If the table size is already the minimum possible, this method does nothing.
	 *
	 * @return true if there was enough memory to trim the map.
	 * @see #trim(int)
	 */
	public boolean trim() {
		return trim(size);
	}
	/**
	 * Rehashes this map if the table is too large.
	 *
	 * <p>
	 * Let <var>N</var> be the smallest table size that can hold
	 * <code>max(n,{@link #size()})</code> entries, still satisfying the load
	 * factor. If the current table size is smaller than or equal to <var>N</var>,
	 * this method does nothing. Otherwise, it rehashes this map in a table of size
	 * <var>N</var>.
	 *
	 * <p>
	 * This method is useful when reusing maps. {@linkplain #clear() Clearing a map}
	 * leaves the table size untouched. If you are reusing a map many times, you can
	 * call this method with a typical size to avoid keeping around a very large
	 * table just because of a few large transient maps.
	 *
	 * @param n
	 *            the threshold for the trimming.
	 * @return true if there was enough memory to trim the map.
	 * @see #trim()
	 */
	public boolean trim(final int n) {
		final int l = HashCommon.nextPowerOfTwo((int) Math.ceil(n / DEFAULT_LOAD_FACTOR));
		if (l >= this.n || size > maxFill(l, DEFAULT_LOAD_FACTOR))
			return true;
		try {
			rehash(l);
		} catch (OutOfMemoryError cantDoIt) {
			return false;
		}
		return true;
	}
	/**
	 * Rehashes the map.
	 *
	 * <p>
	 * This method implements the basic rehashing strategy, and may be overridden by
	 * subclasses implementing different rehashing strategies (e.g., disk-based
	 * rehashing). However, you should not override this method unless you
	 * understand the internal workings of this class.
	 *
	 * @param newN
	 *            the new size
	 */
	@SuppressWarnings("unchecked")
	protected void rehash(final int newN) {
		var key = this.key;
		var value = this.value;
		final int mask = newN - 1; // Note that this is used by the hashing macro
		var newKey = new long[newN + 1];
		var newValue = new WidgetInstance<?>[newN + 1];
		int i = n, pos;
		for (int j = realSize(); j-- != 0;) {
			while (((key[--i]) == (0)));
			if (!((newKey[pos = (int) it.unimi.dsi.fastutil.HashCommon.mix((key[i])) & mask]) == (0)))
				while (!((newKey[pos = (pos + 1) & mask]) == (0)));
			newKey[pos] = key[i];
			newValue[pos] = value[i];
		}
		newValue[newN] = value[n];
		n = newN;
		this.mask = mask;
		maxFill = maxFill(n, DEFAULT_LOAD_FACTOR);
		this.key = newKey;
		this.value = newValue;
	}
	/**
	 * Returns a deep copy of this map.
	 *
	 * <p>
	 * This method performs a deep copy of this hash map; the data stored in the
	 * map, however, is not cloned. Note that this makes a difference only for
	 * object keys.
	 *
	 * @return a deep copy of this map.
	 */
	@Override
	public NonNullLong2WidgetInstanceHashMap clone() {
		NonNullLong2WidgetInstanceHashMap c;
		try {
			c = (NonNullLong2WidgetInstanceHashMap) super.clone();
		} catch (CloneNotSupportedException cantHappen) {
			throw new InternalError();
		}
		c.containsNullKey = containsNullKey;
		c.key = key.clone();
		c.value = value.clone();
		return c;
	}

	@Override
	public boolean equals(Object o) {
		return super.equals(o);
	}

	/**
	 * Returns a hash code for this map.
	 *
	 * This method overrides the generic method provided by the superclass. Since
	 * {@code equals()} is not overriden, it is important that the value returned by
	 * this method is the same value as the one returned by the overriden method.
	 *
	 * @return a hash code for this map.
	 */
	@Override
	public int hashCode() {
		int h = 0;
		for (int j = realSize(), i = 0; j-- != 0;) {
			while (((key[i]) == (0)))
				i++;
			int t = it.unimi.dsi.fastutil.HashCommon.long2int(key[i]);
			t ^= (value[i] == null ? 0 : (value[i]).hashCode());
			h += t;
			i++;
		}
		// Zero / null keys have hash zero.
		if (containsNullKey)
			h += (value[n] == null ? 0 : (value[n]).hashCode());
		return h;
	}

	private class ValueCollection extends AbstractObjectCollection<WidgetInstance<?>> {
		@Override
		public ObjectIterator<WidgetInstance<?>> iterator() {
			return new ValueIterator();
		}

		@Override
		public ObjectSpliterator<WidgetInstance<?>> spliterator() {
			return new ValueSpliterator();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public void forEach(final Consumer<? super WidgetInstance<?>> consumer) {
			if (containsNullKey) {
				consumer.accept(value[n]);
			}
			for (int pos = n; pos-- != 0; )
				if (!((key[pos]) == (0))) {
					consumer.accept(value[pos]);
				}
		}

		@Override
		public int size() {
			return size;
		}

		@Override
		public boolean contains(Object v) {
			return containsValue(v);
		}

		@Override
		public void clear() {
			NonNullLong2WidgetInstanceHashMap.this.clear();
		}
	}
}
