package dev.pfaff.altargui.util;

import java.util.function.Consumer;

/**
 * @see java.util.stream.Sink
 */
public interface Sink<T> extends Consumer<T> {
	/**
	 * Resets the sink state to receive a fresh data set.  This must be called
	 * before sending any data to the sink.  After calling {@link #end()},
	 * you may call this method to reset the sink for another calculation.
	 * @param size The exact size of the data to be pushed downstream, if
	 * known or {@code -1} if unknown or infinite.
	 *
	 * <p>Prior to this call, the sink must be in the initial state, and after
	 * this call it is in the active state.
	 */
	default void begin(long size) {}

	/**
	 * Indicates that all elements have been pushed.  If the {@code Sink} is
	 * stateful, it should send any stored state downstream at this time, and
	 * should clear any accumulated state (and associated resources).
	 *
	 * <p>Prior to this call, the sink must be in the active state, and after
	 * this call it is returned to the initial state.
	 */
	default void end() {}

	/**
	 * Indicates that this {@code Sink} does not wish to receive any more data.
	 * <p>
	 * <strong>Note</strong>: The default implementation always returns false.
	 *
	 * @return true if cancellation is requested
	 */
	default boolean cancellationRequested() {
		return false;
	}

	/**
	 * Accepts an int value.
	 * <p>
	 * <strong>Note</strong>: The default implementation throws IllegalStateException.
	 *
	 * @throws IllegalStateException if this sink does not accept int values
	 */
	default void accept(int value) {
		throw new IllegalStateException("called wrong accept method");
	}

	/**
	 * Accepts a long value.
	 * <p>
	 * <strong>Note</strong>: The default implementation throws IllegalStateException.
	 *
	 * @throws IllegalStateException if this sink does not accept long values
	 */
	default void accept(long value) {
		throw new IllegalStateException("called wrong accept method");
	}

	/**
	 * Accepts a double value.
	 * <p>
	 * <strong>Note</strong>: The default implementation throws IllegalStateException.
	 *
	 * @throws IllegalStateException if this sink does not accept double values
	 */
	default void accept(double value) {
		throw new IllegalStateException("called wrong accept method");
	}
}
