package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.widgets.state.StatefulWidgetInstance;
import dev.pfaff.altargui.handlers.CharTypedHandler;
import dev.pfaff.altargui.handlers.KeyPressHandler;
import dev.pfaff.altargui.handlers.MousePressHandler;
import dev.pfaff.altargui.handlers.MouseReleaseHandler;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMod;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.listenable.AbstractValue;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Value;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.ref.WeakReference;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

import static dev.pfaff.altargui.GuiGlobals.HOST;
import static dev.pfaff.altargui.widgets.MultiChildWidget.row;

public record TextFieldWidget(@Nullable TextFieldController controller, @NonNull Consumer<String> onChanged) implements Widget {

	public static final int LINE_HEIGHT = 9;

	public TextFieldWidget() {
		this(null, ignored -> {});
	}

	public TextFieldWidget withController(@Nullable TextFieldController controller) {
		return new TextFieldWidget(controller, onChanged);
	}

	public TextFieldWidget withOnChanged(@NonNull Consumer<String> onChanged) {
		return new TextFieldWidget(controller, onChanged);
	}

	@Override
	public @NonNull WidgetInstance<TextFieldWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<TextFieldWidget> implements CharTypedHandler, KeyPressHandler, MousePressHandler, MouseReleaseHandler {
		private TextFieldController controller;
		@Nullable
		private Consumer<String> onChanged;
		@Nullable
		private ListenerDisposer onChangedDisposer;
		private final BitSet buttonsPressed = new BitSet(3);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, TextFieldWidget widget, boolean init) {
			boolean dirty = init;
			if (init || (widget.controller != null && widget.controller != controller)) {
				if (this.onChangedDisposer != null) {
					this.onChangedDisposer.close();
					this.onChangedDisposer = null;
				}
				controller = Objects.requireNonNullElseGet(widget.controller, TextFieldController::new);
			}
			if (!controller.currentState.refersTo(this)) {
				dirty = true;
				controller.currentState = new WeakReference<>(this);
			}
			if (widget.onChanged != this.onChanged) {
				if (this.onChangedDisposer != null) {
					this.onChangedDisposer.close();
				}
				this.onChangedDisposer = this.controller.content.listen(this.onChanged = widget.onChanged);
			}
			if (dirty) markNeedsRebuild();
		}

		@Override
		protected @NonNull Widget build() {
			var v = controller.buffer.toString();
			Widget widget;
			if (controller.focused) {
				int cursorI = controller.cursor;
				int selectionExtent = controller.selectionExtent;
				int selectionStart = selectionExtent == -1 ? cursorI : Math.min(cursorI, selectionExtent);
				int selectionEnd = selectionExtent == -1 ? cursorI : Math.max(cursorI, selectionExtent);
				var cursor = RectWidget.colored(Color.WHITE)
						.withAbsoluteWidth(1)
						.atRelativePosition(-0.5f, 0f)
						.withAbsoluteWidth(cursorI == 0 || cursorI == v.length() ? 0 : 1);
				Widget widget1, widget2;
				if (selectionExtent != -1) {
					var selected = TextWidget.of(v.substring(selectionStart, selectionEnd))
							.withBackgroundColor(Color.fromRgba(0x0090ff_0f));
					widget1 = cursorI < selectionExtent ? cursor : selected;
					widget2 = cursorI < selectionExtent ? selected : cursor;
				} else {
					widget1 = cursor;
					widget2 = Widgets.EMPTY;
				}
				widget = row(List.of(TextWidget.of(v.substring(0, selectionStart)),
								   widget1,
								   widget2,
								   TextWidget.of(v.substring(selectionEnd))));
			} else {
				widget = row(List.of(TextWidget.of(v)));
			}
			widget = widget.withAbsoluteHeight(LINE_HEIGHT).withMargin(2);
			widget = widget.withBackground(controller.focused ? RectWidget.GENERIC_HEAVY_HUD_BACKGROUND : RectWidget.GENERIC_HUD_BACKGROUND)
					.withBorder(Color.fromArgb(controller.focused ? 0x99_bcbcbc : 0x99_222222), 2f);
			return widget;
		}

		@Override
		public boolean charTyped(int codePoint) {
			if (!controller.focused) return false;
			return controller.insertAtCursor(codePoint);
		}

		@Override
		public boolean keyPressed(InputKey keyCode, int scanCode, InputMods modifiers) {
			if (!controller.focused) return false;
			boolean hasPrimary = modifiers.has(InputMod.PRIMARY);
			boolean hasShift = modifiers.has(InputMod.SHIFT);
			switch (keyCode) {
				case TAB -> {
					// TODO: autocompletion
				}
				case ENTER -> {
					// TODO: onSubmit
				}
				case LEFT -> {
					controller.extendSelection(-1, hasShift);
					return true;
				}
				case RIGHT -> {
					controller.extendSelection(+1, hasShift);
					return true;
				}
				case HOME -> {
					controller.selectTo(0, hasShift);
					return true;
				}
				case END -> {
					controller.selectTo(controller.buffer.length(), hasShift);
					return true;
				}
				case BACKSPACE -> {
					controller.backspace(hasPrimary);
					return true;
				}
				case DELETE -> {
					controller.delete(hasPrimary);
					return true;
				}
				case V -> {
					if (hasPrimary) {
						controller.pasteAsync();
						return true;
					}
				}
				case X, C -> {
					if (hasPrimary) {
						controller.copyAsync(keyCode == InputKey.X);
						return true;
					}
				}
			}

			return false;
		}

		@Override
		public boolean mousePressed(double mouseX, double mouseY, int button) {
			if (isInside(mouseX, mouseY)) {
				this.buttonsPressed.set(button);
				return true;
			} else {
				return false;
			}
		}

		@Override
		public boolean mouseReleased(double mouseX, double mouseY, int button) {
			var wasPressedInside = this.buttonsPressed.get(button);
			if (isInside(mouseX, mouseY)) {
				if (wasPressedInside) {
					this.buttonsPressed.clear(button);
					if (button == 0) {
						if (!this.controller.focused) {
							this.controller.focused = true;
							markNeedsRebuild();
						}
					}
				}
				return true;
			} else {
				if (!wasPressedInside) {
					if (this.controller.focused) {
						this.controller.focused = false;
						markNeedsRebuild();
					}
				} else {
					this.buttonsPressed.clear(button);
				}
				return false;
			}
		}

		private void markNeedsRebuildAccessible() {
			this.markNeedsRebuild();
		}
	}

	public static final class TextFieldController {
		private static final WeakReference<@Nullable State> NULL_STATE = new WeakReference<>(null);

		private WeakReference<State> currentState = NULL_STATE;

		private final @NonNull StringBuilder buffer;

		private final ContentValue content;

		private int cursor;
		private int selectionExtent;

		private boolean focused;

		public TextFieldController() {
			this.buffer = new StringBuilder();
			this.content = new ContentValue(this);
		}

		public TextFieldController(@NonNull String initialValue) {
			this.buffer = new StringBuilder(initialValue);
			this.content = new ContentValue(this);
		}

		public Value<String> content() {
			return this.content;
		}

		public boolean selectTo(int position, boolean isShiftDown) {
			if (position == cursor) return false;

			if (position >= 0 && position <= buffer.length()) {
				if (isShiftDown && selectionExtent == -1) {
					selectionExtent = cursor;
				}

				cursor = position;
				markVisuallyDirty();
				return true;
			}
			return false;
		}

		public boolean extendSelection(int delta, boolean isShiftDown) {
			if (!isShiftDown && selectionExtent != -1) {
				if (delta == -1) {
					cursor = Math.min(cursor, selectionExtent);
					selectionExtent = -1;
					markVisuallyDirty();
					return true;
				} else if (delta == 1) {
					cursor = Math.max(cursor, selectionExtent);
					selectionExtent = -1;
					markVisuallyDirty();
					return true;
				}
			}
			return selectTo(cursor + delta, isShiftDown);
		}

		private boolean deleteSelection() {
			if (selectionExtent != -1 && selectionExtent != cursor) {
				buffer.delete(selectionStart(), selectionEnd());
				cursor = selectionStart();
				selectionExtent = -1;
				markContentDirty(true);
				return true;
			}
			return false;
		}

		public boolean backspace(boolean isCtrlDown) {
			if (cursor == 0) return false;
            if (isCtrlDown) {
                buffer.delete(0, cursor);
                selectionExtent = -1;
                cursor = 0;
				markContentDirty(true);
                return true;
            } else {
                if (selectionExtent != -1) {
                    return deleteSelection();
                } else if (cursor > 0) {
                    buffer.deleteCharAt(--cursor);
					markContentDirty(true);
                    return true;
                }
            }
			return false;
        }

		public boolean delete(boolean isCtrlDown) {
            if (buffer.isEmpty()) return false;
            if (isCtrlDown) {
                var len = buffer.length();
                buffer.delete(cursor, len);
                selectionExtent = -1;
				markContentDirty(true);
                return true;
            } else {
                if (selectionExtent != -1) {
                    deleteSelection();
					markContentDirty(true);
                    return true;
                } else if (cursor < buffer.length()) {
                    buffer.deleteCharAt(cursor);
					markContentDirty(true);
                    return true;
                }
            }
            return false;
		}

		public int selectionStart() {
			return Math.min(cursor, selectionExtent);
		}

		public int selectionEnd() {
			return Math.max(cursor, selectionExtent);
		}

		public String getSelection() {
			if (selectionExtent == -1 || selectionExtent == cursor) return "";
			return buffer.substring(selectionStart(), selectionEnd());
		}

		public void pasteAsync() {
			Thread.startVirtualThread(() -> {
				var s = HOST.getClipboardString();
				if (s == null) return;
				if (selectionStart() != -1) {
					buffer.replace(selectionStart(),
											  selectionEnd(),
											  s);
					cursor = selectionStart() + s.length();
					selectionExtent = -1;
					markContentDirty(true);
				} else if (!s.isEmpty()) {
					buffer.insert(cursor, s);
					cursor += s.length();
					markContentDirty(true);
				}
			});
		}

		public void copyAsync(boolean cut) {
			var s = getSelection();
			if (!s.isEmpty()) {
				Thread.startVirtualThread(() -> HOST.setClipboardString(s));
				if (cut) {
					buffer.delete(selectionStart(), selectionEnd());
					cursor = selectionStart();
					selectionExtent = -1;
					markContentDirty(true);
				}
			}
		}

		private void markVisuallyDirty() {
			var state = currentState.get();
			if (state != null) state.markNeedsRebuildAccessible();
		}

		private void markContentDirty(boolean notify) {
			markVisuallyDirty();
			if (notify) this.content.notifyListeners();
		}

		public boolean insertAtCursor(int codePoint) {
			if (!Character.isDefined(codePoint)) return false;
			if (Character.isISOControl(codePoint)) return false;

			// TODO: make sure we avoid all control characters.
			if (selectionExtent != -1) {
				buffer.replace(selectionStart(),
						selectionEnd(),
						Character.toString(codePoint));
				cursor = selectionStart() + 1;
				selectionExtent = -1;
			} else {
				buffer.insert(cursor++, Character.toString(codePoint));
			}
			markContentDirty(true);
			return true;
		}

		public boolean insertAtCursor(char chr) {
			return insertAtCursor((int) chr);
		}

		public void setFocused(boolean focused) {
			if (this.focused != focused) {
				this.focused = focused;
				markVisuallyDirty();
			}
		}

		private static final class ContentValue extends AbstractValue<String> {
			private final TextFieldController controller;
			private final StringBuilder buffer;

			private ContentValue(TextFieldController controller) {
				this.controller = controller;
				this.buffer = controller.buffer;
			}

			@Override
			public String read() {
				return this.buffer.toString();
			}

			@Override
			public boolean writeNoNotify(String value) {
				boolean updated = !value.contentEquals(this.buffer);
				this.buffer.setLength(0);
				this.buffer.append(value);
				this.controller.markContentDirty(false);
				return updated;
			}

			private void notifyListeners() {
				forEachListener(listener -> {
					switch (listener) {
						case Runnable runnable -> runnable.run();
						case Consumer consumer -> consumer.accept(read());
						default -> {}
					}
				});
			}
		}
	}
}
